using System;
using System.Text;
using BenjaminSchroeter.Flow.DataTypes;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// The base class for all generators that generate code .
    /// </summary>
    public abstract class BlockGeneratorBase
    {
        protected HardwareDescriptionModel hardware;
        
        /// <summary>
        /// All code must be written to this string builder
        /// </summary>
        internal protected StringBuilder sb;

        /// <summary>
        /// Initializes a new instance of the <see cref="BlockGeneratorBase"/> class.
        /// </summary>
        /// <param name="hardware">The hardware.</param>
        protected BlockGeneratorBase(HardwareDescriptionModel hardware)
        {
            this.hardware = hardware;
        }

        /// <summary>
        /// Generates the behavior of the element represented by this generator
        /// </summary>
        /// <returns></returns>
        public string GenerateBehavior()
        {
            sb = new StringBuilder(); 
            generateBehavior();
            return sb.ToString();
        }

        public virtual void generateBehavior()
        {
            throw new NotImplementedException();
        }

        #region generation helpers
        /// <summary>
        /// Creates a Variable declaration in the output code.
        /// </summary>
        protected void VariableDeclaration(DataTypeString type, Store store, string variableName)
        {
            string ctype = type.AsCType(store);
            string appendix = "";
            if (type.AsDslType(store) == DslDataType.String)
                appendix = "[STRING_LENGTH]";

            sb.AppendFormatLine("{0} {1}{2};", ctype, variableName, appendix);
        }

        /// <summary>
        /// Creates a Function call in the output code.
        /// </summary>
        protected void FunctionCall(string call, CallArguments args)
        {
                sb.AppendFormatLine("{0}({1});", call, args.GetArgumentList());
        }
        /// <summary>
        /// Creates a Function call in the output code with a return value.
        /// </summary>
        /// <param name="returnVariable">The returnvalue of the method will be assigend to the given variable. The Variable must be declared somewhere before in the code.</param>
        protected void FunctionCall(string call, CallArguments args, string returnVariable)
        {
            sb.AppendFormatLine("{2} = {0}({1});", call, args.GetArgumentList(), returnVariable);
        }
        /// <summary>
        /// Creates a Function call in the output code.
        /// </summary>
        protected void FunctionCall(CallParameter call, CallArguments args)
        {
            if (call.HasFunction)
                FunctionCall(call.Function, args);
        }
        /// <summary>
        /// Creates a Function call with a single argument in the output code.
        /// </summary>
        protected void FunctionCall(CallParameter call, string arg, DslDataType type)
        {
            if (call.HasFunction)
            {
                AddCallArgumentType addingType = call.ParameterReferenceType == ""
                                                     ? AddCallArgumentType.ByValue
                                                     : AddCallArgumentType.ByReference;
                if (type == DslDataType.String)
                    addingType = AddCallArgumentType.ByValue;

                CallArguments args = new CallArguments();
                args.Add(1, arg, addingType); // first and singel argument
                sb.AppendFormatLine("{0}({1});", call.Function, args.GetArgumentList());
            }
        }

        /// <summary>
        /// Creates a Comment in the output code.
        /// </summary>
        protected void Comment(string str)
        {
            sb.AppendLine("// " + str);
        }

        /// <summary>
        /// Creates an Assignments in the output code.
        /// 
        /// For strings <code>memcpy()</code> is used with a fixed length of STRING_LENGTH.
        /// </summary>
        protected void Assignment(string from, string into, DslDataType dslType)
        {
            Assignment(from, into, dslType, -1);
        }
        /// <summary>
        /// Creates an Assignments in the output code.
        /// 
        /// For strings <code>memcpy()</code> is used.
        /// </summary>
        /// <param name="stringLength">Only for dslType == String this parameter is used to copy only that much bytes. Can be -1 for a default value of STRING_LENGTH.</param>
        protected void Assignment(string from, string into, DslDataType dslType, int stringLength)
        {
            switch (dslType)
            {
                case DslDataType.Boolean:
                case DslDataType.Number:
                case DslDataType.Other:
                    sb.AppendFormatLine("{0} = {1};", into, from);
                    break;
                case DslDataType.String:
                    // determin how much bytes to copy
                    string length;
                    if (stringLength < 0)
                        length = "STRING_LENGTH"; // default value for all strings
                    else
                        length = stringLength.ToString();

                    sb.AppendFormatLine("memcpy({0}, {1}, {2});", into, from, length);
                    break;
                case DslDataType.Record:
                    throw new NotImplementedException("Cannot generate assignment for Records.");
                default:
                    throw new ArgumentOutOfRangeException("dslType");
            }
        }
        #endregion

        #region datatypes
        protected string ParameterValue(bool b)
        {
            return ParameterValue( b ? "true" : "false" , DslDataType.Boolean);
        }
            
        protected string ParameterValue(string str, DslDataType type)
        {
            switch (type)
            {
                case DslDataType.Boolean:
                    return str;
                
                case DslDataType.Number:
                    return str;

                case DslDataType.String:
                    return "\"" + str + "\"";

                case DslDataType.Record:
                case DslDataType.Other:
                default:
                    throw new ArgumentOutOfRangeException("type");
            }
        }
        #endregion
    }
}
