using System;
using System.Collections.Generic;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using Input=BenjaminSchroeter.Flow.DataFlow.Input;
using Output=BenjaminSchroeter.Flow.DataFlow.Output;
using Variable=BenjaminSchroeter.Flow.DataFlow.Variable;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// This factory creates Generators (see <see cref="BlockGeneratorBase"/>)
    /// for given elements from the DataFlow.
    /// </summary>
    public class BlockGeneratorFactory
    {
        /// <summary>
        /// A list of UniqueNameProvider is maintained. 
        /// One for each function call (scope). And one special (<see cref="GLOBAL"/>) for the global scope.
        /// </summary>
        private Dictionary<string ,UniqueNameProvider> nameProviders;
        public const string GLOBAL = "FunctionNames";
        
        private HardwareDescriptionModel hardware;

        /// <summary>
        /// Initializes a new instance of the <see cref="BlockGeneratorFactory"/> class.
        /// </summary>
        /// <param name="hardware">The HardwareDescriptionModel.</param>
        /// <param name="dataflow">The DataFlowModel.</param>
        public BlockGeneratorFactory(HardwareDescriptionModel hardware, DataFlowModel dataflow)
        {
            nameProviders = new Dictionary<string, UniqueNameProvider>();
            this.hardware = hardware;

            var x = from t in hardware.DataTypes
                    select t.Name;
            dataflow.DataTypes = x.ToArray();
        }

        /// <summary>
        /// Gets the generator for a FlowElement.
        /// </summary>
        /// <param name="flowElement">The flow element.</param>
        /// <param name="nameScope">The scope used by the UniqueNameProvider.</param>
        /// <returns></returns>
        public BlockGeneratorBase GetGenerator(FlowElement flowElement, string nameScope)
        {
            // find the name provider to use
            UniqueNameProvider nameProvider = GetNameProvider(nameScope);

            // elements from hardware description
            if (flowElement is Input)
                return new InputGenerator((Input) flowElement, nameProvider, hardware);

            if (flowElement is InputEvent || flowElement is FunctionCall)
                return new EventGenerator((FieldedDefinitionBase)flowElement, nameProvider, hardware);

            if (flowElement is Output)
                return new OutputGenerator((Output)flowElement, nameProvider, hardware);

            if (flowElement is OutputEvent || flowElement is FunctionReturn)
                return new OutputEventGenerator((FieldedDefinitionBase)flowElement, nameProvider, hardware);

            if (flowElement is Record)
                return new RecordGenerator((Record)flowElement, nameProvider, hardware);

            if (flowElement is Variable)
            {
                Variable v = (Variable) flowElement;
                if (v.FromHardwareDescription)
                    return new HardwareVariableGenerator(v, nameProvider, hardware);
                else
                    return new DatastructureVariableGenerator(v, nameProvider, hardware);
            }

            // transformations
            if (flowElement is Operation)
                return new OperationGenerator((Operation)flowElement, nameProvider, hardware);

            if (flowElement is Formula)
                return new FormulaGenerator((Formula) flowElement, nameProvider, GetNameProvider(GLOBAL), hardware);

            if (flowElement is CodeBlock)
                return new CodeBlockGenerator((CodeBlock)flowElement, nameProvider, GetNameProvider(GLOBAL), hardware);

            // misc / MasterTrigger
            if (flowElement is MasterTrigger)
                return new EmptyGenerator();



            // if we reach this position: a FlowElement is used without a Generator implementation
            throw new NotImplementedException();
        }

        /// <summary>
        /// Gets the generator for an InputEvent
        /// </summary>
        /// <param name="flowElement">The flow element.</param>
        /// <param name="nameScope">The scope used by the UniqueNameProvider.</param>
        /// <returns></returns>
        public EventGenerator GetEventGenerator(FieldedDefinitionBase flowElement, string nameScope)
        {
            // find the name provider to use
            UniqueNameProvider nameProvider = GetNameProvider(nameScope);
            
            // and create the generator
            return new EventGenerator(flowElement, nameProvider, hardware);
        }

        /// <summary>
        /// Gets the name provider for the given scope.
        /// </summary>
        /// <param name="nameScope">The name scope.</param>
        /// <returns></returns>
        public UniqueNameProvider GetNameProvider(string nameScope)
        {
            if (!nameProviders.ContainsKey(nameScope))
                nameProviders.Add(nameScope, new UniqueNameProvider());
                
            return nameProviders[nameScope];
        }
    }
}
