using System.Collections.Generic;
using System.Linq;
using System.Text;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Code Generator for CodeBlocks
    /// </summary>
    public class CodeBlockGenerator : DataFlowBlockGenerationBase
    {
        private new CodeBlock flowElement;
        private UniqueNameProvider globalNameProiver;

        /// <summary>
        /// Initializes a new instance of the <see cref="CodeBlockGenerator"/> class.
        /// </summary>
        public CodeBlockGenerator(CodeBlock flowElement, UniqueNameProvider nameProvider, UniqueNameProvider globalNameProiver, HardwareDescriptionModel hardware)
            : base(flowElement, nameProvider, hardware)
        {
            this.flowElement = flowElement;
            this.globalNameProiver = globalNameProiver;
        }

        /// <summary>
        /// Generates the behavior.
        /// </summary>
        public override void generateBehavior()
        {
            CallArguments callArgs = new CallArguments();

            // collect the incomming and outgoing links
            int i = 1;
            foreach (CodeBlockInputField field in flowElement.Inputs)
            {
                // find incommong link for this field
                FlowElementReferencesTargetFlowElements link = IncomingLink(field.Guid);
                if (link == null)
                    continue;

                callArgs.Add(i++,nameProvider.GetSourceName(link),AddCallArgumentType.ByValue);
            }
            foreach (CodeBlockOutputField field in flowElement.Outputs)
            {
                // a new variable
                string fieldVariable = nameProvider.GetName(field);
                VariableDeclaration(field.Type, field.Store, fieldVariable);

                if ( field.Type.AsDslType(field.Store) != DslDataType.String )
                    callArgs.Add(i++, fieldVariable, AddCallArgumentType.ByReference);
                else
                    callArgs.Add(i++, fieldVariable, AddCallArgumentType.ByValue);
            }

            // call the CodeBlock function
            FunctionCall(globalNameProiver.GetName(flowElement), callArgs);
        }

        /// <summary>
        /// Generates the function declaration for this code block.
        /// 
        /// Can be used at the top of the code file or in a header file.
        /// </summary>
        /// <remarks>
        /// external void Name(type1 in1, type2 in2, type3* out1, type4* out2);
        /// </remarks>
        /// <param name="prefix">Name of the model this codeblock is placed in.</param>
        public string GenerateFunctionSignature(string prefix)
        {
            return GenerateFunctionSignature(prefix, flowElement, globalNameProiver);   
        }

        /// <summary>
        /// Generates the name of the function used for this codeblock.
        /// </summary>
        /// <param name="prefix">Name of the model this codeblock is placed in.</param>
        public string GenerateFunctionName(string prefix)
        {
            return GenerateFunctionName(prefix, flowElement, globalNameProiver);
        }
          
        public static string GenerateFunctionSignature(string prefix, CodeBlock flowElement, UniqueNameProvider globalNameProiver)
        {
            if (globalNameProiver == null)
                globalNameProiver = new UniqueNameProvider(); 
            
            StringBuilder sb = new StringBuilder();

            sb.Append("void ");
            sb.Append(GenerateFunctionName(prefix,flowElement,globalNameProiver));
            sb.Append("(");

            List<string> parameter = new List<string>();
            // inputs
            foreach (CodeBlockInputField f in flowElement.Inputs)
            {
                // strings must be called as a pointer
                string argumentType ="";
                DslDataType dslDataType = f.Type.AsDslType(f.Store);
                if ( dslDataType == DslDataType.String )
                    argumentType = "*";

                parameter.Add(string.Format("{0}{2} {1}", f.Type.AsCType(f.Store), f.Name, argumentType));
            }

            // outputs
            foreach (CodeBlockOutputField f in flowElement.Outputs)
                parameter.Add(string.Format("{0}* {1}", f.Type.AsCType(f.Store), f.Name));

            sb.Append(string.Join(", ", parameter.ToArray()));

            sb.Append(")");

            return sb.ToString();
        }

        public static string GenerateFunctionContent(CodeBlock flowElement, string comment)
        {
            StringBuilder sb = new StringBuilder();

            var incommingRecords = from l in FlowElementReferencesTargetFlowElements.GetLinksToSourceFlowElements(flowElement)
                                   where l.Type.AsDslType(l.Store) == DslDataType.Record
                                         && l.SourceFlowElement is Record
                                         && l.SourceField == null
                                   select new { Record = l.SourceFlowElement as Record,
                                                Parameter = (l.TargetField as CodeBlockFieldBase).Name};

            foreach (var recordLink in incommingRecords)
            {
                // check the record typ
                string inputVariableName = UniqueNameProvider.FormatCodeName(recordLink.Parameter);
                sb.AppendFormatLine("if( !IS_USER_STRUCT_TYPE({0}, {1}) ) return;", inputVariableName, recordLink.Record.RecordId );

                string recordVariable = UniqueNameProvider.FormatCodeName(recordLink.Parameter + "Expanded");
                // cast to the record struct 
                sb.AppendFormatLine("EXPAND_USER_STRUCT({0}, {1}, {2});", recordVariable, RecordGenerator.RecordType(recordLink.Record), inputVariableName);
            }

            sb.AppendLine();
            sb.AppendLine(comment);

            return sb.ToString();
        }
            
        public static string GenerateFunctionName(string prefix, CodeBlock flowElement, UniqueNameProvider globalNameProiver)
        {
            if (globalNameProiver == null)
                globalNameProiver = new UniqueNameProvider();

            return globalNameProiver.GetName(flowElement, prefix + "_");
        }
    }
}
