using System;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Variable=BenjaminSchroeter.Flow.DataFlow.Variable;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Generates code for a Variable defined in the Datastructure Model
    /// </summary>
    public class DatastructureVariableGenerator : DataFlowBlockGenerationBase
    {
        public const string DECLARATION_PREFIX = "var_";

        private new Variable flowElement;
        private DslDataType dslType;

        /// <summary>
        /// Initializes a new instance of the <see cref="DatastructureVariableGenerator"/> class.
        /// </summary>
        internal DatastructureVariableGenerator(Variable flowElement, UniqueNameProvider nameProvider, HardwareDescriptionModel hardware)
            : base(flowElement, nameProvider, hardware)
        {
            this.flowElement = flowElement;
        }

        /// <summary>
        /// Generates the behavior.
        /// </summary>
        public override void generateBehavior()
        {
            dslType = flowElement.Type.AsDslType(flowElement.Store);

            // each of this functions only generates something
            // if there is a incomming / outgoing link
            GenerateSet();
            GenerateGet();
        }

        /// <summary>
        /// Generates the set of a value to the variable if there is an incomming link.
        /// </summary>
        private void GenerateSet()
        {
            // find incommong link
            FlowElementReferencesTargetFlowElements link = IncomingLink(Guid.Empty);
            if (link == null)
                return;

            string variable = nameProvider.GetSourceName(link);

            if (dslType == DslDataType.Record)
                sb.AppendFormatLine("StoreRecord(&{0}, {1});", variable, GetVariableName(flowElement.Name));
            else
                Assignment(variable, GetVariableName(flowElement.Name), dslType);
        }

        /// <summary>
        /// Generates the get of a value to the variable if there is an outgoing link.
        /// </summary>
        private void GenerateGet()
        {
            // any outgoing links?
            if (flowElement.TargetFlowElements.Count == 0)
                return;

            string variable = nameProvider.GetName(flowElement);
            
            VariableDeclaration(flowElement.Type, flowElement.Store, variable);
            if (dslType == DslDataType.Record)
                sb.AppendFormatLine("ReadRecord({0}, &{1});", GetVariableName(flowElement.Name), variable);
            else
                Assignment(GetVariableName(flowElement.Name), variable, dslType);
        }

        #region declaration of the variable
        /// <summary>
        /// Creates the Declaration for the specified variable.
        /// </summary>
        public static string Declaration(BenjaminSchroeter.Flow.Datastructures.Variable v, HardwareDescriptionModel hardware, bool withDefaultValue)
        {
            if (v.Type == null)
                return "";

            DslDataType dslType = v.Type.AsDslType(v.Store);

            string cType = v.Type.AsCType(v.Store);
            if (dslType == DslDataType.Record)
                cType = "uint8_t";

            // default / initial value
            string value = "";
            if ( withDefaultValue )
            {
                value = GetValueLiteral(dslType, v.InitialValue);
                if (value != "")
                    value = " = " + value;
            }

            // appendix for strings (char arrays)
            string appendix = "";
            if (dslType == DslDataType.String)
                appendix = "[STRING_LENGTH]";
            else if (dslType == DslDataType.Record)
                appendix = "[RECORD_LENGTH]";
            
            return string.Format("{0} {1}{2}{3};", cType, GetVariableName(v.Name), appendix, value);
        }
        
        /// <summary>
        /// Gets the name of a Variable in the code.
        /// </summary>
        public static string GetVariableName(string name)
        {
            return DECLARATION_PREFIX + UniqueNameProvider.FormatCodeName(name);
        }
        #endregion

        public static string GetValueLiteral(DslDataType dslType, string valueString)
        {
            // default value
            string value = "";
            switch (dslType)
            {
                case DslDataType.Number:
                case DslDataType.Boolean:
                    if (valueString.Trim() != "")
                        value = valueString;
                    break;

                case DslDataType.String:
                    value = "\"" + valueString + "\"";
                    break;
            }
            return value;
        }
    }
}