using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using FieldBase=BenjaminSchroeter.Flow.HardwareDescription.FieldBase;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Code generator for Events
    /// </summary>
    public class EventGenerator : DataFlowBlockGenerationBase
    {
        private new FieldedDefinitionBase flowElement;
        private BenjaminSchroeter.Flow.HardwareDescription.Event hardwareDefinition;

        /// <summary>
        /// Initializes a new instance of the <see cref="EventGenerator"/> class.
        /// </summary>
        public EventGenerator(FieldedDefinitionBase flowElement, UniqueNameProvider nameProvider, HardwareDescriptionModel hardware) : base(flowElement, nameProvider, hardware)
        {
            this.flowElement = flowElement;

            if ( flowElement is InputEvent)
                this.hardwareDefinition = flowElement.Definition(hardware) as BenjaminSchroeter.Flow.HardwareDescription.Event;
        }

        /// <summary>
        /// Creates the event register, that is the call to the init function of this event
        /// </summary>
        /// <param name="prefix">The prefix.</param>
        /// <returns></returns>
        public string CreateEventRegister(string prefix)
        {
            InputEvent e = flowElement as InputEvent;
            Debug.Assert(e != null);

            // in the c-file for a dataflow the event register is only necessary 
            // if the Multiplicity > 1. Otherwise the call will be generated to
            // the main.c file.
            if (hardwareDefinition.Multiplicity > 1)
            {
                SingletonEventRegisterGenerator eventGen = new SingletonEventRegisterGenerator(hardware);
                return eventGen.CreateEventRegister(hardwareDefinition, e.Parameter, prefix + "_Callback" );
            }
            else // hardwareDefinition.Multiplicity == 1
            {
                return string.Format("// registering event '{0}' see main.c", this.flowElement.Name);
            }
        }

        /// <summary>
        /// Gets the name of the Callback method.
        /// </summary>
        /// <returns></returns>
        public string EventMethodName()
        {
            return "Callback_" + nameProvider.GetName(flowElement);
        }

        /// <summary>
        /// Creates the parameter list of the event callback function.
        /// </summary>
        /// <param name="withType">if set to <c>true</c> the parameter list will be created with the type declaration.</param>
        /// <returns></returns>
        public string CreateEventCallbackParameter(bool withType)
        {
            Dictionary<int, Field> mainCallParameter = new Dictionary<int, Field>();

            if (flowElement is InputEvent)
            {
                // for events the CallParameter data from hardwaredefintion is needed
                foreach (EventField field in hardwareDefinition.Fields)
                {
                    Field dataFlowField = flowElement.Fields.FirstOrDefault(f => f.Guid.Equals(field.Guid));
                    if (dataFlowField == null)
                        continue;

                    CallParameter fieldCall = field.CallParameter;
                    if (fieldCall.HasParameter)
                        mainCallParameter.Add(fieldCall.ParameterPos, dataFlowField);
                }
            }
            else if ( flowElement is FunctionCall)
            {
                int i = 1;
                foreach (Field field in flowElement.Fields)
                    mainCallParameter.Add(i++, field);
            }

            var parameter = from p in mainCallParameter 
                            orderby p.Key 
                            select (withType ? ParameterTypeString(p.Value) : "") + nameProvider.GetName(p.Value);
            return string.Join(", ", parameter.ToArray());
        }
        private string ParameterTypeString(Field field)
        {
            string s = field.Type.AsCType(field.Store);

            DslDataType type = field.Type.AsDslType(field.Store);
            if (type == DslDataType.String || type == DslDataType.Record)
                s += "*";
            
            s += " ";
            return s;
        }

        /// <summary>
        /// Generates the behavior.
        /// </summary>
        /// <remarks>An event has no behavior...</remarks>
        public override void generateBehavior()
        {
            Comment("no code nessesary: see function parameters");

            // use the function parameter for the field values
            foreach (Field field in this.flowElement.Fields)
            {
                // the function parameter were createt at first
                // so i can simply use the FormatCOdeName function
                string variableName = UniqueNameProvider.FormatCodeName(field.Name);
                // but for records we have to handle a pointer
                if (field.Type.AsDslType(field.Store) == DslDataType.Record)
                    variableName = string.Format("(*{0})", variableName);

                nameProvider.ExchangeName(field, variableName);
            }
        }
    }
}