using System.Linq;
using System.Text;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Generates the code for a formula used in the dataflow model.
    /// </summary>
    public class FormulaGenerator : DataFlowBlockGenerationBase
    {
        private new Formula flowElement;
        private UniqueNameProvider globalNameProiver;

        /// <summary>
        /// Initializes a new instance of the <see cref="FormulaGenerator"/> class.
        /// </summary>
        public FormulaGenerator(Formula flowElement, UniqueNameProvider nameProvider, UniqueNameProvider globalNameProiver, HardwareDescriptionModel hardware) 
            : base(flowElement, nameProvider, hardware)
        {
            this.flowElement = flowElement;
            this.globalNameProiver = globalNameProiver;
        }

        /// <summary>
        /// Generates the behavior: the call of the function.
        /// </summary>
        public override void generateBehavior()
        {
            // create a variable for the return value
            string variable = nameProvider.GetName(flowElement);
            VariableDeclaration(this.flowElement.GetTypeValue(), this.flowElement.Store, variable);

            if (Inline())
            {
                this.Assignment(this.flowElement.FormulaText, variable, this.flowElement.Type.AsDslType(this.flowElement.Store));
            }
            else
            {
                // find all parameter form incomming links
                var parameter =
                    from l in FlowElementReferencesTargetFlowElements.GetLinksToSourceFlowElements(flowElement)
                    orderby l.TargetAnnotation
                    select nameProvider.GetSourceName(l);

                // call the function
                CallArguments args = new CallArguments();
                foreach (string s in parameter)
                    args.Add(s, AddCallArgumentType.ByValue);
                this.FunctionCall(globalNameProiver.GetName(flowElement), args, variable);
            }
        }

        /// <summary>
        /// Generates the function used for the formula.
        /// </summary>
        public string GenerateFunction(string prefix)
        {
            if (Inline())
                return ""; // do not generate a function

            StringBuilder sb = new StringBuilder();

            sb.Append( flowElement.GetTypeValue().AsCType(flowElement.Store)  );
            sb.Append(" ");
            sb.Append(globalNameProiver.GetName(flowElement, prefix));
            sb.Append("(");

            var parameter = from l in FlowElementReferencesTargetFlowElements.GetLinksToSourceFlowElements(flowElement)
                            orderby l.TargetAnnotation
                            select l.Type.AsCType(l.Store) + " " + l.TargetAnnotation;
            sb.Append(string.Join(", ", parameter.ToArray()));
            
            sb.Append(")");
            sb.AppendLine();
            sb.AppendLine("{");
            sb.AppendFormatLine("\treturn {0};", flowElement.FormulaText);
            sb.AppendLine("}");

            return sb.ToString();            
        }

        /// <summary>
        /// Determins if this formula can be inlined. 
        /// 
        /// If a fromula has no incomming connectors it can only contain
        /// a constant expression. For such an expression no method is needed.
        /// </summary>
        /// <returns></returns>
        private bool Inline()
        {
            return this.flowElement.SourceFlowElements.Count == 0;
        }
    }
}
