using System;
using System.Diagnostics;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Generates the code for records
    /// </summary>
    class OutputEventGenerator : DataFlowBlockGenerationBase
    {
        private new FieldedDefinitionBase flowElement;
        private int recordId;

        /// <summary>
        /// Initializes a new instance of the <see cref="OutputEventGenerator"/> class.
        /// </summary>
        public OutputEventGenerator(FieldedDefinitionBase flowElement, UniqueNameProvider nameProvider, HardwareDescriptionModel hardware)
            : base(flowElement, nameProvider, hardware)
        {
            Debug.Assert(flowElement is OutputEvent || flowElement is FunctionReturn);

            this.flowElement = flowElement;

            if (flowElement is OutputEvent)
                recordId = (flowElement as OutputEvent).InternalRecordId;
            else if (flowElement is FunctionReturn)
                recordId = (flowElement as FunctionReturn).InternalRecordId;
        }

        /// <summary>
        /// Generates the behavior.
        /// </summary>
        public override void generateBehavior()
        {
            string variable = nameProvider.GetName(flowElement);

            // create a new record struct
            sb.AppendFormatLine("CREATE_USER_STRUCT_INSTANCE({0}, {1});", RecordType(), variable);

            // set inputs (or default values)
            foreach (OptionalbleFieldWithDefault field in flowElement.Fields)
            {
                // name (a->b) of the current field
                string assignmentLeft = nameProvider.GetName(field, variable + "->");

                FlowElementReferencesTargetFlowElements link = this.IncomingLink(field.Guid);
                DslDataType dslType = field.Type.AsDslType(field.Store);
                if (link == null)
                {
                    // use default value for this field
                    string value = DatastructureVariableGenerator.GetValueLiteral(dslType, field.DefaultValue);

                    if ( value != "" )
                        this.Assignment(value, assignmentLeft, dslType, field.StringLength);
                }
                else
                {
                    // set the value of the field
                    this.Assignment(nameProvider.GetSourceName(link), assignmentLeft, dslType, field.StringLength);
                }
            }

            // get the internal record id of the output event
            

            // if there is an output from the head of this record shape
            //     i have to encpasulate the record now
            string encRecordName = "encapsulated_" + variable;
            sb.AppendFormatLine("ENCAPSULATE_USER_STRUCT({0}, {1}, {2});", encRecordName, recordId , variable);

            sb.AppendFormatLine("SubmitFlowRecordResponse(&{0});", encRecordName);

        }

        private string RecordType()
        {
            return RecordType(flowElement);
        }

        public static string RecordType(FieldedDefinitionBase flowElement)
        {
            string prefix = "";
            if (flowElement is OutputEvent)
                prefix = "oe_";
            else if (flowElement is FunctionReturn)
                prefix = "f_ret_";

            return prefix + UniqueNameProvider.FormatCodeName(flowElement.Name) + "_t";            
        }
    }
}
