using System;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Generates the code for records
    /// </summary>
    class RecordGenerator : DataFlowBlockGenerationBase
    {
        private new Record flowElement;

        /// <summary>
        /// Initializes a new instance of the <see cref="RecordGenerator"/> class.
        /// </summary>
        public RecordGenerator(Record flowElement, UniqueNameProvider nameProvider, HardwareDescriptionModel hardware)
            : base(flowElement, nameProvider, hardware)
        {
            this.flowElement = flowElement;
        }

        /// <summary>
        /// Generates the behavior.
        /// </summary>
        public override void generateBehavior()
        {
            string recordVariable = nameProvider.GetName(flowElement);

            FlowElementReferencesTargetFlowElements inputToHead = this.IncomingLink(Guid.Empty);

            // writing to the Record
            if ( inputToHead == null )
            {
                // create a new record struct
                sb.AppendFormatLine("CREATE_USER_STRUCT_INSTANCE({0}, {1});", RecordType(), recordVariable);

                // set inputs (or default values)
                foreach (OptionalbleFieldWithDefault field in flowElement.Fields)
                {
                    // name (a->b) of the current field
                    string assignmentLeft = nameProvider.GetName(field, recordVariable + "->");

                    FlowElementReferencesTargetFlowElements link = this.IncomingLink(field.Guid);
                    DslDataType dslType = field.Type.AsDslType(field.Store);
                    if (link == null)
                    {
                        // use default value for this field
                        string value = DatastructureVariableGenerator.GetValueLiteral(dslType, field.DefaultValue);

                        if ( value != "" )
                            this.Assignment(value, assignmentLeft, dslType, field.StringLength);
                    }
                    else
                    {
                        // set the value of the field
                        this.Assignment(nameProvider.GetSourceName(link), assignmentLeft, dslType, field.StringLength);
                    }
                }
            }
            else
            {
                // check the record typ
                string inputVariableName = nameProvider.GetSourceName(inputToHead);
                sb.AppendFormatLine("if( !IS_USER_STRUCT_TYPE({0}, {1}) ) return;",inputVariableName, flowElement.RecordId);

                // cast to the record struct 
                sb.AppendFormatLine("EXPAND_USER_STRUCT({0}, {1}, {2});", recordVariable, RecordType(), inputVariableName);

                // register all fields at the nameprovider
                //  so other elements can query it for the name
                foreach (OptionalbleFieldWithDefault field in flowElement.Fields)
                {
                    // name (a->b) of the current field
                    nameProvider.GetName(field, recordVariable + "->");
                }
            }

            // if there is an output from the head of this record shape
            //     i have to encpasulate the record now
            FlowElementReferencesTargetFlowElements outptuFromHead = this.OutgoingLink(Guid.Empty);
            if ( outptuFromHead != null )
            {
                string encRecordName = "encapsulated_" + recordVariable;
                sb.AppendFormatLine("ENCAPSULATE_USER_STRUCT({0}, {1}, {2});", encRecordName, flowElement.RecordId, recordVariable);
                nameProvider.ExchangeName(flowElement, encRecordName);
            }

        }

        private string RecordType()
        {
            return RecordType(flowElement);
        }

        public static string RecordType(Record record)
        {
            return UniqueNameProvider.FormatCodeName(record.Name) + "_t";            
        }
    }
}
