using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Dsl.DynamicProperties;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// Creates the registration methods for Events with Multiplicity == 1.
    /// </summary>
    public class SingletonEventRegisterGenerator : BlockGeneratorBase
    {
        private struct CallHint
        {
            public string Name;
            public int Priority;
        }

        private Dictionary<Guid, List<CallHint>> methodCalls;
        private UniqueNameProvider provider;
            
        /// <summary>
        /// Initializes a new instance of the <see cref="SingletonEventRegisterGenerator"/> class.
        /// </summary>
        public SingletonEventRegisterGenerator(HardwareDescriptionModel hardware) : base(hardware)
        {
            methodCalls = new Dictionary<Guid, List<CallHint>>();
            provider = new UniqueNameProvider();
        }

        /// <summary>
        /// Opens the given DataFlowModel and search for events that must be registerd
        /// </summary>
        /// <param name="filename">The filename of a DataFlowModel.</param>
        public void ReadEventsFromDataFlow(string filename)
        {
            using (Store store = new Store())
            {
                // load the model
                DataFlowModel model;
                store.LoadDomainModels(new[] {typeof (CoreDesignSurfaceDomainModel), typeof (DataFlowDomainModel)});
                using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                {
                    model = DataFlowSerializationHelper.Instance.LoadModel(store, filename, null, null);
                    t.Commit();
                }

                // find all events
                var events = model.TriggerFlow.FlowElements.OfType<InputEvent>();

                // remember the events and the init calls
                foreach (InputEvent e in events)
                {
                    // events with multiplicity > 1 are handled in the flows
                    Event hardwareEvent = FindEvent(e.Guid);
                    if ( hardwareEvent.Multiplicity > 1 )
                        continue;

                    if (!methodCalls.ContainsKey(e.Guid))
                        methodCalls.Add(e.Guid, new List<CallHint>());

                    string method = string.Format("{0}_Callback_{1}", Path.GetFileNameWithoutExtension(filename),e.Name);
                    method = UniqueNameProvider.FormatCodeName(method);
                    methodCalls[e.Guid].Add(new CallHint { Name = method, Priority = model.Priority }); 
                }
            }
        }

        /// <summary>
        /// Generates the callback functions for all events and call the functions of the Dataflow
        /// </summary>
        /// <returns></returns>
        public string GenerateCallbackFunctions()
        {
            StringBuilder sb = new StringBuilder();

            foreach (KeyValuePair<Guid, List<CallHint>> pair in methodCalls)
            {
                Event e = FindEvent(pair.Key);
                if ( e == null )
                {
                    Debug.Fail("Event not found in hardware description.");
                    continue;
                }
                
                UniqueNameProvider nameProviderForMethod = new UniqueNameProvider();
                string parameterDef = CreateEventCallbackParameter(e, true, nameProviderForMethod);
                string parameterList = CreateEventCallbackParameter(e, false, nameProviderForMethod);

                sb.AppendFormatLine("void {0}({1})", provider.GetName(e.Guid, "Callback_" + e.Name), parameterDef);
                sb.AppendLine("{");
                foreach (CallHint hint in pair.Value.OrderBy( x=> x.Priority))
                    sb.AppendFormatLine("    {0}({1});", hint.Name, parameterList);
                sb.AppendLine("}");
                sb.AppendLine();
            }

            return sb.ToString();
        }

        /// <summary>
        /// Creates the event register calls.
        /// </summary>
        /// <returns></returns>
        public string CreateEventRegister()
        {
            StringBuilder sb = new StringBuilder();

            foreach (KeyValuePair<Guid, List<CallHint>> call in methodCalls)
            {
                Event e = FindEvent(call.Key);
                if (e != null)
                    sb.Append(CreateEventRegister(e, null, "Callback"));
            }

            return sb.ToString();            
        }

        public string CreateEventRegister(Event eventDef, DynamicPropertyCollection userParameter, string callbackMethodPrefix)
        {
            sb = new StringBuilder();

            // collect the parameter of the shape
            Dictionary<int, string> callParameter = new Dictionary<int, string>();
            foreach (Parameter p in eventDef.Parameters)
            {
                CallParameter parameterCall = p.CallParameter;
                if (parameterCall.HasParameter)
                {
                    string value = "";
                    if (userParameter == null)
                    {
                        // use the parameter form the Event defintions
                        value = p.Value;
                    }
                    else
                    {
                        // use the giben user Parameter
                        if (userParameter.ExistsProperty(p.Name))
                        {
                            object o = userParameter.GetPropertyValue(p.Name);
                            if (o != null)
                                value = o.ToString();
                        }                    
                    }

                    callParameter.Add(parameterCall.ParameterPos, ParameterValue(value, p.Type.AsDslType(p.Store)));
                }
            }
            // the enabled property of this shape
            if (eventDef.CallParameterInit.HasEnabledParameter)
                callParameter.Add(  eventDef.CallParameterInit.EnabledParameterPos, ParameterValue(eventDef.Enabled) );

            var sortedParameter = (from p in callParameter orderby p.Key select p.Value);

            // create the list of function parameter
            List<string> parameter = new List<string>();
            foreach (string p in sortedParameter)
                parameter.Add(p);

            // add the last parameter
            parameter.Add(string.Format("(void*) {0}", provider.GetName(eventDef.Guid, callbackMethodPrefix + "_" + eventDef.Name))); // last parameter: function pointer

            // call the function
            CallArguments args = new CallArguments();
            foreach (string s in parameter)
                args.Add(s, AddCallArgumentType.ByValue);
            
            this.FunctionCall(eventDef.CallParameterInit.Function, args);

            return sb.ToString();            
        }

        private Event FindEvent(Guid guid)
        {
            foreach (Platform platform in hardware.Platforms)
            {
                Event e = platform.Events.FirstOrDefault(x => x.Guid == guid);
                if ( e != null )
                    return e;
            }

            return null;
        }

        private string CreateEventCallbackParameter(Event eventDef, bool withType, UniqueNameProvider nameProvider)
        {
            Dictionary<int, EventField> mainCallParameter = new Dictionary<int, EventField>();

            foreach (EventField field in eventDef.Fields)
            {
                CallParameter fieldCall = field.CallParameter;
                if (fieldCall.HasParameter)
                    mainCallParameter.Add(fieldCall.ParameterPos, field);
            }

            var parameter = from p in mainCallParameter 
                            orderby p.Key
                            select (withType ? ParameterTypeString(p.Value) : "") 
                                    + nameProvider.GetName(p.Value.Guid, p.Value.Name);
            return string.Join(", ", parameter.ToArray());
        }
        private string ParameterTypeString(EventField field)
        {
            string s = field.Type.AsCType(field.Store);

            DslDataType type = field.Type.AsDslType(field.Store);
            if (type == DslDataType.String || type == DslDataType.Record)
                s += "*";

            s += " ";
            return s;
        }

    }
}
