using System;
using System.Collections.Generic;
using System.Text;
using BenjaminSchroeter.Flow.DataFlow;

namespace BenjaminSchroeter.Flow.CodeGen
{
    /// <summary>
    /// This class can provide unique names for elemets used in the 
    /// generated source code. If a wish-name is already in use
    /// a number will be appended (e.g. name_1).
    /// 
    /// It is also possible to receive the name for a already
    /// created element.
    /// </summary>
    public class UniqueNameProvider
    {
        /// <summary>
        /// This dictionary stores for every object the given name.
        /// </summary>
        private Dictionary<object, string> names;
        /// <summary>
        /// This dictionary stores for every name prefix the last give number.
        /// </summary>
        private Dictionary<string, int> index;

        /// <summary>
        /// Initializes a new instance of the <see cref="UniqueNameProvider"/> class.
        /// </summary>
        public UniqueNameProvider()
        {
            Clear();
        }

        /// <summary>
        /// Clears this instance.
        /// </summary>
        private void Clear()
        {
            names = new Dictionary<object, string>();
            index = new Dictionary<string, int>();
        }

        /// <summary>
        /// Gets the name for the given FlowElement.
        /// 
        /// The same FlowElement receives always the same name.
        /// </summary>
        public string GetName(FlowElement key)
        {
            return this.GetName(key, "");
        }

        /// <summary>
        /// Gets the name the given guid.
        /// </summary>
        public string GetName(Guid key, string prefix)
        {
            if (names.ContainsKey(key))
                return names[key];

            string n = prefix;
            n = FormatCodeName(n);
            n = addIndex(n);

            names.Add(key, n);
            return n;
        }

        /// <summary>
        /// Gets the name for the given FlowElement and uses the prefix 
        /// as part of the name.
        /// </summary>
        public string GetName(FlowElement key, string prefix)
        {
            if (names.ContainsKey(key))
                return names[key];

            string n = "unknown";
            if (key is MasterTrigger)
                n = "MasterTrigger";
            else if (key is DefinitionBase)
                n = (key as DefinitionBase).Name;
            else if (key is Operation)
                n = "op";
            else if (key is Formula)
                n = "Formula";
            else if (key is CodeBlock)
                n = (key as CodeBlock).Name;

            n = prefix + n;
            n = FormatCodeName(n);
            n = addIndex(n);

            names.Add(key, n);
            return n;
        }

        /// <summary>
        /// Gets the name for a FieldBase element.
        /// </summary>
        public string GetName(FieldBase key)
        {
            return GetName(key, "");    
        }

        /// <summary>
        /// Gets the name for a FieldBase element.
        /// </summary>
        public string GetName(FieldBase key, string prefix)
        {
            if (names.ContainsKey(key))
                return names[key];

            string n = "unknown";
            if (key is Field)
                n = (key as Field).Name;
            else if (key is CodeBlockFieldBase)
                n = (key as CodeBlockFieldBase).Name;

            n = FormatCodeName(n);
            n = prefix + n;
            n = addIndex(n);

            names.Add(key, n);
            return n;
        }

        /// <summary>
        /// Exchanges the name of an already registerd FlowElement.
        /// Every request for this element will return the new name from now
        /// </summary>
        public void ExchangeName(FlowElement key, string newName)
        {
            names[key] = newName;
        }
        /// <summary>
        /// Exchanges the name of an already registerd FieldBase.
        /// Every request for this element will return the new name from now
        /// </summary>
        public void ExchangeName(FieldBase key, string newName)
        {
            names[key] = newName;
        }

        /// <summary>
        /// Gets the name of the source of the given link.
        /// </summary>
        public string GetSourceName(FlowElementReferencesTargetFlowElements link)
        {
            if (link.SourceEntry.Equals(new Guid()))
                return GetName(link.SourceFlowElement);
            else
                return GetName(link.SourceField);
        }

        /// <summary>
        /// Adds a index to the given string if the 
        /// string is already in use as a name.
        /// </summary>
        private string addIndex(string n)
        {
            if (index.ContainsKey(n))
            {
                int i = index[n];
                index[n] = i + 1;
                return n + "_" + i;
            }
            else
            {
                index.Add(n, 1);
                return n;
            }
        }

        /// <summary>
        /// Formats a string in a way to use it as a name in sourcecode.
        /// </summary>
        /// <remarks>Only Letters, Numbers and the Underscore are allowed.</remarks>
        public static string FormatCodeName(string s)
        {
            bool first = true;

            StringBuilder sb = new StringBuilder(s.Length);
            foreach (char c in s)
                if (   (char.IsLetter(c) && ((byte)c) < 127) // no umlaute ;)
                    || c == '_'
                    || (char.IsNumber(c) && !first))
                {
                    sb.Append(c);
                    first = false;
                }

            if (sb.Length == 0)
                return "NoName";

            return sb.ToString();
        }
    }
}
