using System;
using System.Collections.Generic;
using BenjaminSchroeter.Dsl.CompartmentMapping;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.DataFlow
{
    /// <summary>
    /// Partial class with a few methods used by the classes created by the
    /// CompartmentMapping. See Mapping.xml / Mapping_xGen.cs in this folder.
    /// </summary>
    partial class FlowElementReferencesTargetFlowElementsBuilderInstance
    {
        /// <summary>
        /// Make a connection between the given pair of source and target elements
        /// </summary>
        /// <param name="source">The model element to use as the source of the connection</param>
        /// <param name="sourcePartType">The type of the source. (head or entry)</param>
        /// <param name="sourceEntry">The model element to use as the source entry of the conneciton</param>
        /// <param name="target">The model element to use as the target of the connection</param>
        /// <param name="targetPartType">The type of the target. (head or entry)</param>
        /// <param name="targetEntry">The model elment to use as the target entry of the connection</param>
        /// <returns>
        /// A link representing the created connection
        /// </returns>
        /// <remarks>This method is only called if both, source and targent, are compartment shapes.</remarks>
        protected override ElementLink CreateElementLink(FlowElement source, 
                                                            SelectedCompartmentPartType sourcePartType,
                                                            FieldBase sourceEntry, 
                                                            FlowElement target,
                                                            SelectedCompartmentPartType targetPartType,
                                                            FieldBase targetEntry)
        {
            // deletes other incomming links, since only one input link is allowed (for exceptions see the method itself)
            RemoveOtherIncommingLinks(target, targetPartType, targetEntry);

            FlowElementReferencesTargetFlowElements link = new FlowElementReferencesTargetFlowElements(source, target);

            if (sourcePartType == SelectedCompartmentPartType.None || sourcePartType == SelectedCompartmentPartType.Head)
                link.SourceEntry = Guid.Empty;
            else if (sourcePartType == SelectedCompartmentPartType.Element)
                link.SourceEntry = sourceEntry.Guid;

            if (targetPartType == SelectedCompartmentPartType.None || targetPartType == SelectedCompartmentPartType.Head)
                link.TargetEntry = Guid.Empty;
            else if (targetPartType == SelectedCompartmentPartType.Element)
                link.TargetEntry = targetEntry.Guid;

            // some special case: formula (see method itself)
            if (target is Formula)
                CreateTargetAnnotation((Formula) target, link);

            return link;
        }

        /// <summary>
        /// Checks if the given connection has the entry as source
        /// </summary>
        /// <param name="entry">the entry if a compartment shape</param>
        /// <param name="connection">the connection</param>
        /// <returns>
        /// true, if the connection has the entry as source
        /// </returns>
        /// <remarks>You have to implement this method, because only you can know how to store this information in your connection class.</remarks>
        public override bool IsEntryConnectionSource(FieldBase entry, FlowElementReferencesTargetFlowElements connection)
        {
            if (entry == null)
                return connection.SourceEntry == Guid.Empty;

            return connection.SourceEntry == entry.Guid;
        }

        /// <summary>
        /// Checks if the given connection has the entry as target
        /// </summary>
        /// <param name="entry">the entry if a compartment shape</param>
        /// <param name="connection">the connection</param>
        /// <returns>
        /// true, if the connection has the entry as target
        /// </returns>
        /// <remarks>You have to implement this method, because only you can know how to store this information in your connection class.</remarks>
        public override bool IsEntryConnectionTarget(FieldBase entry, FlowElementReferencesTargetFlowElements connection)
        {
            if (entry == null)
                return connection.TargetEntry == Guid.Empty;

            return connection.TargetEntry == entry.Guid;
        }

        #region CanAccept...
        /// <summary>
        /// Test whether a given model element is acceptable to this ConnectionBuilder as the source of a connection.
        /// </summary>
        /// <remarks>
        /// A main part of the logic of allowed connections is implemented in this method.
        /// </remarks>
        /// <param name="candidate">The model element to test.</param>
        /// <param name="partType">The type of the source. (head or entry)</param>
        /// <param name="candidateEntry">The model element to use as the source entry of the conneciton</param>
        /// <returns>
        /// Whether the element can be used as the source of a connection.
        /// </returns>
        protected override bool CanAcceptAsCompartmentSource(FlowElement candidate, SelectedCompartmentPartType partType, FieldBase candidateEntry)
        {
            bool isTriggerFlow = candidate.FlowBase is TriggerFlow;

            return candidate is FunctionCall && partType == SelectedCompartmentPartType.Element
                   || candidate is FunctionCall && partType == SelectedCompartmentPartType.Head && isTriggerFlow
                   || candidate is Record && partType == SelectedCompartmentPartType.Element
                   || candidate is Record && partType == SelectedCompartmentPartType.Head
                   || candidate is Input && partType == SelectedCompartmentPartType.Element
                   || candidate is InputEvent && partType == SelectedCompartmentPartType.Element
                   || candidate is InputEvent && partType == SelectedCompartmentPartType.Head && isTriggerFlow
                   || candidate is CodeBlock && partType == SelectedCompartmentPartType.Element && candidateEntry is CodeBlockOutputField
                   || candidate is Variable && (candidate as Variable).Readable
                   || candidate is Operation
                   || candidate is Formula;
        }

        /// <summary>
        /// Test whether a given model element is acceptable to this ConnectionBuilder as the target of a connection
        /// </summary>
        /// <remarks>
        /// A main part of the logic of allowed connections is implemented in this method.
        /// </remarks>
        /// <param name="candidate"></param>
        /// <param name="partType"></param>
        /// <param name="candidateEntry"></param>
        /// <returns>
        /// Whether the elements can be used as the source and target of a connection
        /// </returns>
        protected override bool CanAcceptAsCompartmentTarget(FlowElement candidate, SelectedCompartmentPartType partType, FieldBase candidateEntry)
        {
            return candidate is MasterTrigger
                   || candidate is FunctionReturn && partType == SelectedCompartmentPartType.Element
                   || candidate is Record && partType == SelectedCompartmentPartType.Element
                   || candidate is Record && partType == SelectedCompartmentPartType.Head
                   || candidate is Output && partType == SelectedCompartmentPartType.Element
                   || candidate is OutputEvent && partType == SelectedCompartmentPartType.Element
                   || candidate is CodeBlock && partType == SelectedCompartmentPartType.Element && candidateEntry is CodeBlockInputField
                   || candidate is Variable && (candidate as Variable).Writable
                   || candidate is Operation && (candidate as Operation).AcceptIncomingConnection()
                   || candidate is Formula;
                   
        }

        /// <summary>
        /// Test whether a given pair of model elements are acceptable to this ConnectionBuilder as the source and target of a connection
        /// </summary>
        /// <param name="sourceElement">The model element to use as the source of the connection</param>
        /// <param name="sourcePartType">The type of the source. (head or entry)</param>
        /// <param name="sourceEntry">The model element to use as the source entry of the conneciton</param>
        /// <param name="targetElement">The model element to use as the target of the connection</param>
        /// <param name="targetPartType">The type of the target. (head or entry)</param>
        /// <param name="targetEntry">The model elment to use as the target entry of the connection</param>
        /// <returns>
        /// Whether the elements can be used as the source and target of a connection.
        /// </returns>
        protected override bool CanAcceptAsCompartmentSourceAndTarget(FlowElement sourceElement, SelectedCompartmentPartType sourcePartType, FieldBase sourceEntry, FlowElement targetElement, SelectedCompartmentPartType targetPartType, FieldBase targetEntry)
        {
            bool isTriggerFlow = sourceElement.FlowBase is TriggerFlow;

            // the head of a record can go directly to the master trigger
            if (sourceElement is Record && sourcePartType == SelectedCompartmentPartType.Head
                && targetElement is MasterTrigger)
                return true;

            // the fields of a record and variables must not be set in the trigger flow
            if (isTriggerFlow && targetElement is Record && targetPartType == SelectedCompartmentPartType.Element)
                return false;
            if (isTriggerFlow && targetElement is Variable )
                return false;


            // check datatypes (Link is only allowed if the datatype is the same)
            if (!(targetElement is Formula))  // but formulas accept every type as an input 
            {
                string sourceType = DataTypeHelper.GetDataType(sourceElement, sourceEntry);
                string targetType = DataTypeHelper.GetDataType(targetElement, targetEntry);
                if (sourceType != targetType)
                    return false;
            }

            // the head of an event can only target the master trigger or operations
            if (sourceElement is InputEvent && sourcePartType == SelectedCompartmentPartType.Head
                && !(targetElement is MasterTrigger || targetElement is Operation))
                return false;
                
            // allow only connection within the same flow
            // (e.g. no connections from TriggerFlow to ProgramFlow allowed)
            return sourceElement.FlowBase == targetElement.FlowBase;
        }
        #endregion

        /// <summary>
        /// Creates the target annotation for the connection.
        /// These annotations will be used for connections to forumal shapes.
        /// A unique name for every new incomming connection will be found.
        /// </summary>
        /// <param name="formula">The formula.</param>
        /// <param name="newlink">The newlink.</param>
        private void CreateTargetAnnotation(Formula formula, FlowElementReferencesTargetFlowElements newlink)
        {
            // a list of annotations already in use for the given formula
            HashSet<string> usedNames = new HashSet<string>();

            foreach (FlowElementReferencesTargetFlowElements link in FlowElementReferencesTargetFlowElements.GetLinksToSourceFlowElements(formula))
                usedNames.Add(link.TargetAnnotation);

            // find a single char (starting at 'a')
            string newName = "";
            for(byte i = 97 ; i<=122;i++)
            {
                newName = ((char) i).ToString();
                if (!usedNames.Contains(newName))
                    break;
            }

            // setup the annotation
            newlink.TargetAnnotation = newName;
        }

        /// <summary>
        /// Removes other incomming links when creating a new one.
        /// Each input can only have one incomming connection.
        /// </summary>
        /// <remarks>
        /// Exceptions of this rule are: MasterTrigger, Formula and Operation
        /// </remarks>
        /// <param name="target">The target.</param>
        /// <param name="type">The type.</param>
        /// <param name="entry">The entry.</param>
        private void RemoveOtherIncommingLinks(FlowElement target, SelectedCompartmentPartType type, FieldBase entry)
        {
            // these elements are allowed to have multiple input links
            if ( target is MasterTrigger 
                || target is Formula
                || target is Operation)
                return;
            
            // find other incomming links the shape
            List<FlowElementReferencesTargetFlowElements> todelete = new List<FlowElementReferencesTargetFlowElements>();
            foreach (FlowElementReferencesTargetFlowElements link in FlowElementReferencesTargetFlowElements.GetLinksToSourceFlowElements(target))
            {
                // check whether this link is relay the link of oure filed
                if (type == SelectedCompartmentPartType.None || type == SelectedCompartmentPartType.Head)
                {
                    if (link.TargetEntry.Equals(Guid.Empty))
                        todelete.Add(link);
                }
                else if (type == SelectedCompartmentPartType.Element)
                {
                    if (link.TargetEntry.Equals(entry.Guid))
                        todelete.Add(link);
                }
            }

            // delete the links
            foreach (FlowElementReferencesTargetFlowElements link in todelete)
                link.Delete();
        }
    }
}
