using System;
using System.Diagnostics;
using BenjaminSchroeter.Flow.DataTypes;
using BenjaminSchroeter.Flow.HardwareDescription.Data;

namespace BenjaminSchroeter.Flow.DataFlow
{
    /// <summary>
    /// Some Helper methods for the Datatypes of model elements
    /// </summary>
    public static class DataTypeHelper
    {
        /// <summary>
        /// Gets the DataType of a given FlowElement with consideration the selected entry.
        /// </summary>
        /// <returns>The name of the DataType.</returns>
        public static string GetDataType(FlowElement element, FieldBase entry)
        {
            // Head of compartment shape or a regular shape            
            if (entry == null)
            {
                Variable v = element as Variable;
                if (v != null)
                    return v.Type.Value;

                // operation are always boolean
                Operation o = element as Operation;
                if (o != null)
                    if (o.Kind == OperationKind.LogicalAnd
                        || o.Kind == OperationKind.LogicalNot
                        || o.Kind == OperationKind.LogicalOr
                        || o.Kind == OperationKind.LogicalXor
                        || o.Kind == OperationKind.True
                        || o.Kind == OperationKind.False)
                        return GetDefaultDataType(DslDataType.Boolean, element);

                // for events / trigger flows
                if (element is FunctionCall
                    || element is InputEvent)
                    return GetDefaultDataType(DslDataType.Boolean, element);

                if (element is Formula)
                    return (element as Formula).Type.Value;          

                if (element is Record)
                    return GetDefaultDataType(DslDataType.Record, element);

                if (element is MasterTrigger)
                    return GetDefaultDataType(DslDataType.Boolean, element);
            }
            else
            {
                if (entry is Field)
                    return (entry as Field).Type.Value;

                if (entry is CodeBlockFieldBase)
                    return (entry as CodeBlockFieldBase).Type.Value;
            }

            Debug.Fail("Cannot get datatype for connection.");
            return "UNKNOWN";
        }

        /// <summary>
        /// Gets the default DataType for a given DslDataType
        /// </summary>
        /// <param name="element">Used to access the DataTypes of the DataFlowModel.</param>
        public static string GetDefaultDataType(DslDataType dslType, FlowElement element)
        {
            return DataTypeConverter.GetDefaultDataType(element.Store, dslType);
        }

        /// <summary>
        /// Converts a DslDataType to a .net Type
        /// </summary>
        /// <remarks>e.g. used for formula evaluation.</remarks>
        /// <returns></returns>
        public static Type DslType2DotNet(DslDataType dslDataType)
        {
            switch (dslDataType)
            {
                case DslDataType.Boolean:
                    return typeof (bool);
                case DslDataType.Number:
                    return typeof (int);
                case DslDataType.String:
                    return typeof (string);
                case DslDataType.Record:
                    return typeof (object);
                case DslDataType.Other:
                    return typeof (object);
            }

            return typeof(void);
        }

        /// <summary>
        /// Gets the DataType name for a given .net type
        /// </summary>
        /// <param name="type">The type.</param>
        /// <param name="element">The element.</param>
        /// <returns></returns>
        public static string GetTypeName(Type type, FlowElement element)
        {
            DslDataType t;
            if (type == typeof(bool))
                t = DslDataType.Boolean;
            else if (type == typeof(int))
                t = DslDataType.Number;
            else if (type == typeof(string))
                t = DslDataType.String;
            else if (type == typeof(object))
                t = DslDataType.Record;
            else 
                t = DslDataType.Other;

            return DataTypeConverter.GetDefaultDataType(element.Store, t);
        }
    }
}
