using System.Collections.Generic;
using BenjaminSchroeter.Flow.Datastructures.Data;

namespace BenjaminSchroeter.Flow.DataFlow.Update
{
    /// <summary>
    /// Integrates changes in the DatastructureModelData into a DataFlow Model
    /// </summary>
    public class DatastructuresUpdater : UpdaterBase
    {
        public void Update(DataFlowModel dataflow, DatastructureModelData datastructures)
        {
            UpdateFlowBase(dataflow.TriggerFlow, datastructures);
            UpdateFlowBase(dataflow.ProgramFlow, datastructures);
        }
        /// <summary>
        /// Looks for differences between the Datastructures and the elements of the given FlowBase
        /// </summary>
        private void UpdateFlowBase(FlowBase flow, DatastructureModelData datastructure)
        {
            // a list with elements to delete later on 
            // (deleting on the fly is not allowed inside the foreach loop)
            List<FlowElement> toDelete = new List<FlowElement>();

            // check ever element
            foreach (FlowElement flowElement in flow.FlowElements)
            {
                Variable variable = flowElement as Variable;
                if (variable != null && variable.FromHardwareDescription == false)
                    if (!UpdateVariable(datastructure.FindVariable(variable.Guid), variable))
                        toDelete.Add(flowElement);

                Record record = flowElement as Record;
                if (record != null)
                    if (!UpdateRecord(datastructure.FindRecord(record.Guid), record))
                        toDelete.Add(flowElement);

                OutputEvent outputEvent = flowElement as OutputEvent;
                if (outputEvent != null)
                    if (!UpdateOutputEvent(datastructure.FindOutputEvent(outputEvent.Guid), outputEvent))
                        toDelete.Add(flowElement);

                FunctionCall functionCall = flowElement as FunctionCall;
                if (functionCall != null)
                    if (!UpdateFunctionCall(datastructure.FindFunction(functionCall.Guid), functionCall))
                        toDelete.Add(flowElement);

                FunctionReturn functionReturn = flowElement as FunctionReturn;
                if (functionReturn != null)
                    if (!UpdateFunctionReturn(datastructure.FindFunction(functionReturn.Guid), functionReturn))
                        toDelete.Add(flowElement);
            }

            // now, do the deletion
            foreach (FlowElement element in toDelete)
                element.Delete();
        }

        /// <summary>
        /// Handels changes of Name and Description as well as the situation where the base element was deleted from the datastructure definitions
        /// </summary>
        /// <returns>false, if the elment was deleted from the datastructure definitions</returns>
        private bool UpdateDefinitionBase(BaseDomainClassData data, DefinitionBase definitionBase)
        {
            if (data == null)
            {
                Report(UpdaterMessagePriority.High, "The {1} '{0}' was deleted from the Datastructrues and will be deleted from Dataflow, too.", definitionBase.Name, definitionBase.GetType().Name);
                return false;
            }

            if (definitionBase.Name != data.Name)
            {
                Report(UpdaterMessagePriority.Low, "Name of {2} '{0}' has changed to '{1}'.", definitionBase.Name, data.Name, definitionBase.GetType().Name);
                definitionBase.Name = data.Name;
            }

            if (definitionBase.Description != data.Description)
                definitionBase.Description = data.Description;

            return true;
        }

        private bool UpdateVariable(VariableData data, Variable variable)
        {
            if (!UpdateDefinitionBase(data, variable))
                return false;

            if (variable.Type != data.Type)
            {
                string oltType = variable.Type != null ? variable.Type.Value : "NULL";
                string newType = data.Type != null ? data.Type.Value : "NULL";
                Report(UpdaterMessagePriority.Normal, "Type of Variable '{0}' has changed from '{1}' to '{2}'.", variable.Name, oltType, newType);
                variable.Type = data.Type != null ? data.Type.Clone() : null;
            }

            if (variable.ExternalVisible != data.ExternalVisible)
                variable.ExternalVisible = data.ExternalVisible;
            if (variable.ExternalWritable != data.ExternalWritable)
                variable.ExternalWritable = data.ExternalWritable;

            return true;
        }

        private bool UpdateOutputEvent(OutputEventData data, OutputEvent outputEvent)
        {
            if (!UpdateDefinitionBase(data, outputEvent))
                return false;

            if (outputEvent.ExternalVisible != data.ExternalVisible)
                outputEvent.ExternalVisible = data.ExternalVisible;
            if (outputEvent.InternalRecordId != data.InternalRecordId)
                outputEvent.InternalRecordId = data.InternalRecordId;

            UpdateFields(data.Fields, outputEvent);

            return true;
        }

        private bool UpdateRecord(RecordData data, Record record)
        {
            if (!UpdateDefinitionBase(data, record))
                return false;

            if (record.RecordId != data.RecordId)
                record.RecordId = data.RecordId;

            UpdateFields(data.Fields, record);

            return true;
        }

        private bool UpdateFunctionCall(FunctionData data, FunctionCall functionCall)
        {
            if (!UpdateDefinitionBase(data, functionCall))
                return false;

            if (functionCall.ExternalVisible != data.ExternalVisible)
                functionCall.ExternalVisible = data.ExternalVisible;
            if (functionCall.InternalRecordId != data.InternalRecordId)
                functionCall.InternalRecordId = data.InternalRecordId;

            UpdateFields(data.Parameters, functionCall);

            return true;
        }

        private bool UpdateFunctionReturn(FunctionData data, FunctionReturn functionReturn)
        {
            if (!UpdateDefinitionBase(data, functionReturn))
                return false;

            if (functionReturn.ExternalVisible != data.ExternalVisible)
                functionReturn.ExternalVisible = data.ExternalVisible;
            if (functionReturn.InternalRecordId != data.InternalRecordIdForReturn)
                functionReturn.InternalRecordId = data.InternalRecordIdForReturn;

            UpdateFields(data.ReturnValues, functionReturn);

            return true;
        }
    }
}
