using System.Collections.Generic;
using BenjaminSchroeter.Flow.HardwareDescription.Data;

namespace BenjaminSchroeter.Flow.DataFlow.Update
{
    /// <summary>
    /// Integrates changes in the HardwareDescription into a DataFlow Model
    /// </summary>
    public class HardwareUpdater : UpdaterBase
    {
        /// <summary>
        /// Updates the specified dataflow from the given HardwareDescription.
        /// </summary>
        /// <param name="dataflow">The dataflow.</param>
        /// <param name="hardware">The hardware.</param>
        public void Update(DataFlowModel dataflow, HardwareDescriptionModelData hardware)
        {
            UpdateFlowBase(dataflow.TriggerFlow, hardware);
            UpdateFlowBase(dataflow.ProgramFlow, hardware);
        }

        /// <summary>
        /// Looks for differences between the Hardware Definition and the elements of the given FlowBase
        /// </summary>
        private void UpdateFlowBase(FlowBase flow, HardwareDescriptionModelData hardware)
        {
            // a list with elements to delete later on 
            // (deleting on the fly is not allowed inside the foreach loop)
            List<FlowElement> toDelete = new List<FlowElement>();

            // check ever element
            foreach (FlowElement flowElement in flow.FlowElements)
            {
                Variable variable = flowElement as Variable;
                if (variable != null && variable.FromHardwareDescription)
                    if (!UpdateVariable(hardware.FindVariable(variable.Guid), variable))
                        toDelete.Add(flowElement);

                Input input = flowElement as Input;
                if (input != null)
                    if (!UpdateInput(hardware.FindInput(input.Guid), input))
                        toDelete.Add(flowElement);

                InputEvent inputEvent = flowElement as InputEvent;
                if (inputEvent != null)
                    if (!UpdateInputEvent(hardware.FindEvent(inputEvent.Guid), inputEvent))
                        toDelete.Add(flowElement);

                Output output = flowElement as Output;
                if (output != null)
                    if (!UpdateOutput(hardware.FindOutput(output.Guid), output))
                        toDelete.Add(flowElement);
            }

            // now, do the deletion
            foreach (FlowElement element in toDelete)
                element.Delete();
        }

        /// <summary>
        /// Handels changes of Name and Description as well as the situation where the base element was deleted from the hardware description
        /// </summary>
        /// <returns>false, if the elment was deleted from the hardware description</returns>
        private bool UpdateDefinitionBase(EnableableDomainClassData data, DefinitionBase definitionBase)
        {
            if (data == null || data.Enabled == false)
            {
                Report(UpdaterMessagePriority.High, "The {1} '{0}' was deleted from the Hardware Definition and will be deleted from Dataflow, too.", definitionBase.Name, definitionBase.GetType().Name);
                return false;
            }

            if (definitionBase.Name != data.Name)
            {
                Report(UpdaterMessagePriority.Low, "Name of {2} '{0}' has changed to '{1}'.", definitionBase.Name, data.Name, definitionBase.GetType().Name);
                definitionBase.Name = data.Name;
            }

            if (definitionBase.Description != data.Description)
                definitionBase.Description = data.Description;

            return true;
        }

        private bool UpdateVariable(VariableData data, Variable variable)
        {
            if (!UpdateDefinitionBase(data, variable))
                return false;

            if (variable.Type != data.Type)
            {
                string oltType = variable.Type != null ? variable.Type.Value : "NULL";
                string newType = data.Type != null ? data.Type.Value : "NULL";
                Report(UpdaterMessagePriority.Normal, "Type of Variable '{0}' has changed from '{1}' to '{2}'.", variable.Name, oltType, newType);
                variable.Type = data.Type != null ? data.Type.Clone() : null;
            }

            if (variable.Readable != data.IsReadable)
                variable.Readable = data.IsReadable;
            if (variable.Writable != data.IsWritable)
                variable.Writable = data.IsWritable;

            return true;
        }

        private bool UpdateOutput(OutputData data, Output output)
        {
            if (!UpdateDefinitionBase(data, output))
                return false;

            UpdateFields(data.Fields, output);

            return true;
        }

        private bool UpdateInputEvent(EventData data, InputEvent inputEvent)
        {
            if (!UpdateDefinitionBase(data, inputEvent))
                return false;

            UpdateFields(data.Fields, inputEvent);

            return true;
        }

        private bool UpdateInput(InputData data, Input input)
        {
            if (!UpdateDefinitionBase(data, input))
                return false;

            UpdateFields(data.Fields, input);

            return true;
        }
    }
}
