using System;
using System.Collections.Generic;
using System.Linq;
using BenjaminSchroeter.Flow.Datastructures.Data;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using FieldBaseData=BenjaminSchroeter.Flow.Datastructures.Data.FieldBaseData;

namespace BenjaminSchroeter.Flow.DataFlow.Update
{
    /// <summary>
    /// Priorities of Messages from the Updaters
    /// </summary>
    public enum UpdaterMessagePriority
    {
        /// <summary>
        /// 
        /// </summary>
        Low,
        /// <summary>
        /// 
        /// </summary>
        Normal,
        /// <summary>
        /// 
        /// </summary>
        High
    }

    /// <summary>
    /// Base class to build updaters to integrate the changes of a model in anoter one 
    /// </summary>
    public abstract class UpdaterBase
    {
        /// <summary>
        /// Occurs when a new (error) message is created.
        /// </summary>
        public event Action<UpdaterMessagePriority, string> Message;

        /// <summary>
        /// Reports a new (error) message.
        /// </summary>
        /// <param name="priority">The priority.</param>
        /// <param name="s">The s.</param>
        /// <param name="parameter">The parameter.</param>
        protected void Report(UpdaterMessagePriority priority, string s, params string[] parameter)
        {
            if (Message == null)
                return;

            string text = string.Format(s, parameter);
            Message(priority, text);
        }

        /// <summary>
        /// Handels the updating of the FieldCollection of FieldedDefinitionBase objects.
        /// 
        /// The properties of each field will be updated, fields will be deleted and added.
        /// </summary>
        /// <remarks>
        /// To write this method all field classes (in the HardwareDescrition: FieldBase and EnableableFieldBase and in Datastructures FieldBase)
        /// must implement the IFieldBase in the DataService.
        /// 
        /// If the DataModel does not provide an enabled property, always true is returend.
        /// </remarks>
        /// <typeparam name="T">The Field Type of the DataService Data classes. See remarks.</typeparam>
        /// <param name="fieldDataCollection">The list of fields from Dataservice (from HardwareDescription oder Datastructures)</param>
        /// <param name="definitionBase">The Element of the Dataflow to update</param>
        protected void UpdateFields<T>(IEnumerable<T> fieldDataCollection, FieldedDefinitionBase definitionBase) where T : IFieldBase
        {
            // find fields that do not longer exist in hardware definition
            // (delete later, since not allowed in foreach)
            List<Field> toDelete = new List<Field>();

            foreach (Field field in definitionBase.Fields)
            {
                // find a definition from dataservice
                T fieldData = fieldDataCollection.FirstOrDefault(x => x.Guid.Equals(field.Guid));
                if (fieldData == null || fieldData.Enabled == false)
                {
                    Report(UpdaterMessagePriority.High, "The Field '{0}' of '{1}' was deleted from the Hardware Definition and will be deleted from Dataflow, too.", field.Name, definitionBase.Name);
                    toDelete.Add(field);
                }
                else
                {
                    // check name and content
                    if (field.Name != fieldData.Name)
                    {
                        Report(UpdaterMessagePriority.Low, "Name of Field '{0}' of '{1}' has changed to '{2}'.", field.Name, definitionBase.Name, fieldData.Name);
                        field.Name = fieldData.Name;
                    }

                    if (field.Type != fieldData.Type)
                    {
                        Report(UpdaterMessagePriority.Low, "Type of Field '{0}' of '{1}' has changed from '{2}' to '{3}'.", field.Name, definitionBase.Name, field.Type.Value, fieldData.Type.Value);
                        field.Type = fieldData.Type.Clone();
                    }

                    if (field.Description != fieldData.Description)
                        field.Description = fieldData.Description;

                    if (fieldData is FieldBaseData)
                    {
                        if (field.StringLength != (fieldData as FieldBaseData).StringLength)
                            field.StringLength = (fieldData as FieldBaseData).StringLength;                        
                    }

                    if (field is OptionalbleField && fieldData is OutputFieldData)
                    {
                        OptionalbleField of = (OptionalbleField)field;
                        OutputFieldData od = fieldData as OutputFieldData;

                        if (of.Optional != od.Optional)
                            of.Optional = od.Optional;
                    }

                    if (field is OptionalbleFieldWithDefault
                        && (fieldData is RecordFieldData || fieldData is OutputEventFieldData || fieldData is ReturnValueData))
                    {
                        OptionalbleFieldWithDefault of = (OptionalbleFieldWithDefault)field;
                        OptionalFieldData od = fieldData as OptionalFieldData;

                        if (of.Optional != od.Optional)
                            of.Optional = od.Optional;

                        if (of.DefaultValue != od.DefaultValue)
                            of.DefaultValue = od.DefaultValue;
                    }
                }
            }
            // now delete the fields markted for deletion
            foreach (Field field in toDelete)
                field.Delete();

            // find new fields in hardware definition
            foreach (T fieldData in fieldDataCollection)
                if (fieldData.Enabled)
                {
                    // does this field exist in the datafloe element?
                    bool fieldExists = definitionBase.Fields.FirstOrDefault(x => x.Guid.Equals(fieldData.Guid)) != null;
                    if (!fieldExists)
                    {
                        // create new Field
                        // for Output and Record special types with special property are used
                        Field f;
                        if ( definitionBase is Output )
                        {
                            OptionalbleField fo = new OptionalbleField(definitionBase.Partition);
                            OutputFieldData outputFieldData = fieldData as OutputFieldData;
                            if ( outputFieldData != null)
                                fo.Optional = outputFieldData.Optional;
                            f = fo;
                        }
                        else if (definitionBase is Record)
                        {
                            OptionalbleFieldWithDefault fod = new OptionalbleFieldWithDefault(definitionBase.Partition);
                            RecordFieldData recordFieldData = fieldData as RecordFieldData;
                            if ( recordFieldData != null)
                            {
                                fod.Optional = recordFieldData.Optional;
                                fod.DefaultValue = recordFieldData.DefaultValue;
                            }
                            f = fod;
                        }
                        else
                            f = new Field(definitionBase.Partition);
                        f.Name = fieldData.Name;
                        f.Guid = fieldData.Guid;
                        f.Type = fieldData.Type.Clone();
                        f.Description = fieldData.Description;
                        // add the filed
                        definitionBase.Fields.Add(f);

                        Report(UpdaterMessagePriority.Low, "Added Field '{0}' to '{1}'.", f.Name, definitionBase.Name);
                    }
                }
        }
    }
}