using System;
using System.Diagnostics;
using BenjaminSchroeter.Dsl.DynamicProperties;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.DataFlow
{
    // the functions in this file control what other elements can be merged to one flow

    partial class FlowBase
    {
        #region can merge
        private bool CanMergeOperation(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            // do not allow reparenting from another swinlane
            if (!IsMergeingFromToolbox(prototype))
                return false;

            // if Operation.KindDomainPropertyId == null the operation is not valid
            return element.GetPropertyValue(Operation.KindDomainPropertyId) != null;
        }

        private bool CanMergeFormula(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            // do not allow reparenting from another swinlane
            return IsMergeingFromToolbox(prototype);
        }
        
        private bool CanMergeCodeBlock(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            // do not allow reparenting from another swinlane
            return IsMergeingFromToolbox(prototype);
        }
        
        private bool CanMergeVariable(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            return CanMergeDefinitionBase(element, prototype);
        }

        private bool CanMergeFunctionCall(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            return CanMergeDefinitionBase(element, prototype);
        }

        private bool CanMergeInput(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            return CanMergeDefinitionBase(element, prototype);
        }

        private bool CanMergeRecord(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            return CanMergeDefinitionBase(element, prototype);
        }

        private bool CanMergeInputEvent(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            // input event are not allowed in the ProgramFlow 
            // (if someone adds them to the TriggerFlow they will be created in
            //  the ProgramFlow, too, but the user is not allowed to to this
            //  merge by himself)
            if (this is ProgramFlow)
                return false;

            return CanMergeDefinitionBase(element, prototype);
        }

        // Used by all elements that inherit form DefinitionBase
        protected bool CanMergeDefinitionBase(ProtoElementBase element, ElementGroupPrototype prototype)
        {
            // do not allow reparenting from another swinlane
            if (!IsMergeingFromToolbox(prototype))
                return false;
            
            // check if the Guid is set
            // allow the merge only if the Guid is not Empty
            // (all Elementes in the Toolbar have a valid Guid, new Elements from DSL Explorer context menu doesn't)
            object o = element.GetPropertyValue(DefinitionBase.GuidDomainPropertyId);
            Guid guidOfDefinitionBase = (Guid) (o ?? Guid.Empty);

            return !guidOfDefinitionBase.Equals(Guid.Empty);
        }

        /// <summary>
        /// Determines whether this element is mergeing from the toolbox.
        /// </summary>
        protected bool IsMergeingFromToolbox(ElementGroupPrototype prototype)
        {
            return CopyOriginContext.GetCopyOrigin(prototype).Equals("Toolbox",StringComparison.InvariantCultureIgnoreCase);
        }
        #endregion

        #region merge InputEvent
        private void MergeRelateInputEvent(ModelElement element, ElementGroup group)
        {
            // if an input event is added to the TriggerFlow the event should be
            // created in the ProgramFlow as well (if it has fields)

            TriggerFlow flow = this as TriggerFlow;
            if (flow != null)
            {
                // add the input to this flow (this is a trigger flow since the merge only allows trigger flows)
                InputEvent input = (InputEvent)element;
                if ( input.Parameter != null)
                {
                    // clone the parameter value (to avoid the same refence for all events)
                    DynamicPropertyCollectionTypeConverter converter = new DynamicPropertyCollectionTypeConverter();
                    string s = (string) converter.ConvertTo(input.Parameter,typeof(string));
                    input.Parameter = (DynamicPropertyCollection) converter.ConvertFrom(s);
                }

                flow.FlowElements.Add(input);
                
                // and add a copy to the program flow
                if ( input.Fields.Count > 0 )
                {
                    InputEvent copy = (InputEvent)input.Copy();
                    copy.Parameter = null;
                    flow.DataFlowModel.ProgramFlow.FlowElements.Add(copy);
                }
            }
        }

        private void MergeDisconnectInputEvent(ModelElement element)
        {
            Debug.Fail("should not be called");
        }
        #endregion

        #region merge Function
        private void MergeRelateFunctionCall(ModelElement element, ElementGroup group)
        {
            // if a function call is added to the Trigger or ProgramFlow
            // some more elements must be added to the other flow.

            FunctionCall fcall = (FunctionCall) element;
            
            TriggerFlow tflow = this as TriggerFlow;
            if ( tflow != null ) // call added to trigger flow
            {
                // add call and trigger flow and program flow
                AddFunction(tflow, fcall);
                AddFunction(tflow.DataFlowModel.ProgramFlow, fcall);
            }

            ProgramFlow pflow = this as ProgramFlow;
            if ( pflow != null ) // call added to program flow
            {
                // add call and trigger flow and program flow
                AddFunction(pflow, fcall);
                AddFunction(pflow.DataFlowModel.TriggerFlow, fcall);
            }
        }

        private void AddFunction(TriggerFlow flow, FunctionCall fcall)
        {
            FunctionCall fcallcopy = (FunctionCall)fcall.Copy();
            flow.FlowElements.Add(fcallcopy);
        }
        private void AddFunction(ProgramFlow flow, FunctionCall fcall)
        {
            if (fcall.Fields.Count > 0)
                flow.FlowElements.Add((FunctionCall)fcall.Copy());

            if (fcall.FunctionReturn.Fields.Count > 0 )
                flow.FlowElements.Add((FunctionReturn) fcall.FunctionReturn.Copy());
        }

        private void MergeDisconnectFunctionCall(ModelElement element)
        {
            Debug.Fail("should not be called");
        }
        #endregion
    }
}