using System;
using System.Text.RegularExpressions;
using BenjaminSchroeter.Flow.DataTypes;
using Microsoft.VisualStudio.Modeling.Validation;

namespace BenjaminSchroeter.Flow.DataFlow
{
    [ValidationState(ValidationState.Enabled)]
    partial class FlowElementReferencesTargetFlowElements
    {
        private static readonly Regex validCNameRegex = new Regex("[a-zA-Z_][a-zA-Z0-9_]*");
        
        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateAnnotations(ValidationContext context)
        {
            Formula formula = this.TargetFlowElement as Formula;
            if ( formula != null)
            {
                // this is a link to a formula shape
                // a annotation is needed
                if ( this.TargetAnnotation == null || this.TargetAnnotation.Trim() == "" )
                {
                    string msg = string.Format("The annotation of the connection to the Formula '{0}' must not be empty.",formula.FormulaText);
                    context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateAnnotations", this);
                }
                else
                {
                    // the name is not empty, but is es a valid code name
                    Match m = validCNameRegex.Match(this.TargetAnnotation);
                    if (m.Value != this.TargetAnnotation)
                    {
                        string msg = string.Format("The annotation name '{0}' is not a valid identifier.", this.TargetAnnotation);
                        context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateAnnotations", this);
                    }
                }
            }
            
            
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateGuids(ValidationContext context)
        {
            // search the source and target guid in the source and target elements

            if ( this.SourceFlowElement == null )
            {
                string msg = "The Source of the Connection is not a valid Shape.";
                context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateGuids", this);                
            }
            if (this.TargetFlowElement == null)
            {
                string msg = "The Target of the Connection is not a valid Shape.";
                context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateGuids", this);
            }

            if ( this.SourceFlowElement != null && !this.SourceEntry.Equals(Guid.Empty))
                if (this.SourceField == null)
                {
                    string msg = string.Format("The Source Guid of the Connection can not be found in the Source Shape '{0}': {1}.", this.SourceFlowElement, this.SourceEntry);
                    context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateGuids", this);
                }
            if (this.TargetFlowElement != null && !this.TargetEntry.Equals(Guid.Empty))
                if (this.TargetField == null)
                {
                    string msg = string.Format("The Target Guid of the Connection can not be found in the Target Shape '{0}': {1}.", this.TargetFlowElement, this.TargetEntry);
                    context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateGuids", this);
                }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateType(ValidationContext context)
        {
            // a formula accepts every datatype
            if ( this.TargetFlowElement is Formula)
                return;

            // and the master trigger can accept a record head
            if ( this.SourceFlowElement is Record && this.SourceEntry.Equals(Guid.Empty)
                 && this.TargetFlowElement is MasterTrigger )
                return;

            DataTypeString from = this.GetTypeValue();
            DataTypeString to = this.GetTargetTypeValue();
            if ( from.Value != to.Value )
            {
                string msg = string.Format("The connection from '{0}' to '{1}' is invalid, since the Datatypes of Source ('{2}') and Target ('{3}') are not the same.", this.SourceFlowElement, this.TargetFlowElement, from.Value, to.Value);
                context.LogError(msg, "FlowElementReferencesTargetFlowElements.ValidateType", this);
            }
        }
    }
}
