using System;
using System.Collections.Generic;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow.Graphs;
using BenjaminSchroeter.Dsl.Graphs;
using Microsoft.VisualStudio.Modeling.Validation;

namespace BenjaminSchroeter.Flow.DataFlow
{
    [ValidationState(ValidationState.Enabled)]
    partial class TriggerFlow
    {
        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateInputEventCount(ValidationContext context)
        {
            int numberOfInputEvents = this.FlowElements.OfType<InputEvent>().Count();
            int numberOfFunctionCalls = this.FlowElements.OfType<FunctionCall>().Count();
            if (numberOfInputEvents + numberOfFunctionCalls == 0 )
            {
                string msg = "At least one Event or Function Call is needed for the trigger part of this model.";
                context.LogError(msg, "TriggerFlow.ValidateInputEventCount", this);
            }

            if ( numberOfFunctionCalls > 1 )
            {
                string msg = "Only one Function Call is allowed in the trigger part of this model.";
                context.LogError(msg, "TriggerFlow.ValidateInputEventCount", this.FlowElements.OfType<FunctionCall>().ToArray());
            }

            if ( numberOfFunctionCalls > 0 && numberOfInputEvents > 0)
            {
                string msg = "Only one Function Call or Events are allowed in the trigger part of this model, not both.";
                context.LogError(msg, "TriggerFlow.ValidateInputEventCount", this);                
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateEventUniqueness(ValidationContext context)
        {
            var events = this.FlowElements.OfType<InputEvent>();
            HashSet<Guid> guids = new HashSet<Guid>();
            HashSet<Guid> duplicate = new HashSet<Guid>();

            foreach (InputEvent ev in events)
            {
                if (!guids.Contains(ev.Guid))
                    guids.Add(ev.Guid);
                else
                    duplicate.Add(ev.Guid);
            }

            foreach (Guid guid in duplicate)
            {
                var duplicateEvents = this.FlowElements.OfType<InputEvent>().Where(x => x.Guid.Equals(guid));
                string msg = string.Format("The Event '{0}' must only have one instance within the trigger part of this model.", duplicateEvents.First().Name );
                context.LogError(msg, "TriggerFlow.ValidateEventUniqueness", duplicateEvents.ToArray());
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateEventGraphComponentsOnlyOneEvent(ValidationContext context)
        {
            // two parts of the graph that start from an event must not share any
            // flow elements besides the master trigger
            UndirectedDataFlowGraph triggerGraph = new UndirectedDataFlowGraph(this); 
            MasterTrigger t = (MasterTrigger) triggerGraph.Vertices().FirstOrDefault(x => x is MasterTrigger);
            triggerGraph.RemoveVertice(t);
            DepthFirstSearch<FlowElement, FlowElementReferencesTargetFlowElements> dfs = new DepthFirstSearch<FlowElement, FlowElementReferencesTargetFlowElements>(triggerGraph);

            foreach (HashSet<FlowElement> elements in dfs.ConnectedComponents)
            {
                if ( elements.OfType<InputEvent>().Count() > 1 )
                {
                    string msg = "Each sub graph of the tigger part must contain only one event.";
                    context.LogError(msg, "TriggerFlow.ValidateEventGraphComponentsOnlyOneEvent", elements.ToArray());
                }
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateEventGraphComponentsLeadToTrigger(ValidationContext context)
        {
            // every element of the graph mit lead to the trigger
            UndirectedDataFlowGraph triggerGraph = new UndirectedDataFlowGraph(this);
            DepthFirstSearch<FlowElement, FlowElementReferencesTargetFlowElements> dfs = new DepthFirstSearch<FlowElement, FlowElementReferencesTargetFlowElements>(triggerGraph);

            foreach (HashSet<FlowElement> elements in dfs.ConnectedComponents)
            {
                if (elements.OfType<MasterTrigger>().Count() == 0)
                {
                    // the elements of this subgraph do not lead to the master trigger
                    foreach (FlowElement element in elements)
                    {
                        string type = element.GetType().Name;
                        string msg = string.Format("The {0} '{1}' is not connected with the rest of the trigger part of the model. It must be connected with the master trigger in some way.", type,element.ToString());
                        context.LogError(msg, "TriggerFlow.ValidateEventGraphComponentsLeadToTrigger", element );
                    }
                }
            }
        }
    }
}
