using System;
using System.Collections;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.IO;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Modeling.Shell;

namespace BenjaminSchroeter.Flow.DataFlow
{
    /// <summary>
    /// Handler for the "Open Code Block Editor..." comamnd
    /// </summary>
    internal class OpenCodeBlockEditorCommandHandler
    {
        private const string guidDataflowString = "73598FC8-D6B1-4c67-89EB-BFD892E735C3";
        public static readonly Guid guidDataflow = new Guid(guidDataflowString);
        public const uint cmdId = 0x0100;

        /// <summary>
        /// Queries the status for this comamnd.
        /// 
        /// It will be only visible if a single code block is selected and
        /// a CodeBlockEditorService is available
        /// </summary>
        public static void QueryStatus(object sender, ICollection currentDocumentSelection, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            if (!IsCodeBlockEditorAvailable(serviceProvider))
            {
                cmd.Enabled = cmd.Visible = false;
                return;
            }

            cmd.Enabled = cmd.Visible = GetCodeBlockFromSelection(currentDocumentSelection) != null;
        }

        /// <summary>
        /// Opens the Code Block Editor whenever the command is invoked
        /// </summary>
        public static void Invoke(ICollection currentDocumentSelection, DocData docData, IServiceProvider serviceProvider)
        {
            // if the code block editor service is not available -> exit
            if (!IsCodeBlockEditorAvailable(serviceProvider))
                return;

            // find the codeblock
            CodeBlock codeBlock = GetCodeBlockFromSelection(currentDocumentSelection);
            Debug.Assert(codeBlock != null, "No CodeBlock selected");

            // get the service
            ICodeBlockEditor editor = serviceProvider.GetService(typeof (SCodeBlockEditor)) as ICodeBlockEditor;
            if ( editor == null )
                return;

            // and open the editor
            editor.OpenCodeBlockEditor(codeBlock, Path.GetFileNameWithoutExtension(docData.FileName), docData.Hierarchy);
        }

        /// <summary>
        /// Gets the single selected CodeBlock from current document selection.
        /// </summary>
        /// <param name="currentDocumentSelection">The current document selection.</param>
        /// <returns></returns>
        private static CodeBlock GetCodeBlockFromSelection(ICollection currentDocumentSelection)
        {
            if (currentDocumentSelection.Count == 1)
                foreach (ModelElement modelElement in currentDocumentSelection)
                {
                    if (modelElement is CodeBlockShape)
                    {
                        // get the codeblock from selected shape 
                        return (modelElement as CodeBlockShape).ModelElement as CodeBlock;
                    }
                 
                    // get the codeblock from seleced field (on double click)
                    if (modelElement is CodeBlockInputField)
                        return (modelElement as CodeBlockInputField).CodeBlock;
                    if (modelElement is CodeBlockOutputField)
                        return (modelElement as CodeBlockOutputField).CodeBlock;

                    // get the codeblock if one of the ElementListCompartments of the CodeBlockShape is selected
                    if ( modelElement is ElementListCompartment)
                        return (modelElement as ElementListCompartment).ParentShape.ModelElement as CodeBlock;
                }
            
            return null;
        }

        // caches this value:
        //  -1 = not cached
        //   0 = CoodeBlockEditor is not available
        //   1 = it is available
        private static int codeBlockEditorAvailibility = -1;
        private static bool IsCodeBlockEditorAvailable(IServiceProvider serviceProvider)
        {
            if ( codeBlockEditorAvailibility == -1 )
            {
                ICodeBlockEditor editor = serviceProvider.GetService(typeof(SCodeBlockEditor)) as ICodeBlockEditor;
                codeBlockEditorAvailibility = editor == null ? 0 : 1;
            }

            return codeBlockEditorAvailibility == 1;
        }
    }
}