using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.Flow.DataFlow.Update;
using BenjaminSchroeter.Flow.Datastructures.Data;
using BenjaminSchroeter.Flow.Datastructures.DataService;
using BenjaminSchroeter.Flow.DataTypes;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using BenjaminSchroeter.Flow.HardwareDescription.DataService;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Shell;

namespace BenjaminSchroeter.Flow.DataFlow
{
    partial class DataFlowDocView
    {
        protected override bool LoadView()
        {
            try
            {
                InitErrorList();

                bool open = base.LoadView();
                if (open)
                {
                    // get the HierarchyItem of this diagram
                    HierarchyItem hierarchyItem = GetHierarchyItem();
                    using (Transaction t = this.DocData.Store.TransactionManager.BeginTransaction("Update base changes"))
                    {
                        // integrate the changes form hardware definition...
                        FindHardwareChanges(hierarchyItem);
                        // ... and datastructures
                        FindDatastructureChanges(hierarchyItem);

                        // do a commit only if it is nessesary 
                        if (t.HasPendingChanges)
                            t.Commit();
                        else
                            t.Rollback();

                        // if there were errors -> show them
                        if (errorList.Tasks.Count > 0)
                        {
                            // these events will clear the error list on saving the document or closing the editor
                            this.DocData.DocumentSaved += ClearErrorList;
                            this.DocData.DocumentClosing += ClearErrorList;

                            // show the error list and the form (it will be visible unter the modal message box)
                            errorList.Show();
                            this.Frame.ShowNoActivate();

                            // maybe the document is not realy open now
                            // (closed by the rtd event listener in the UiSupportPackage)
                            if (this.DocData == null)
                                return false;

                            // inform the user with a messagebox
                            string name =  Path.GetFileName(this.DocData.FileName);
                            DialogResult d = MessageBox.Show(
@"Changes detected in the Hardware Description or Datastructures used by " + name + ".\n" + 
@"This file was changed to reflect these changes.
See Error List for Details.

Do you realy want to open the file?

(If you open it, you are still able to close it without saving!)", "Changes in " + name, MessageBoxButtons.YesNo, MessageBoxIcon.Question);

                            return (d == DialogResult.Yes);
                        }
                    }
                }
                else
                    return false;
            }
            catch (Exception e)
            {
                // otherwise the DSL framwork will catch the exeption
                Debug.Fail(e.Message,e.ToString());
                throw;
            }

            return true;
        }

        #region error list
        private ErrorListProvider errorList;

        private void InitErrorList()
        {
            errorList = new ErrorListProvider(this)
                            {
                                ProviderName = "Dataflow",
                                ProviderGuid = new Guid("61B55AE2-4853-4a48-823E-A1B4CF84F0BA")
                            };
        }

        /// <summary>
        /// Removes all tasks of this file from the error list 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ClearErrorList(object sender, EventArgs e)
        {
            List<Task> toDelete = new List<Task>();
            foreach (Task task in errorList.Tasks)
            {
                if (task.Document.Equals(this.DocData.FileName,StringComparison.InvariantCultureIgnoreCase) )
                    toDelete.Add(task);
            }

            foreach (Task task in toDelete)
                this.errorList.Tasks.Remove(task);
        }

        /// <summary>
        /// Callback used by the model updater.
        /// </summary>
        /// <param name="priority">The priority.</param>
        /// <param name="text">The text.</param>
        private void updater_Message(UpdaterMessagePriority priority, string text)
        {
            Task t = new ErrorTask
            {
                ErrorCategory = TaskErrorCategory.Message,
                Text = text,
                Document = this.DocData.FileName
            };

            errorList.Tasks.Add(t);
        }
        #endregion

        #region Update elements from Hardware description and datastructures
        /// <summary>
        /// Looks for differences between the Hardware Definition and this Dataflow Model
        /// </summary>
        /// <param name="hierarchyItem">The HierarchyItem of the Dataflow</param>
        private void FindHardwareChanges(HierarchyItem hierarchyItem)
        {
            // get the Dataflow model
            DataFlowModel dataflow = (DataFlowModel)this.DocData.RootElement;

            // get the data service
            IHardwareDescriptionDataService dataservice = this.GetService(typeof(SHardwareDescriptionDataService)) as IHardwareDescriptionDataService;
            Debug.Assert(dataservice != null, "IHardwareDescriptionDataService");

            HardwareDescriptionModelData hardware;
            try
            {
                hardware = dataservice.GetHardwareDescription(hierarchyItem.VsHierarchy);
            }
            catch 
            {
                // no message will be shown here, since a message of this error will be displayed from toolbox creation as well
                return;
            }

            HardwareUpdater updater = new HardwareUpdater();
            updater.Message += updater_Message;
            updater.Update(dataflow,hardware);

            RegisterDataTypesFromHardwareDescription(dataflow, hardware);
        }

        private void RegisterDataTypesFromHardwareDescription(DataFlowModel dataflow, HardwareDescriptionModelData hardware)
        {
            foreach (DataTypeData data in hardware.DataTypes)
                DataTypeConverter.AddDataType(this.DocData.Store, data.Name, data.CType, data.DslType);
                    
            dataflow.DataTypes = hardware.DataTypes.Select(x => x.Name).ToArray();
        }

        /// <summary>
        /// Looks for differences between the Datastructures and this Dataflow Model
        /// </summary>
        /// <param name="hierarchyItem">The HierarchyItem of the Dataflow</param>
        private void FindDatastructureChanges(HierarchyItem hierarchyItem)
        {
            // get the dataflow model
            DataFlowModel dataflow = (DataFlowModel)this.DocData.RootElement;

            // get the data service
            IDatastructureDataService dataservice = this.GetService(typeof(SDatastructureDataService)) as IDatastructureDataService;
            Debug.Assert(dataservice != null, "IDatastructureDataService");

            DatastructureModelData datastructs = dataservice.GetAllDatastructures(hierarchyItem.VsHierarchy);

            DatastructuresUpdater updater = new DatastructuresUpdater();
            updater.Message += updater_Message;
            updater.Update(dataflow, datastructs);
        }
        #endregion
    }
}