using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Design;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.Flow.Datastructures.Data;
using BenjaminSchroeter.Flow.Datastructures.DataService;
using BenjaminSchroeter.Dsl.DynamicProperties;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using BenjaminSchroeter.Flow.HardwareDescription.DataService;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Design;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Shell.Interop;
using IDataObject=Microsoft.VisualStudio.OLE.Interop.IDataObject;
using ParameterData=BenjaminSchroeter.Flow.Datastructures.Data.ParameterData;
using VariableData=BenjaminSchroeter.Flow.HardwareDescription.Data.VariableData;

namespace BenjaminSchroeter.Flow.DataFlow
{
    partial class DataFlowDocView
    {
        private const string PLACEHOLDER_NAME = "Placeholder";
        private const string DATASTRUCTURES_TOOLBOX_NAME = "Custom Datastructures";
        private string ToolboxItemKey;
        private IVsToolbox2 vsToolbox2;
        private IToolboxService toolboxService;

        protected override void OnCreate()
        {
            base.OnCreate();

            // get services
            vsToolbox2 = this.GetService(typeof(IVsToolbox)) as IVsToolbox2;
            Debug.Assert(this.vsToolbox2 != null, "IVsToolbox2");

            toolboxService = base.GetService(typeof(IToolboxService)) as IToolboxService;
            Debug.Assert(this.toolboxService != null, "IToolboxService");

            try
            {
                HierarchyItem hierarchyItem = GetHierarchyItem();
                if (hierarchyItem != null)
                {
                    // create a toolbox key for this project
                    string projectName = "UNKNOWN";
                    HierarchyItem rootItem = hierarchyItem.GetRoot();
                    if (rootItem != null)
                        projectName = rootItem.Name;
                    ToolboxItemKey = DataFlowToolboxHelperBase.ToolboxFilterString + "." + projectName;

                    // Create store and load domain model types.
                    using (Store store = new Store(this))
                    {
                        store.LoadDomainModels(typeof(CoreDesignSurfaceDomainModel), typeof(DataFlowDomainModel));

                        // Open transaction so we can create model elements corresponding to toolbox items.
                        using (Transaction t = store.TransactionManager.BeginTransaction("CreateToolboxItems"))
                        {
                            CreateDatastructureToolbar(store, hierarchyItem);
                            CreateHardwareToolbar(store, hierarchyItem);

                            t.Rollback();
                        }
                    }
                }

                RemoveEmptyDummyTabs();
            }
            catch (Exception e)
            {
                Debug.Fail(e.Message,e.ToString());
                throw;
            }

            // no longer needed
            vsToolbox2 = null;
            toolboxService = null;
        }

        /// <summary>
        /// Creates toolbar items for every element of all datastructure models in the current project.
        /// </summary>
        /// <param name="store">The store.</param>
        /// <param name="hierarchyItem">The hierarchy item.</param>
        private void CreateDatastructureToolbar(Store store, HierarchyItem hierarchyItem)
        {
            IDatastructureDataService dataservice = this.GetService(typeof(SDatastructureDataService)) as IDatastructureDataService;
            Debug.Assert(dataservice != null, "IDatastructureDataService");

            // get the datastructures for this project
            DatastructureModelData datastructs = dataservice.GetAllDatastructures(hierarchyItem.VsHierarchy);

            // add a tab in the toolbox
            vsToolbox2.AddTab(DATASTRUCTURES_TOOLBOX_NAME); 
            
            // find all the current Toolbox Items
            Dictionary<string, ModelingToolboxItem> list = GetVersionedModelingToolboxItems(DATASTRUCTURES_TOOLBOX_NAME);

            // add a dummy entry to the tab (otherwise: empty tabs are always visible)
            AddDummyToolboxItem(DATASTRUCTURES_TOOLBOX_NAME, list);

            // add new toolbox items
            foreach (OutputEventData data in datastructs.OutputEvents)
                AddToolboxItem(DATASTRUCTURES_TOOLBOX_NAME, data.Name, CreateElementToolPrototype(store, data), Resources.EventSmall, list, data.ModelVersion, ToolboxItemKey);
            foreach (RecordData data in datastructs.Records)
                AddToolboxItem(DATASTRUCTURES_TOOLBOX_NAME, data.Name, CreateElementToolPrototype(store, data), Resources.RecordSmall, list, data.ModelVersion, ToolboxItemKey);
            foreach (BenjaminSchroeter.Flow.Datastructures.Data.VariableData data in datastructs.Variables)
                AddToolboxItem(DATASTRUCTURES_TOOLBOX_NAME, data.Name, CreateElementToolPrototype(store, data), Resources.VariableSmall, list, data.ModelVersion, ToolboxItemKey);
            foreach (FunctionData data in datastructs.Functions)
                AddToolboxItem(DATASTRUCTURES_TOOLBOX_NAME, data.Name, CreateElementToolPrototype(store, data), Resources.FunctionSmall, list, data.ModelVersion, ToolboxItemKey);

            // remove not longer needed ones
            foreach (VersionedModelingToolboxItem item in list.Values)                                          
                toolboxService.RemoveToolboxItem(item, DATASTRUCTURES_TOOLBOX_NAME);
        }

        /// <summary>
        /// Creates toolbar items for every element of the hardware decription of the current project.
        /// </summary>
        /// <param name="store">The store.</param>
        /// <param name="hierarchyItem">The hierarchy item.</param>
        private void CreateHardwareToolbar(Store store, HierarchyItem hierarchyItem)
        {
            IHardwareDescriptionDataService dataservice = this.GetService(typeof(SHardwareDescriptionDataService)) as IHardwareDescriptionDataService;
            Debug.Assert(dataservice != null, "IHardwareDescriptionDataService");

            HardwareDescriptionModelData hardware;
            try
            {
                hardware = dataservice.GetHardwareDescription(hierarchyItem.VsHierarchy);
            }
            catch (WrongNumberOfHardwareDescriptionException e)
            {
                MessageBox.Show(e.Message, "Error loading Dataflow", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            foreach (PlatformData platform in hardware.Platforms)
            {
                // add one tab for every platform (that is every driver)
                string toolboxTabName = platform.Name;
                vsToolbox2.AddTab(toolboxTabName); 

                Dictionary<string, ModelingToolboxItem> list = GetVersionedModelingToolboxItems(toolboxTabName);

                // add a dummy entry to the tab (otherwise: empty tabs are always visible)
                AddDummyToolboxItem(toolboxTabName, list);

                // add new toolbox items
                foreach (InputData data in platform.Inputs.Where(x => x.Enabled && x.Fields.Any( f => f.Enabled ) ))
                    AddToolboxItem(toolboxTabName, data.Name, CreateElementToolPrototype(store, data), Resources.InputSmall, list, data.ModelVersion, ToolboxItemKey);
                foreach (OutputData data in platform.Outputs.Where(x => x.Enabled && x.Fields.Any(f => f.Enabled)))
                    AddToolboxItem(toolboxTabName, data.Name, CreateElementToolPrototype(store, data), Resources.OutputSmall, list, data.ModelVersion, ToolboxItemKey);
                foreach (EventData data in platform.Events.Where(x => x.Enabled))
                    AddToolboxItem(toolboxTabName, data.Name, CreateElementToolPrototype(store, data, hardware), Resources.EventSmall, list, data.ModelVersion, ToolboxItemKey);
                foreach (VariableData data in platform.Variables.Where(x => x.Enabled))
                    AddToolboxItem(toolboxTabName, data.Name, CreateElementToolPrototype(store, data), Resources.VariableSmall, list, data.ModelVersion, ToolboxItemKey);

                // remove not longer needed ones
                foreach (VersionedModelingToolboxItem item in list.Values)
                    toolboxService.RemoveToolboxItem(item, toolboxTabName);
            }
        }

        /// <summary>
        /// Adds a new toolbox item to the toolbox.
        /// </summary>
        /// <param name="toolboxTabName">Name of the toolbox tab.</param>
        /// <param name="itemName">Name of the item.</param>
        /// <param name="prototype">The prototype for the toolbox item.</param>
        /// <param name="bitmap">Icon.</param>
        /// <param name="oldItems">The old items.</param>
        /// <param name="currentVersion">The current version.</param>
        /// <param name="toolboxItemKey">The toolbox item key.</param>
        private void AddToolboxItem(string toolboxTabName, 
            string itemName, 
            ElementGroupPrototype prototype, 
            Bitmap bitmap,
            Dictionary<string, ModelingToolboxItem> oldItems, 
            int currentVersion, 
            string toolboxItemKey)
        {
            // if an element with the same name already exists
            VersionedModelingToolboxItem oldItem = null;
            if (oldItems.ContainsKey(itemName))
            {
                oldItem = oldItems[itemName] as VersionedModelingToolboxItem;

                // remove it from list: we will leave it (if version is the same) or create a new in this method
                oldItems.Remove(itemName);

                if (oldItem != null && oldItem.ModelVersion == currentVersion)
                {
                    // we don't need to change the toolbox item.
                    return;
                }
            }

            VersionedModelingToolboxItem item = new VersionedModelingToolboxItem(
                itemName,        // Unique identifier (non-localized) for the toolbox item.
                1, // Position relative to other items in the same toolbox tab.
                itemName, // Localized display name for the item.
                bitmap,   // Image displayed next to the toolbox item.
                toolboxTabName, // Unique identifier (non-localized) for the toolbox item tab.
                toolboxTabName,// Localized display name for the toolbox tab.
                "f1help", // F1 help keyword for the toolbox item.  
                "",// Localized tooltip text for the toolbox item.
                prototype, // ElementGroupPrototype (data object) representing model element on the toolbox.
                new[]{
                         // Collection of ToolboxItemFilterAttribute objects that determine visibility of the toolbox item.
						//new global::System.ComponentModel.ToolboxItemFilterAttribute(ToolboxFilterString, global::System.ComponentModel.ToolboxItemFilterType.Require) 
                         new ToolboxItemFilterAttribute(toolboxItemKey, System.ComponentModel.ToolboxItemFilterType.Require ),});
                       //  new ToolboxItemFilterAttribute(DataFlowToolboxHelperBase.ToolboxFilterString , System.ComponentModel.ToolboxItemFilterType.Require ),});

            item.IsTransient = true;
            item.ModelVersion = currentVersion;
            item.Project = toolboxItemKey;

            if ( oldItem != null)
                toolboxService.RemoveToolboxItem(oldItem, toolboxTabName);

            toolboxService.AddToolboxItem(item,toolboxTabName);
        }

        #region create prototypes for hardware description itmes
        private static ElementGroupPrototype CreateElementToolPrototype(Store store, InputData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);
            
            Input i = (Input) store.ElementFactory.CreateElement(Input.DomainClassId);
            i.Name = data.Name;
            i.Description = data.Description;
            i.Guid = data.Guid;
            
            foreach (InputFieldData fieldData in data.Fields.Where(x => x.Enabled))
            {
                Field f = new Field(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                i.Fields.Add(f);
            }
            
            elementGroup.AddGraph(i, true);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, VariableData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);
             
            Variable v = (Variable)store.ElementFactory.CreateElement(Variable.DomainClassId);
            v.Name = data.Name;
            v.Description = data.Description;
            v.Guid = data.Guid;
            v.Type = data.Type != null ? data.Type.Clone() : null;
            v.ExternalWritable = false;
            v.ExternalVisible = false;
            v.Readable = data.IsReadable;
            v.Writable = data.IsWritable;
            v.FromHardwareDescription = true;
          
            elementGroup.AddGraph(v, true);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, EventData data, HardwareDescriptionModelData hardware)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);
             
            InputEvent e = (InputEvent)store.ElementFactory.CreateElement(InputEvent.DomainClassId);
            e.Name = data.Name;
            e.Description = data.Description;
            e.Guid = data.Guid;
            if (data.Multiplicity > 1)
            {
                DynamicPropertyCollection props = new DynamicPropertyCollection();
                foreach (BenjaminSchroeter.Flow.HardwareDescription.Data.ParameterData parameter in data.Parameters)
                {
                    DataTypeData datatype = hardware.DataTypes.FirstOrDefault(x => x.Name == parameter.Type.Value);
                    Type type = DataTypeHelper.DslType2DotNet(datatype.DslType);

                    object defaultValue = null;
                    try
                    {
                        defaultValue = TypeDescriptor.GetConverter(type).ConvertFromString(parameter.Value);
                    }
                    catch { }

                    props.AddProperty(parameter.Name, type, defaultValue);
                }

                e.Parameter = props;
            }

            foreach (EventFieldData fieldData in data.Fields)
            {
                Field f = new Field(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                e.Fields.Add(f);
            }

            elementGroup.AddGraph(e, true);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, OutputData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);
             
            Output o = (Output)store.ElementFactory.CreateElement(Output.DomainClassId);
            o.Name = data.Name;
            o.Description = data.Description;
            o.Guid = data.Guid;

            foreach (OutputFieldData fieldData in data.Fields.Where(x => x.Enabled))
            {
                OptionalbleField f = new OptionalbleField(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                f.Optional = fieldData.Optional;
                o.Fields.Add(f);
            }

            elementGroup.AddGraph(o, true);
            return elementGroup.CreatePrototype();
        }
        #endregion

        #region create prototypes for datastructure itmes
        private static ElementGroupPrototype CreateElementToolPrototype(Store store, FunctionData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);

            // Function call
            FunctionCall fcall = (FunctionCall)store.ElementFactory.CreateElement(FunctionCall.DomainClassId);
            fcall.Name = data.Name;
            fcall.Description = data.Description;
            fcall.Guid = data.Guid;
            fcall.ExternalVisible = data.ExternalVisible;
            fcall.InternalRecordId = data.InternalRecordId;

            foreach (ParameterData fieldData in data.Parameters)
            {
                Field f = new Field(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                f.StringLength = fieldData.StringLength;
                fcall.Fields.Add(f);
            }


            // function return
            FunctionReturn fret = (FunctionReturn)store.ElementFactory.CreateElement(FunctionReturn.DomainClassId);
            fret.Name = data.Name;
            fret.Description = data.Description;
            fret.Guid = data.Guid;
            fret.ExternalVisible = data.ExternalVisible;
            fret.InternalRecordId = data.InternalRecordIdForReturn;

            foreach (ReturnValueData fieldData in data.ReturnValues)
            {
                OptionalbleFieldWithDefault f = new OptionalbleFieldWithDefault(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                f.Optional = fieldData.Optional;
                f.DefaultValue = fieldData.DefaultValue;
                f.StringLength = fieldData.StringLength;
                fret.Fields.Add(f);
            }


            fcall.FunctionReturn = fret;
            elementGroup.AddGraph(fcall, true);           
            elementGroup.AddGraph(fret, false);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, BenjaminSchroeter.Flow.Datastructures.Data.VariableData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);

            Variable v = (Variable)store.ElementFactory.CreateElement(Variable.DomainClassId);
            v.Name = data.Name;
            v.Description = data.Description;
            v.Guid = data.Guid;
            v.Type = data.Type != null ? data.Type.Clone() : null;
            v.ExternalWritable = data.ExternalWritable;
            v.ExternalVisible = data.ExternalVisible;
            v.FromHardwareDescription = false;
            v.Readable = v.Writable = true;
            elementGroup.AddGraph(v, true);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, RecordData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);

            Record r = (Record)store.ElementFactory.CreateElement(Record.DomainClassId);
            r.Name = data.Name;
            r.Description = data.Description;
            r.Guid = data.Guid;
            r.RecordId = data.RecordId;

            foreach (RecordFieldData fieldData in data.Fields)
            {
                OptionalbleFieldWithDefault f = new OptionalbleFieldWithDefault(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                f.Optional = fieldData.Optional;
                f.DefaultValue = fieldData.DefaultValue;
                f.StringLength = fieldData.StringLength;
                r.Fields.Add(f);
            }

            elementGroup.AddGraph(r, true);
            return elementGroup.CreatePrototype();
        }

        private static ElementGroupPrototype CreateElementToolPrototype(Store store, OutputEventData data)
        {
            ElementGroup elementGroup = new ElementGroup(store.DefaultPartition);

            OutputEvent o = (OutputEvent)store.ElementFactory.CreateElement(OutputEvent.DomainClassId);
            o.Name = data.Name;
            o.Description = data.Description;
            o.Guid = data.Guid;
            o.ExternalVisible = data.ExternalVisible;
            o.InternalRecordId = data.InternalRecordId;

            foreach (OutputEventFieldData fieldData in data.Fields)
            {
                OptionalbleFieldWithDefault f = new OptionalbleFieldWithDefault(store.DefaultPartition);
                f.Name = fieldData.Name;
                f.Guid = fieldData.Guid;
                f.Description = fieldData.Description;
                f.Type = fieldData.Type;
                f.Optional = fieldData.Optional;
                f.DefaultValue = fieldData.DefaultValue;
                f.StringLength = fieldData.StringLength;
                o.Fields.Add(f);
            }

            elementGroup.AddGraph(o, true);
            return elementGroup.CreatePrototype();
        }
        #endregion

        /// <summary>
        /// Gets the hierarchy item of this DocView.
        /// </summary>
        /// <returns></returns>
        private HierarchyItem GetHierarchyItem()
        {
            IVsRunningDocumentTable rdt = this.GetService(typeof(IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");

            HierarchyItem hierarchyItem = null;
            foreach (IVsRunningDocumentTableExtensions.RdtInfo info in rdt.GetDocDataForOpenDocuments())
                if (info.DocData == this.DocData)
                {
                    hierarchyItem = info.Item;
                    break;
                }

            return hierarchyItem;
        }

        private Dictionary<string, ModelingToolboxItem> GetVersionedModelingToolboxItems(string toolboxTabName)
        {
            Dictionary<string, ModelingToolboxItem> list = new Dictionary<string, ModelingToolboxItem>();
            foreach (ToolboxItem ti in toolboxService.GetToolboxItems(toolboxTabName))
                if (( (ti is VersionedModelingToolboxItem && (ti as VersionedModelingToolboxItem).Project == ToolboxItemKey)
                    || (ti.DisplayName == PLACEHOLDER_NAME) )
                     && !list.ContainsKey(ti.DisplayName))
                    list.Add(ti.DisplayName, (ModelingToolboxItem)ti);
            return list;
        }
        
        protected override ICollection TargetToolboxItemFilterAttributes
        {
            get
            {
                ICollection original = base.TargetToolboxItemFilterAttributes;

                List<ToolboxItemFilterAttribute> x = new List<ToolboxItemFilterAttribute>(original.Cast<ToolboxItemFilterAttribute>());
                if (this.ToolboxItemKey != "")
                    x.Add(new ToolboxItemFilterAttribute(this.ToolboxItemKey, System.ComponentModel.ToolboxItemFilterType.Require));
                return x;
            }
        }

        #region Dummy items
        private void AddDummyToolboxItem(string toolboxTabName,
            Dictionary<string, ModelingToolboxItem> oldItems)
        {
            if (oldItems.ContainsKey(PLACEHOLDER_NAME))
            {
                oldItems.Remove(PLACEHOLDER_NAME);
                return;
            }

            ModelingToolboxItem item = new ModelingToolboxItem(
                PLACEHOLDER_NAME,        // Unique identifier (non-localized) for the toolbox item.
                10000, // Position relative to other items in the same toolbox tab.
                PLACEHOLDER_NAME, // Localized display name for the item.
                Resources.InputSmall,   // Image displayed next to the toolbox item.
                toolboxTabName, // Unique identifier (non-localized) for the toolbox item tab.
                toolboxTabName,// Localized display name for the toolbox tab.
                "f1help", // F1 help keyword for the toolbox item.  
                "",// Localized tooltip text for the toolbox item.
                null, // ElementGroupPrototype (data object) representing model element on the toolbox.
                new[] { new ToolboxItemFilterAttribute("DUMMY", System.ComponentModel.ToolboxItemFilterType.Require) }); // filter

            toolboxService.AddToolboxItem(item, toolboxTabName);
        }

        /// <summary>
        /// Removes all Tabs from the Toolbox that conain only the dummy item
        /// </summary>
        private void RemoveEmptyDummyTabs()
        {
            List<string> toDelete = new List<string>();

            foreach (string tab in vsToolbox2.GetTabs())
            {
                IEnumToolboxItems pEnum;
                vsToolbox2.EnumItems(tab, out pEnum);
                uint pceltFetched;
                
                // get the first item 
                IDataObject[] rgelt1 = new IDataObject[1];
                pEnum.Next(1, rgelt1, out pceltFetched);   // .Next() seems not to work with a value != 1
                if (pceltFetched == 0)  // there is no first item -> check next tab
                    continue;

                // get the second item 
                IDataObject[] rgelt2 = new IDataObject[1];
                pEnum.Next(1, rgelt2, out pceltFetched);
                if (pceltFetched != 0) // there is a second item  -> check the nxt tab
                    continue;

                // is the first and single item a ToolboxItem?
                if (toolboxService.IsToolboxItem(rgelt1[0]))
                {
                    // and is it also the Dummy item?
                    ToolboxItem t = toolboxService.DeserializeToolboxItem(rgelt1[0]);
                    if (t.DisplayName == PLACEHOLDER_NAME)
                        toDelete.Add(tab);
                }
            }

            // now remove the tabs
            foreach (string tab in toDelete)
                vsToolbox2.RemoveTab(tab);
        }
        #endregion
    }
}

