using System;
using System.Linq;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling.Validation;

namespace BenjaminSchroeter.Flow.Datastructures
{
    [ValidationState(ValidationState.Enabled)]
    partial class Variable
    {
        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateType(ValidationContext context)
        {
            if ( this.Type == null || string.IsNullOrEmpty(this.Type.Value) )
            {
                string msg = string.Format("The Type of the Variable '{0}' must not be empty.", this.Name);
                context.LogError(msg, "Variable.ValidateType", this);
            }
            else
            {
                if ( this.DatastructureModel.DataTypes != null && 
                    !this.DatastructureModel.DataTypes.Contains(this.Type.Value ))
                {
                    string msg = string.Format("The Type ('{1}') of the Variable '{0}' is not defined in the Hardware Description.", this.Name, this.Type.Value);
                    context.LogError(msg, "Variable.ValidateType", this);
                }
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateExternalType(ValidationContext context)
        {
            // if type is null another rule will throw a message
            if ( this.Type == null )
                return;

            if ( !this.ExternalVisible && !this.ExternalWritable )
                return; // every type is allowed

            DslDataType dataType = this.Type.AsDslType(this.Store);
            if ( dataType == DslDataType.Record || dataType == DslDataType.Other )
            {
                string msg = string.Format("A Variable ('{0}') of Record or Other Data Type must not be external visible or external writable.", this.Name);
                context.LogError(msg, "Variable.ValidateExternalType", this);               
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateInitialValue(ValidationContext context)
        {
            // empty initialValue = no initialValue, or for string an empty string
            if (this.InitialValue.Trim() == "")
                return;

            // if type is null another rule will throw a message
            if ( this.Type == null )
                return;

            DslDataType dataType = this.Type.AsDslType(this.Store);
            switch (dataType)
            {
                case DslDataType.Boolean:
                    bool x;
                    if (!Boolean.TryParse(this.InitialValue, out x))
                    {
                        string msg = string.Format("The Initial Value ('{1}') of Variable '{0}' is not a boolean value. Please use 'True' or 'False'.", this.Name, this.InitialValue);
                        context.LogError(msg, "Variable.ValidateInitialValue", this);
                    }
                    break;

                case DslDataType.Number:
                    int i;
                    if (!Int32.TryParse(this.InitialValue, out i))
                    {
                        string msg = string.Format("The Initial Value ('{1}') of Variable '{0}' is not a numeric value.", this.Name, this.InitialValue);
                        context.LogError(msg, "Variable.ValidateInitialValue", this);
                    }
                    break;

                case DslDataType.String:
                    // every string can be parsed to string ;)
                    break;

                case DslDataType.Record:
                case DslDataType.Other:
                    string msg2 = string.Format("The Variable '{0}' cannot have an Initial Value, since of its Data Type.", this.Name);
                    context.LogError(msg2, "Variable.ValidateInitialValue", this);
                    break;

                default:
                    throw new ArgumentOutOfRangeException();
            }
        }

    }
}
