using System;
using System.Collections.Generic;
using System.Diagnostics;
using BenjaminSchroeter.Flow.Datastructures.Data;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.Flow.Datastructures.DataService
{
    /// <summary>
    /// DataService to receive the Datastructures defined in a given project.
    /// </summary>
    public class DatastructureDataService : IDatastructureDataService, SDatastructureDataService
    {
        private readonly IServiceProvider serviceProvider;

        /// <summary>
        /// Initializes a new instance of the <see cref="DatastructureDataService"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        internal DatastructureDataService(IServiceProvider serviceProvider)
        {
            this.serviceProvider = serviceProvider;
        }

        /// <summary>
        /// Gets all datastructures of all Datastructure models in the project of the hierarchy item.
        /// </summary>
        /// <param name="projectHierarchy">The project hierarchy item.</param>
        /// <returns></returns>
        public DatastructureModelData GetAllDatastructures(IVsHierarchy projectHierarchy)
        {
            IVsRunningDocumentTable rdt = (IVsRunningDocumentTable)serviceProvider.GetService(typeof(IVsRunningDocumentTable));
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");

            List<DatastructureModel> allModels = new List<DatastructureModel>();

            // === search the project hierarchy for all models ===
            foreach (HierarchyItem item in projectHierarchy.FindFileItems(file => file.EndsWith("." + Constants.DesignerFileExtension,StringComparison.InvariantCultureIgnoreCase)))
            {
                // === get the models from rdt... ===
                DatastructuresDocData docdata = rdt.GetDocDataForOpenDocument(item) as DatastructuresDocData;
                if (docdata != null && docdata.RootElement is DatastructureModel)
                    allModels.Add((DatastructureModel) docdata.RootElement);
                else
                {
                    // === ...or load them from disk  ===
                    Store store = new Store();
                    store.LoadDomainModels(new[] { typeof(CoreDesignSurfaceDomainModel), typeof(DatastructuresDomainModel) });

                    using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                    { 
                        DatastructureModel model = DatastructuresSerializationHelper.Instance.LoadModel(store, item.FullFileName, null, null);
                        t.Commit();

                        if ( model != null )
                            allModels.Add(model);
                    }
                }
            }            
            
            // === put all data together ===
            DatastructureModelData all = new DatastructureModelData();
            foreach (DatastructureModel model in allModels)
            {
                model.Functions.ForEach(f => all.Functions.Add(f.ToDataClass()));
                model.Records.ForEach(r => all.Records.Add(r.ToDataClass()));
                model.OutputEvents.ForEach(o => all.OutputEvents.Add(o.ToDataClass()));
                model.Variables.ForEach(v => all.Variables.Add(v.ToDataClass()));
            }

            return all;
        }
    }
}
