using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.DataTypes
{
    /// <summary>
    /// A TypeConverter for <see cref="DataTypeString"/> with a standard values list.
    /// The standard values list is shown as a drop down combo box in the properites window.
    /// </summary>
    public class DataTypeConverter : StringConverter 
    {
        /// <summary>
        /// Returns whether this object supports a standard set of values that can be picked from a list, using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <returns>
        /// true if <see cref="M:System.ComponentModel.TypeConverter.GetStandardValues"/> should be called to find a common set of values the object supports; otherwise, false.
        /// </returns>
        public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
        {
            return true;
        }

        /// <summary>
        /// Returns whether the collection of standard values returned from <see cref="M:System.ComponentModel.TypeConverter.GetStandardValues"/> is an exclusive list of possible values, using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <returns>
        /// true if the <see cref="T:System.ComponentModel.TypeConverter.StandardValuesCollection"/> returned from <see cref="M:System.ComponentModel.TypeConverter.GetStandardValues"/> is an exhaustive list of possible values; false if other values are possible.
        /// </returns>
        public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
        {
            return true;
        }

        /// <summary>
        /// Returns a collection of standard values for the data type this type converter is designed for when provided with a format context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context that can be used to extract additional information about the environment from which this converter is invoked. This parameter or properties of this parameter can be null.</param>
        /// <returns>
        /// A <see cref="T:System.ComponentModel.TypeConverter.StandardValuesCollection"/> that holds a standard set of valid values, or null if the data type does not support a standard set of values.
        /// </returns>
        public override StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
        {
            // ensure that the selected elements are all model elements
            //  or otherwise return an empty list
            object[] elements = context.Instance as object[];
            if ( elements!= null )
            {
                if (elements.Length == 0)
                    return  new StandardValuesCollection(new string[] {});

                if (! elements.All(x => x is ModelElement))
                    return  new StandardValuesCollection(new string[] {});

                // return the list for the selected elements
                return new StandardValuesCollection(GetRegisteredDataTypesAsStrings( (elements[0] as ModelElement).Store ));
            }

            // if no elements selectes return an empty list, too
            ModelElement element = context.Instance as ModelElement;
            if ( element == null )
                return  new StandardValuesCollection(new string[] {});

            // return the list for the selected elements
            return new StandardValuesCollection(GetRegisteredDataTypesAsStrings(element.Store));
        }

        #region convert from and to
        /// <summary>
        /// Gets a value indicating whether this converter can convert an object in the given source type to a string using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="sourceType">A <see cref="T:System.Type"/> that represents the type you wish to convert from.</param>
        /// <returns>
        /// true if this converter can perform the conversion; otherwise, false.
        /// </returns>
        public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
        {
            if (sourceType == typeof(string))
                return true;

            return base.CanConvertFrom(context, sourceType);
        }

        /// <summary>
        /// Converts the specified value object to a <see cref="T:System.String"/> object.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="culture">The <see cref="T:System.Globalization.CultureInfo"/> to use.</param>
        /// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
        /// <returns>
        /// An <see cref="T:System.Object"/> that represents the converted value.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException">The conversion could not be performed. </exception>
        public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
        {
            if (value.GetType() == typeof(string))
                return new DataTypeString(value.ToString().Trim());

            return base.ConvertFrom(context, culture, value);
        }

        /// <summary>
        /// Returns whether this converter can convert the object to the specified type, using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="destinationType">A <see cref="T:System.Type"/> that represents the type you want to convert to.</param>
        /// <returns>
        /// true if this converter can perform the conversion; otherwise, false.
        /// </returns>
        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType == typeof(string))
                return true;

            return base.CanConvertTo(context, destinationType);
        }

        /// <summary>
        /// Converts the given value object to the specified type, using the specified context and culture information.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="culture">A <see cref="T:System.Globalization.CultureInfo"/>. If null is passed, the current culture is assumed.</param>
        /// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
        /// <param name="destinationType">The <see cref="T:System.Type"/> to convert the <paramref name="value"/> parameter to.</param>
        /// <returns>
        /// An <see cref="T:System.Object"/> that represents the converted value.
        /// </returns>
        /// <exception cref="T:System.ArgumentNullException">The <paramref name="destinationType"/> parameter is null. </exception>
        /// <exception cref="T:System.NotSupportedException">The conversion cannot be performed. </exception>
        public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
        {
            if (destinationType == typeof(string) && value is DataTypeString)
                return (value as DataTypeString).Value;

            return base.ConvertTo(context, culture, value, destinationType);
        }
        #endregion

        #region datastructure
        private struct Data
        {
            public string Name;
            public string CName;
            public DslDataType DslName;
        }


        /// <summary>
        /// The Cache holds for every store (the guid of the store is used here)
        /// a list of DataTypes defiend in this store.
        /// 
        /// The Cache is static and every model registers its elements with the cache on loading.
        /// </summary>
        private static Dictionary<Guid, HashSet<Data>> cache = new Dictionary<Guid, HashSet<Data>>();

        /// <summary>
        /// Adds a data type to the cache.
        /// </summary>
        /// <param name="store">The store.</param>
        /// <param name="type">The type.</param>
        public static void AddDataType(Store store, string name, string cname, DslDataType type)
        {
            RemoveDataType(store,name);

            Data d = new Data {Name = name, CName = cname, DslName = type};
            GetSet(store).Add(d);
        }

        /// <summary>
        /// Removes a data type form the cache.
        /// </summary>
        /// <param name="store">The store.</param>
        /// <param name="name">The name.</param>
        public static void RemoveDataType(Store store, string name)
        {
            GetSet(store).RemoveWhere(x => x.Name == name);
        }

        /// <summary>
        /// Removes all data types of one store from the caceh.
        /// </summary>
        /// <param name="store">The store.</param>
        public static void RemoveDataTypes(Store store)
        {
            cache.Remove(store.Id);
        }

        public static string GetDataTypeCName(Store store, string name)
        {            
            return GetSet(store).FirstOrDefault(x => x.Name == name).CName ?? "";
        }
        public static DslDataType GetDataTypeDslType(Store store, string name)
        {
            HashSet<Data> set = GetSet(store);
            if (!set.Any(x => x.Name == name))
                return DslDataType.Other;
            
            return set.First(x => x.Name == name).DslName;
        }

        public static string GetDefaultDataType(Store store, DslDataType dslType)
        {
            HashSet<Data> set = GetSet(store);
            if (!set.Any(d => d.DslName == dslType))
                return "UNKNOWN";
            
            return set.First(d => d.DslName == dslType).Name;
        }

        private static HashSet<Data> GetRegisteredDataTypes(Store store)
        {
            if (!cache.ContainsKey(store.Id))
                return new HashSet<Data>();

            return cache[store.Id];
        }

        public static string[] GetRegisteredDataTypesAsStrings(Store store)
        {
            var a = from x in GetRegisteredDataTypes(store)
                    orderby x.Name
                    select x.Name;

            return a.ToArray();
        }

        private static HashSet<Data> GetSet(Store store)
        {
            if (store == null) return new HashSet<Data>();

            if (cache.ContainsKey(store.Id))
                return cache[store.Id];

            HashSet<Data> set = new HashSet<Data>();
            cache.Add(store.Id, set);
            return set;
        }
        #endregion
    }
}
