using System;
using System.ComponentModel;
using System.Text;
using System.Text.RegularExpressions;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Type Converter to edit the <see cref="DriverCompatibility"/> in the properties window.
    /// </summary>
    public class DriverCompatibilityTypeConverter : ExpandableObjectConverter
    {
        /// <summary>
        /// Returns whether this converter can convert the object to the specified type, using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="destinationType">A <see cref="T:System.Type"/> that represents the type you want to convert to.</param>
        /// <returns>
        /// true if this converter can perform the conversion; otherwise, false.
        /// </returns>
        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType == typeof(string))
                return true;

            return base.CanConvertTo(context, destinationType);
        }

        /// <summary>
        /// Converts the given value object to the specified type, using the specified context and culture information.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="culture">A <see cref="T:System.Globalization.CultureInfo"/>. If null is passed, the current culture is assumed.</param>
        /// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
        /// <param name="destinationType">The <see cref="T:System.Type"/> to convert the <paramref name="value"/> parameter to.</param>
        /// <returns>
        /// An <see cref="T:System.Object"/> that represents the converted value.
        /// </returns>
        /// <exception cref="T:System.ArgumentNullException">The <paramref name="destinationType"/> parameter is null. </exception>
        /// <exception cref="T:System.NotSupportedException">The conversion cannot be performed. </exception>
        public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
        {
            if (destinationType == typeof(string))
            {
                // this is more or less a workaround:
                // when a Object is shown in the PropertiyGrid there should be shown nothing in the root element
                // the TypeComverter to string is only used to serialize the data
                if (context != null && context.ToString().Contains("PropertyGrid"))
                    return "";

                return ConvertValueToString((DriverCompatibility)value);
            }

            return base.ConvertTo(context, culture, value, destinationType);
        }

        /// <summary>
        /// Returns whether this converter can convert an object of the given type to the type of this converter, using the specified context.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="sourceType">A <see cref="T:System.Type"/> that represents the type you want to convert from.</param>
        /// <returns>
        /// true if this converter can perform the conversion; otherwise, false.
        /// </returns>
        public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
        {
            if (sourceType == typeof(string))
                return true;

            return base.CanConvertFrom(context, sourceType);
        }

        /// <summary>
        /// Converts the given object to the type of this converter, using the specified context and culture information.
        /// </summary>
        /// <param name="context">An <see cref="T:System.ComponentModel.ITypeDescriptorContext"/> that provides a format context.</param>
        /// <param name="culture">The <see cref="T:System.Globalization.CultureInfo"/> to use as the current culture.</param>
        /// <param name="value">The <see cref="T:System.Object"/> to convert.</param>
        /// <returns>
        /// An <see cref="T:System.Object"/> that represents the converted value.
        /// </returns>
        /// <exception cref="T:System.NotSupportedException">The conversion cannot be performed. </exception>
        public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
        {
            if (value is string)
                return ConvertValueFromString((string) value);

            return base.ConvertFrom(context, culture, value);
        }

        /// <summary>
        /// Converts the value to string. This string represantation is used to write the data to the xml file.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns></returns>
        private static string ConvertValueToString(DriverCompatibility input)
        {
            StringBuilder sb = new StringBuilder();
            foreach (Guid guid in input.CompatibleHardware)
            {
                sb.Append("[");
                sb.Append(guid);
                sb.AppendLine("]");
            }

            foreach (DriverCompatibility.EnableableRelation relation in input.EnableableRelations)
            {
                sb.Append("(");
                sb.Append(relation.DriverElement);
                sb.Append(",");
                sb.Append(relation.HardwareElement);
                sb.AppendLine(")");
            }

            return sb.ToString();
        }

        /// <summary>
        /// Converts the value from string. This is used to read the class from an xml file.
        /// </summary>
        /// <param name="input">The input.</param>
        /// <returns></returns>
        private static DriverCompatibility ConvertValueFromString(string input)
        {
            Regex hardwareRegex = new Regex(@"\[([0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12})\]"); // any guid in []
            Regex enableableRelationRegex = new Regex(@"\(([0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}),([0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12})\)"); // any group like "(guid,guid)"

            DriverCompatibility x = new DriverCompatibility();
            foreach (Match o in hardwareRegex.Matches(input))
            {
                Guid g = new Guid(o.Groups[1].Value);
                x.CompatibleHardware.Add(g);
            }

            foreach (Match o in enableableRelationRegex.Matches(input))
            {
                Guid driverGuid = new Guid(o.Groups[1].Value);
                Guid hardwareGuid = new Guid(o.Groups[2].Value);

                x.EnableableRelations.Add( new DriverCompatibility.EnableableRelation() {DriverElement = driverGuid,HardwareElement = hardwareGuid});
            }

            return x;
        }
    }
}