using System;
using System.Collections.Generic;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Modeling.Validation;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Error List Observer for the Global Validation
    /// </summary>
    public class GlobalValidationErrorListObserver : ErrorListObserver
    {
        /// <summary>
        /// Gets or sets the filename that will be shown in the error list next to reported errors.
        /// </summary>
        /// <value>The filename.</value>
        public string Filename { get; set;}
        /// <summary>
        /// Gets or sets the project that will be shown in the error list next to reported errors.
        /// </summary>
        /// <value>The project.</value>
        public IVsHierarchy Project { get; set; }

        /// <summary>
        /// Initializes a new instance of the <see cref="GlobalValidationErrorListObserver"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public GlobalValidationErrorListObserver(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
            Clear();
        }

        /// <summary>
        /// Adds the message to the error lost.
        /// </summary>
        /// <remarks>As filename for this error the global property Filename will be used.</remarks>
        /// <param name="text">The text.</param>
        public void AddMessage(string text)
        {
            GlobalValidationMessage m = new GlobalValidationMessage(text);
            OnValidationMessageAdded(m);
        }

        /// <summary>
        /// Adds the message to the error list mapped to the given file.
        /// </summary>
        /// <param name="text">The text.</param>
        /// <param name="file">The file.</param>
        public void AddMessage(string text, string file)
        {
            GlobalValidationMessage m = new GlobalValidationMessage(text);
            OnValidationMessageAdded(m,file);
        }

        /// <summary>
        /// Called when validation message is added.
        /// </summary>
        protected override void OnValidationMessageAdded(ValidationMessage addedMessage)
        {
            OnValidationMessageAdded(addedMessage,Filename);
        }

        /// <summary>
        /// Called when validation message is added.
        /// </summary>
        private void OnValidationMessageAdded(ValidationMessage addedMessage, string file)
        {
            ValidationTask task = new ValidationTask(addedMessage);
            task.Document = file;
            task.HierarchyItem = Project;

            tasks.Add(task);
        }

        private List<ValidationTask> tasks;
        /// <summary>
        /// Gets the number or found errors.
        /// </summary>
        /// <value>The number or found errors.</value>
        public int Count
        {
            get
            {
                return tasks.Count;
            }
        }

        /// <summary>
        /// Clears the error list
        /// </summary>
        public void Clear()
        {
            this.tasks = new List<ValidationTask>();
        }

        /// <summary>
        /// Writes the found errors to the TaskProvider
        /// </summary>
        public void Write()
        {
            this.TaskProvider.SuspendRefresh();

            foreach (ValidationTask task in tasks)
            {
                this.TaskProvider.Tasks.Add(task);
            }

            this.TaskProvider.ResumeRefresh();
        }

    }

    internal class GlobalValidationMessage :  ValidationMessage
    {
        static readonly ValidationContext context = new ValidationContext(ValidationCategories.Custom,(ModelElement)null);

        protected internal GlobalValidationMessage(string description) : base(context, description)
        {
        }
    }
}