using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using BenjaminSchroeter.Flow;
using BenjaminSchroeter.Flow.UiSupportPackage;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Implementation of the Global Validation Service
    /// </summary>
    public class GlobalValidationService : IGlobalValidation, SGlobalValidation
    {
        private readonly IServiceProvider serviceProvider;
        private IVsRunningDocumentTable rdt;
        private static GlobalValidationErrorListObserver errorListObserver;

        /// <summary>
        /// Initializes a new instance of the <see cref="GlobalValidationService"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public GlobalValidationService(IServiceProvider serviceProvider)
        {
            this.serviceProvider = serviceProvider;

            if (errorListObserver == null)
                errorListObserver = new GlobalValidationErrorListObserver(serviceProvider);

            rdt = serviceProvider.GetService(typeof(IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");
        }

        /// <summary>
        /// Does the Validation.
        /// </summary>
        /// <returns>The number of found errors.</returns>
        public int Validate()
        {
            ClearErrorList();

            // get all needed services
            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");

            // get the selected project
            object[] projects = dte.ActiveSolutionProjects as object[];
            if (projects == null || projects.Length == 0)
                return 0;  // no project selected
            
            // if errors are found in the project structure, 
            // the details validation (that is the validation of
            // the single dsl files) will be skipped
            bool withDetailValidation = true;
            foreach (Project p in projects)
            {
                bool projectLevelErrorDetected;
                ValidateProject(p, withDetailValidation, out projectLevelErrorDetected);
                if (projectLevelErrorDetected)
                    withDetailValidation = false;
            }
            
            ShowErrorList();

            return errorListObserver.Count;
        }

        /// <summary>
        /// Validates a single project.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <param name="withDetailValidation">if set to <c>true</c> the DSL files of the project will be validated, too.</param>
        /// <param name="projectLevelErrorsDetected">returns true, if errors on the project level are found</param>
        private void ValidateProject(Project project, bool withDetailValidation, out bool projectLevelErrorsDetected)
        {
            // save all open files (only in the selected projects)
            if ( rdt != null)
                rdt.SaveAllDocumentsFromProject(project,true);

            // find all files from this project
            List<ProjectItem> items =  project.FindAllFiles();
            IEnumerable<string> hardwareInfo = GetProjectFiles(items, Const.FileExtensionHardwareInfo);
            // some special case: if there is no .hardwareinfo file, use the .hardware file instead
            if ( hardwareInfo.Count() == 0 )
                hardwareInfo = GetProjectFiles(items, Const.FileExtensionHardwareDescription);

            IEnumerable<string> datastructures = GetProjectFiles(items, Const.FileExtensionDatastructures);
            IEnumerable<string> dataflows = GetProjectFiles(items, Const.FileExtensionDataFlow);

            // get the hierarchy item from the dte project
            IVsHierarchy hierarchy = DteProjectToHierarchy(project);

            errorListObserver.Project = hierarchy;
            errorListObserver.Filename = "";

            // use the global validator to do the validation
            using (GlobalValidator v = new GlobalValidator(hierarchy, serviceProvider, errorListObserver))
            {
                v.LoadModels(hardwareInfo, datastructures, dataflows);
                v.Validate(withDetailValidation, out projectLevelErrorsDetected);
            }
        }

        private IEnumerable<string> GetProjectFiles(List<ProjectItem> items, string fileExtension)
        {
            return from i in items
                   where i.get_FileNames(0).EndsWith( fileExtension, StringComparison.InvariantCultureIgnoreCase) 
                   select i.get_FileNames(0);
        }

        /// <summary>
        /// Geths the hierarchy item for a Dte.Project instance.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <returns>An hierarchy item</returns>
        private IVsHierarchy DteProjectToHierarchy(Project project)
        {
            IVsSolution solution = serviceProvider.GetService(typeof(IVsSolution)) as IVsSolution;
            Debug.Assert(solution != null, "IVsSolution");

            IVsHierarchy pphier;
            solution.GetProjectOfUniqueName(project.UniqueName, out pphier);

            return pphier;
        }

        #region error list
        /// <summary>
        /// Clears the error list from all items corresponding to the given filename hand hierarchy item.
        /// </summary>
        /// <param name="filename">The filename.</param>
        /// <param name="hierarchy">The hierarchy item.</param>
        public void ClearErrorListItems(string filename, IVsHierarchy hierarchy)
        {
            // hm, better clear all items at this point, too ;-) 
            ClearErrorListItems();
            return;

            if ( errorListObserver == null )
                return;

            TaskProvider.TaskCollection tasks = errorListObserver.TaskProvider.Tasks;

            List<ValidationTask> toDelete = new List<ValidationTask>();
            foreach (object o in tasks)
            {
                ValidationTask t = o as ValidationTask;
                if ( t != null && t.Document.Equals(filename,StringComparison.InvariantCultureIgnoreCase) && t.HierarchyItem == hierarchy )
                    toDelete.Add(t);
            }

            toDelete.ForEach( tasks.Remove );
        }

        /// <summary>
        /// Clears the entire error list.
        /// </summary>
        public void ClearErrorListItems()
        {
            if (errorListObserver == null)
                return;

            TaskProvider.TaskCollection tasks = errorListObserver.TaskProvider.Tasks;
            tasks.Clear();
        }

        private void ClearErrorList()
        {
            errorListObserver.TaskProvider.Tasks.Clear();
            errorListObserver.TaskProvider.Refresh();
        }
        private void ShowErrorList()
        {
            errorListObserver.TaskProvider.Refresh();
            if (errorListObserver.TaskProvider.Tasks.Count > 0)
                errorListObserver.TaskProvider.Show();
        }
        #endregion
    }
}