using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.Datastructures;
using BenjaminSchroeter.Flow.DataTypes;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Modeling.Validation;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Validator to validate the elements of a project.
    /// </summary>
    public class GlobalValidator : IDisposable
    {
        private readonly IVsHierarchy project;
        private readonly string projectFileName;
        private readonly IServiceProvider serviceProvider;

        /// <summary>
        /// Gets or sets the hardware infos files.
        /// </summary>
        /// <value>The hardware infos files.</value>
        public IEnumerable<string> HardwareInfosFiles { get; private set; }
        /// <summary>
        /// Gets or sets the datastructures files.
        /// </summary>
        /// <value>The datastructures files.</value>
        public IEnumerable<string> DatastructuresFiles { get; private set; }
        /// <summary>
        /// Gets or sets the dataflows files.
        /// </summary>
        /// <value>The dataflows files.</value>
        public IEnumerable<string> DataflowsFiles { get; private set; }

        private Dictionary<string, HardwareDescriptionModel> HardwareInfos = new Dictionary<string, HardwareDescriptionModel>();
        private Dictionary<string, DatastructureModel> Datastructures = new Dictionary<string, DatastructureModel>();
        private Dictionary<string, DataFlowModel> Dataflows = new Dictionary<string, DataFlowModel>();
        private Dictionary<string, Store> stores = new Dictionary<string, Store>();
        
        private readonly GlobalValidationErrorListObserver errorListObserver;

        /// <summary>
        /// Initializes a new instance of the <see cref="GlobalValidator"/> class.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <param name="serviceProvider">The service provider.</param>
        /// <param name="errorListObserver">The error list observer.</param>
        public GlobalValidator(IVsHierarchy project, IServiceProvider serviceProvider, GlobalValidationErrorListObserver errorListObserver)
        {
            this.project = project;
            this.errorListObserver = errorListObserver;
            this.serviceProvider = serviceProvider;

            projectFileName = project.GetRoot().FullFileName;
        }

        /// <summary>
        /// Loads the models into this validator.
        /// </summary>
        /// <param name="hardwareInfos">The hardware infos.</param>
        /// <param name="datastructures">The datastructures.</param>
        /// <param name="dataflows">The dataflows.</param>
        public void LoadModels(IEnumerable<string> hardwareInfos, IEnumerable<string> datastructures, IEnumerable<string> dataflows)
        {
            this.HardwareInfosFiles = hardwareInfos;
            this.DatastructuresFiles = datastructures;
            this.DataflowsFiles = dataflows;

            #region hardware
            foreach (string file in this.HardwareInfosFiles)
            {
                if (!File.Exists(file))
                    continue;

                Store store = new Store();
                stores.Add(file, store);
                store.LoadDomainModels(new[] { typeof(CoreDesignSurfaceDomainModel), typeof(HardwareDescriptionDomainModel) });

                using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                {
                    HardwareDescriptionModel model = HardwareDescriptionSerializationHelper.Instance.LoadModel(store, file, null, null);
                    t.Commit();
                    this.HardwareInfos.Add(file, model);
                }
            }
            IEnumerable<DataType> datatypes = null;
            if (HardwareInfos.Count > 0)
            {
                HardwareDescriptionModel hardware = HardwareInfos.First().Value;
                datatypes = hardware.DataTypes;
            }
            #endregion

            #region datastruct
            foreach (string file in this.DatastructuresFiles)
            {
                if (!File.Exists(file))
                    continue;

                Store store = new Store();
                stores.Add(file, store);
                store.LoadDomainModels(new[] { typeof(CoreDesignSurfaceDomainModel), typeof(DatastructuresDomainModel) });

                using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                {
                    DatastructureModel model = DatastructuresSerializationHelper.Instance.LoadModel(store, file, null, null);
                    t.Commit();
                    this.Datastructures.Add(file, model);

                    if (datatypes != null)
                    {
                        foreach (DataType data in datatypes)
                            DataTypeConverter.AddDataType(store, data.Name, data.CType, data.DslType );
                        
                        model.DataTypes = datatypes.Select( x => x.Name ).ToArray();
                    }
                }
            }
            #endregion

            #region dataflow
            foreach (string file in this.DataflowsFiles)
            {
                if (!File.Exists(file))
                    continue;

                Store store = new Store();
                stores.Add(file, store);
                store.LoadDomainModels(new[] { typeof(CoreDesignSurfaceDomainModel), typeof(DataFlowDomainModel) });

                using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                {
                    DataFlowModel model = DataFlowSerializationHelper.Instance.LoadModel(store, file, null, null);
                    t.Commit();
                    this.Dataflows.Add(file, model);

                    if ( datatypes != null )
                    {
                        foreach (DataType data in datatypes)
                            DataTypeConverter.AddDataType(store, data.Name, data.CType, data.DslType);

                        model.DataTypes = datatypes.Select( x => x.Name ).ToArray();
                    }
                }
            }
            #endregion
        }

        #region validation
        /// <summary>
        /// Validates the loaded models.
        /// </summary>
        /// <param name="withDetailValidation">if set to <c>true</c> the DSL files of the project will be validated, too.</param>
        /// <param name="projectLevelErrorsDetected">returns true, if errors on the project level are found</param>
        public void Validate(bool withDetailValidation, out bool projectLevelErrorsDetected)
        {
            errorListObserver.Clear();

            // use every rule
            foreach (GlobalValidationRule rule in AllRules())
            {
                try
                {
                    rule.Validate(HardwareInfos, Datastructures, Dataflows);
                }
                catch
                {
                }
            }

            projectLevelErrorsDetected = errorListObserver.Count != 0;

            // validate the single files only if no errors were found and if withDetailValidation is set
            if (withDetailValidation && !projectLevelErrorsDetected)
                ValidateSingleFilesWithDefaultValidation();

            // write errors to the errorlist
            errorListObserver.Write();
        }

        /// <summary>
        /// Find all rules in this assembly.
        /// </summary>
        /// <returns></returns>
        private IEnumerable<GlobalValidationRule> AllRules()
        {
            // find all types that are non abstrac subclasses of GlobalValidtion Rule
            var rules = this.GetType().Assembly.GetTypes().Where(t => t.IsSubclassOf(typeof (GlobalValidationRule)) && !t.IsAbstract);
            foreach (Type rule in rules)
            {
                // create rule instance
                ConstructorInfo constructor = rule.GetConstructor(new[] {typeof (string), typeof (GlobalValidationErrorListObserver)});
                GlobalValidationRule x = constructor.Invoke(new object[] {this.projectFileName, this.errorListObserver}) as GlobalValidationRule;

                yield return x;
            }
        }

        /// <summary>
        /// Validates the single files with default validation.
        /// </summary>
        public void ValidateSingleFilesWithDefaultValidation()
        {
            IVsRunningDocumentTable rdt = this.serviceProvider.GetService(typeof(IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt!=null,"rdt");

            VsValidationController vc = new VsValidationController(serviceProvider);
            vc.AddObserver(errorListObserver);

            // validate every loaded store 
            foreach (KeyValuePair<string, Store> pair in this.stores)
            {
                // if the model is open in visual studio: clear the error list
                ModelingDocData docdata = rdt.GetDocDataForOpenDocument(pair.Key) as ModelingDocData;
                if (docdata != null)
                    docdata.ClearErrorListItems();

                // now do the validation
                errorListObserver.Filename = pair.Key;
                vc.Validate(pair.Value, ValidationCategories.Open);
            }
        }
        #endregion

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            // Dispose all stores
            foreach (KeyValuePair<string, Store> pair in stores)
            {
                pair.Value.Dispose();
                DataTypeConverter.RemoveDataTypes(pair.Value);
            }

            this.stores = null;
            this.HardwareInfos = null;
            this.Datastructures = null;
            this.Dataflows = null;

        }
    }
}
