using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.Datastructures;
using BenjaminSchroeter.Flow.HardwareDescription;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Global Validation Rule to ensure that each event used in one ore more dataflows is not used more often (in more different dataflows) than allows.
    /// </summary>
    class DataFlowEventCount : GlobalValidationRule
    {
        public DataFlowEventCount(string projectFileName, GlobalValidationErrorListObserver errorListObserver) : base(projectFileName, errorListObserver)
        {
        }

        public override void Validate(Dictionary<string, HardwareDescriptionModel> hardwareInfos, Dictionary<string, DatastructureModel> datastructures, Dictionary<string, DataFlowModel> dataflows)
        {
            if ( hardwareInfos.Count != 1)
                return;

            // find all events in the dataflows and order them by there guid
            // for each guid of a model a list of filenames will be created
            Dictionary<Guid, List<string>> events = new Dictionary<Guid, List<string>>();
            foreach (KeyValuePair<string, DataFlowModel> dataflow in dataflows)
            {
                var allEvents = dataflow.Value.TriggerFlow.FlowElements.Where(x => x is InputEvent);

                foreach (InputEvent e in allEvents)
                {
                    if ( !events.ContainsKey(e.Guid))
                        events.Add(e.Guid, new List<string>());

                    events[e.Guid].Add(dataflow.Key); // the key is the filename of the dataflow
                }
            }

            // check every event in the list (that is every event used in at least one dataflow)
            foreach (KeyValuePair<Guid, List<string>> pair in events)
            {
                Guid eventGuid = pair.Key;
                
                // find this event in the hardwareinfo
                int maxSubscribers = 1;
                string eventName = "";
                foreach (Platform platform in hardwareInfos.First().Value.Platforms)
                {
                    Event hardwareevent = platform.Events.FirstOrDefault(e => e.Guid.Equals(eventGuid));
                    if ( hardwareevent != null)
                    {
                        maxSubscribers = hardwareevent.Multiplicity;
                        eventName = hardwareevent.Name;
                        break;
                    }
                }

                if ( maxSubscribers == 1 )
                    continue; // with a proxy method in the main.c more than one are allowed

                if ( pair.Value.Count() > maxSubscribers )
                {
                    string allModels = string.Join(", ",pair.Value.Select(x => Path.GetFileName(x)).ToArray());
                    LogProjectError("{0} Models are subscribing the Event '{1}' but only {2} are allowed: {3}.",pair.Value.Count(),eventName,maxSubscribers,allModels);
                }
            }
        }
    }
}
