using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.Datastructures;
using BenjaminSchroeter.Flow.HardwareDescription;
using OutputEvent=BenjaminSchroeter.Flow.Datastructures.OutputEvent;
using Record=BenjaminSchroeter.Flow.Datastructures.Record;
using Variable=BenjaminSchroeter.Flow.Datastructures.Variable;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Global Validation Rule to ensure that in all Datastructures the name of the elements and the RecordIds are unique.
    /// </summary>
    class DatastructuresUniqueness : GlobalValidationRule
    {
        private struct ElementInFile
        {
            public ElementInFile(string fileName, string elementName)
            {
                FileName = fileName;
                ElementName = elementName;
            }

            public string FileName;
            public string ElementName;
        }

        // for each type (Varable, Function, OutputEvent and Record)
        //    I maintain a dictionary with the name of these element
        //        and all files that contain such a named element
        private Dictionary<Type, Dictionary<string, List<string>>> names;

        // for each record id
        //     i maintain a list of files that contain such a record and the name of this record
        private Dictionary<int, List<ElementInFile>> recordIds;

        public DatastructuresUniqueness(string projectFileName, GlobalValidationErrorListObserver errorListObserver) : base(projectFileName, errorListObserver)
        {
        }

        public override void Validate(Dictionary<string, HardwareDescriptionModel> hardwareInfos, Dictionary<string, DatastructureModel> datastructures, Dictionary<string, DataFlowModel> dataflows)
        {
            BuildData(datastructures);

            // look for duplicate names
            foreach (KeyValuePair<Type, Dictionary<string, List<string>>> namePair in names)
            {
                Type elementType = namePair.Key;
                string elementTypeName = elementType.Name;

                foreach (KeyValuePair<string, List<string>> pair in namePair.Value)
                {
                    string elementName = pair.Key;
                    List<string> fileNames = pair.Value;

                    if ( fileNames.Count > 1 )
                    {
                        string allfilesnames = string.Join(", ", fileNames.ToArray());
                        LogProjectError("More than one Datastructure Model defines a {0} named '{1}': {2}.", elementTypeName, elementName, allfilesnames);
                    }
                }
            }

            // look for duplicate record ids
            foreach (KeyValuePair<int, List<ElementInFile>> pair in recordIds)
            {
                int id = pair.Key;
                if ( pair.Value.Count > 1 )
                {
                    string allfilesnames = string.Join(", ", pair.Value.Select(a => string.Format("'{0}' in {1}", a.ElementName, Path.GetFileName(a.FileName))).ToArray());
                    LogProjectError("More than one Datastructure Model defines an Element with the RecordId {0}: {1}.", id, allfilesnames);
                }
            }
        }

        /// <summary>
        /// Builds the datastructures described above: names and recordIds.
        /// </summary>
        /// <param name="datastructures">The datastructures.</param>
        private void BuildData(Dictionary<string, DatastructureModel> datastructures)
        {
            names = new Dictionary<Type, Dictionary<string, List<string>>>();
            this.names.Add(typeof(Variable), new Dictionary<string, List<string>>());
            this.names.Add(typeof(Function), new Dictionary<string, List<string>>());
            this.names.Add(typeof(Record), new Dictionary<string, List<string>>());
            this.names.Add(typeof(OutputEvent), new Dictionary<string, List<string>>());

            recordIds = new Dictionary<int, List<ElementInFile>>();

            // iterate over all Datastructure Models
            foreach (KeyValuePair<string, DatastructureModel> datastructure in datastructures)
            {
                string filename = datastructure.Key;
                // find all elements with a name (Variable, Function, OutputEvent)
                foreach (ExternalVisibleDomainClass element in datastructure.Value.Partition.ElementDirectory.AllElements.Where(x => x is ExternalVisibleDomainClass))
                {
                    Type type = element.GetType();
                    string name = element.Name;

                    // add Type, name and Filename to the Datastructure
                    Dictionary<string, List<string>> nameDictionary = this.names[type];
                    if ( !nameDictionary.ContainsKey(name) )
                        nameDictionary.Add(name, new List<string>());
                    nameDictionary[name].Add(filename);  // add filename

                    // add recordid and Filename to the Datastructure
                    int id = element.InternalRecordId;
                    if (!this.recordIds.ContainsKey(id))
                        this.recordIds.Add(id, new List<ElementInFile>());
                    this.recordIds[id].Add(new ElementInFile(filename, name));
                    Function f = element as Function;
                    if (f != null)
                    {
                        id = f.InternalRecordIdForReturn;
                        if (!this.recordIds.ContainsKey(id))
                            this.recordIds.Add(id, new List<ElementInFile>());
                        this.recordIds[id].Add(new ElementInFile(filename, name));
                    }
                }

                // find all records
                foreach (Record element in datastructure.Value.Partition.ElementDirectory.AllElements.Where(x => x is Record))
                {
                    // name 
                    Type type = element.GetType();
                    string name = element.Name;

                    // add Type, name and Filename to the Datastructure
                    Dictionary<string, List<string>> nameDictionary = this.names[type];
                    if (!nameDictionary.ContainsKey(name))
                        nameDictionary.Add(name, new List<string>());
                    nameDictionary[name].Add(filename);  // add filename

                    // add recordid and Filename to the Datastructure
                    int id = element.RecordId;
                    if ( !this.recordIds.ContainsKey(id))
                        this.recordIds.Add(id, new List<ElementInFile>());
                    this.recordIds[id].Add(new ElementInFile(filename,name));
                }
            }
        }
    }
}
