using System.Collections.Generic;
using System.IO;
using System.Linq;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.Flow.DataFlow.Update;
using BenjaminSchroeter.Flow.Datastructures;
using BenjaminSchroeter.Flow.Datastructures.Data;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.GlobalValidation
{
    /// <summary>
    /// Global Validation Rule to ensure that no dataflow has base changes (that are changes 
    /// to the Hardwaredescription or datastructureas) that are not integrated to the dataflow model
    /// </summary>
    class IntegrateBaseChanges : GlobalValidationRule
    {
        public IntegrateBaseChanges(string projectFileName, GlobalValidationErrorListObserver errorListObserver) : base(projectFileName, errorListObserver)
        {
        }

        public override void Validate(Dictionary<string, HardwareDescriptionModel> hardwareInfos, Dictionary<string, DatastructureModel> datastructures, Dictionary<string, DataFlowModel> dataflows)
        {
            if ( hardwareInfos.Count != 1 )
                return;
            
            // get the hardware
            HardwareDescriptionModelData hardware = hardwareInfos.First().Value.ToDataClass();

            // put all datastructurtes together
            DatastructureModelData datastructs = new DatastructureModelData();
            foreach (DatastructureModel model in datastructures.Values)
            {
                model.Functions.ForEach(f => datastructs.Functions.Add(f.ToDataClass()));
                model.Records.ForEach(r => datastructs.Records.Add(r.ToDataClass()));
                model.OutputEvents.ForEach(o => datastructs.OutputEvents.Add(o.ToDataClass()));
                model.Variables.ForEach(v => datastructs.Variables.Add(v.ToDataClass()));
            }

            // do the updates (for base changes)
            foreach (KeyValuePair<string, DataFlowModel> pair in dataflows)
            {
                DataFlowModel dataflow = pair.Value;
                using(Transaction t = dataflow.Store.TransactionManager.BeginTransaction("Integrate Base Changes in Global Validation"))
                {
                    // update the dataflow for changed hardware
                    HardwareUpdater hardwareUpdater = new HardwareUpdater();
                    hardwareUpdater.Update(dataflow, hardware);

                    // update the dataflow for changed datastructures
                    DatastructuresUpdater datastructuresUpdater = new DatastructuresUpdater();
                    datastructuresUpdater.Update(dataflow, datastructs);

                    if (t.HasPendingChanges)
                        // there were some changes
                        LogProjectError("Changes detected in the Hardwaredescription or Datastructures used by '{0}'. Please open the file, review these changes and save it again.", Path.GetFileName(pair.Key));

                    t.Rollback(); // no need to save it
                }
            }
        }
    }
}