using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

namespace BenjaminSchroeter.Flow.Firmwareparser
{
    [Flags]
    public enum DoxygenOutput
    {
        None  =  0,
        Html  =  1,    // doxyfile property GENERATE_HTML
        LaTex =  2,    // doxyfile property GENERATE_LATEX
        Rtf   =  4,    // doxyfile property GENERATE_RTF
        Man   =  8,    // doxyfile property GENERATE_MAN
        Xml   = 16     // doxyfile property GENERATE_XML
    }

    /// <summary>
    /// Managed wrapper to call the Doxygen executable
    /// </summary>
    /// <remarks>
    /// You can configure the options to pass to the doxygen.exe via the public properties.
    /// A new configuration for doxygen will be created in memory and passed to the
    /// doxygen process.
    /// After that, you get the doxygen output as a string.
    /// 
    /// For more about doxygen see: http://www.doxygen.org
    /// </remarks>
    public class DoxygenWrapper
    {
        /// <summary>
        /// Gets or sets the full path to the doxygen executable.
        /// </summary>
        /// <value>The doxygen executable.</value>
        public string DoxygenExecutable { get; set; }

        #region properties mapping doxygen options (see doxygen doku)
        public string ProjectName { get; set; }         // doxyfile property PROJECT_NAME
        public string ProjectNumber { get; set; }       // doxyfile property PROJECT_NUMBER
        public string OutputDirectory { get; set; }     // doxyfile property OUTPUT_DIRECTORY 
        public IList<string> Aliases { get; private set; }      // doxyfile property ALIASES 
        public bool HideUndocMembers { get; set; }      // doxyfile property HIDE_UNDOC_MEMBERS
        public bool HideUndocClasses { get; set; }      // doxyfile property HIDE_UNDOC_CLASSES
        public IList<string> Input { get; private set; }        // doxyfile property INPUT
        public string FilePatterns { get; set; }        // doxyfile property FILE_PATTERNS 
        public DoxygenOutput Generate { get; set; }     // doxyfile property GENERATE_*
        public string XmlOutput { get; set; }           // doxyfile property XML_OUTPUT
        public bool XmlProgrammlisting { get; set; }    // doxyfile property XML_PROGRAMLISTING
        public string InputEncoding { get; set;}        // doxyfile property INPUT_ENCODING
        #endregion

        private string doxyfile;

        /// <summary>
        /// Initializes a new instance of the <see cref="DoxygenWrapper"/> class.
        /// </summary>
        public DoxygenWrapper()
        {
            this.Aliases = new List<string>();
            this.Input = new List<string>();

            this.InputEncoding = "";
        }

        /// <summary>
        /// Runs doxygen an returns its output. Before calling this mehtod make sure
        /// all properties are set proper.
        /// </summary>
        /// <param name="status">The status output from doxygen.</param>
        /// <returns>The parsing output from doxygen.</returns>
        public string RunDoxygen(out string status)
        {
            string output;

            GenerateTemplate();
            ManipulateTemplate();
            GenerateDocumentation(out status, out output);

            return output;
        }

        /// <summary>
        /// Generates an empty template for doxygen
        /// </summary>
        private void GenerateTemplate()
        {
            ExecuteDoxygen("-s -g -", null, true, out this.doxyfile);
        }

        /// <summary>
        /// This method writes all the properties set to this instance to the doxygen configuration file.
        /// </summary>
        /// <remarks>
        /// I have to change many of the template values, but i can leave the old values in the file.
        /// From the Doxygen Documentation "If the same tag is assigned more than once, the last assignment overwrites any earlier assignment."
        /// I have to add my properties at the end of the file!
        /// </remarks>
        private void ManipulateTemplate()
        {
            StringBuilder sb = new StringBuilder(this.doxyfile);

            sb.AppendLine();
            sb.AppendLine("#---------------------------------------------------------------------------");
            sb.AppendLine("# Values from DoxygenWrapper.net overwriting the default values above");
            sb.AppendLine("#---------------------------------------------------------------------------");

            sb.AppendLine("PROJECT_NAME = \"" + this.ProjectName + "\"");
            sb.AppendLine("PROJECT_NUMBER = \"" + this.ProjectNumber + "\"");
            sb.AppendLine("OUTPUT_DIRECTORY = \"" + this.OutputDirectory + "\"");
            foreach (string alias in this.Aliases)
                sb.AppendLine("ALIASES += \"" + alias + "\"");
            sb.AppendLine("HIDE_UNDOC_MEMBERS = " + (this.HideUndocMembers ? "YES" : "NO"));
            sb.AppendLine("HIDE_UNDOC_CLASSES = " + (this.HideUndocClasses ? "YES" : "NO"));
            foreach (string oneInput in this.Input)
                sb.AppendLine("INPUT += \"" + oneInput + "\"");
            sb.AppendLine("FILE_PATTERNS = \"" + this.FilePatterns + "\"");
            sb.AppendLine("XML_OUTPUT = \"" + this.XmlOutput + "\"");
            sb.AppendLine("XML_PROGRAMLISTING = " + (this.XmlProgrammlisting ? "YES" : "NO"));
            if ( this.InputEncoding.Length > 0 )
                sb.AppendLine("INPUT_ENCODING = \"" + this.InputEncoding + "\"");

            sb.AppendLine("GENERATE_HTML = " + (((this.Generate & DoxygenOutput.Html) == DoxygenOutput.Html) ? "YES" : "NO"));
            sb.AppendLine("GENERATE_LATEX = " + (((this.Generate & DoxygenOutput.LaTex) == DoxygenOutput.LaTex) ? "YES" : "NO"));
            sb.AppendLine("GENERATE_RTF = " + (((this.Generate & DoxygenOutput.Rtf) == DoxygenOutput.Rtf) ? "YES" : "NO"));
            sb.AppendLine("GENERATE_MAN = " + (((this.Generate & DoxygenOutput.Man) == DoxygenOutput.Man) ? "YES" : "NO"));
            sb.AppendLine("GENERATE_XML = " + (((this.Generate & DoxygenOutput.Xml) == DoxygenOutput.Xml) ? "YES" : "NO"));

            this.doxyfile = sb.ToString();
        }

        /// <summary>
        /// Generates the documentation output using doxygen.
        /// </summary>
        /// <param name="output">The parsing output.</param>
        /// <param name="err">The error outout of doxygen.</param>
        private void GenerateDocumentation(out string output, out string err)
        {
            ExecuteDoxygen("-", this.doxyfile, false, out output, out err);
        }

        /// <summary>
        /// Executes the doxygen.exe
        /// </summary>
        private void ExecuteDoxygen(string doxygenParameter, string stdInput, bool throwException, out string output)
        {
            string err;
            ExecuteDoxygen(doxygenParameter,stdInput,throwException,out output, out err);
        }
        /// <summary>
        /// Executes the doxygen.exe
        /// </summary>
        private void ExecuteDoxygen(string doxygenParameter, string stdInput, bool throwException, out string output, out string err)
        {
            ProcessStartInfo psi = new ProcessStartInfo(this.DoxygenExecutable, doxygenParameter);
            psi.UseShellExecute = false;
            psi.RedirectStandardOutput = true;
            psi.RedirectStandardError = true;
            psi.RedirectStandardInput = stdInput != null;
            psi.CreateNoWindow = true;

            Process p = new Process();            
            p.StartInfo = psi;

            p.Start(); 
            // redirect input (if input is given)
            if (stdInput != null)
            {
                p.StandardInput.WriteLine(stdInput);
                p.StandardInput.Close();
            }
            
            // redirect output
            StringBuilder sbOutput = new StringBuilder();
            p.OutputDataReceived += delegate(object sender, DataReceivedEventArgs e) { sbOutput.AppendLine(e.Data); };
            p.BeginOutputReadLine();

            // redirect error
            StringBuilder sbErr = new StringBuilder();
            p.ErrorDataReceived += delegate(object sender, DataReceivedEventArgs e) { sbErr.AppendLine(e.Data); };
            p.BeginErrorReadLine();

            // wait
            p.WaitForExit();

            // check for error
            err = sbErr.ToString().Trim();
            if (throwException && err.Length > 0)
                throw new DoxygenException(err);

            output = sbOutput.ToString().Trim();
        }

        /// <summary>
        /// This static function creates a template to use with the <see cref="Aliases"/> property to generate an XML output from doxygen with a parameter
        /// </summary>
        public static string XmlAliasWithParameter(string name)
        {
            return String.Format(@"{0}{{1}}=\xmlonly<{0}>\1</{0}>\endxmlonly",name);
        }

        /// <summary>
        /// This static function creates a template to use with the <see cref="Aliases"/> property to generate an XML output from doxygen without a parameter
        /// </summary>
        public static string XmlAliasWithoutParameter(string name)
        {
            return String.Format(@"{0}=\xmlonly<{0} />\endxmlonly", name);
        }
    }
}
