using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Linq;
using System.Xml.Schema;
using System.Xml.Serialization;
using System.Xml.XPath;
using System.Xml.Xsl;
using BenjaminSchroeter.Flow.Firmwareparser.SemanticChecks;

namespace BenjaminSchroeter.Flow.Firmwareparser
{
    /// <summary>
    /// A parser to analyze the source files of a firmware annotated
    /// with the Flow specific xml tags.
    /// </summary>
    public class FirmwareParser : IDisposable
    {
        private readonly DoxygenWrapper doxygen;
        private readonly IList<ParsingMessage> messages = new List<ParsingMessage>();
        private Stream result;

        /// <summary>
        /// Initializes a new instance of the <see cref="FirmwareParser"/> class.
        /// </summary>
        public FirmwareParser()
        {
            doxygen = new DoxygenWrapper();
            doxygen.InputEncoding = "ISO-8859-1";
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            result.Close();
            result.Dispose();
        }

        #region properties
        /// <summary>
        /// Gets or sets the directory used for temporary files.
        /// </summary>
        /// <remarks>If left empty, a new directory below the systems temp path will be used</remarks>
        public string OutputDirectory
        {
            get
            {
                return doxygen.OutputDirectory;
            }
            set
            {
                doxygen.OutputDirectory = value;
            }
        }

        /// <summary>
        /// Gets or sets the full path to the doxygen executable.
        /// </summary>
        public string DoxygenExecutable
        {
            get
            {
                return doxygen.DoxygenExecutable;
            }
            set
            {
                doxygen.DoxygenExecutable = value;
            }
        }

        /// <summary>
        /// Gets the messages generated by the firmwareparser.
        /// </summary>
        /// <value>The messages.</value>
        public IList<ParsingMessage> Messages
        {
            get { return messages; }
        }
        #endregion

        /// <summary>
        /// Parses the specified files.
        /// </summary>
        /// <param name="files">The files of the FLOW firmware.</param>
        public void Parse(string[] files)
        {
            // check the given output directory
            // or create one if the output directory property is empty
            DirectoryInfo diTemp = null;
            if (string.IsNullOrEmpty(doxygen.OutputDirectory))
            {
                diTemp = new DirectoryInfo(Path.GetTempPath());
                if (!diTemp.Exists)
                    throw new Exception("Temp Path does not exist");
                diTemp = diTemp.CreateSubdirectory(Path.GetRandomFileName());
                doxygen.OutputDirectory = diTemp.FullName;
            }

            // add the files to doxygen
            foreach (string file in files)
                doxygen.Input.Add(file);

            // configure the wrapper to parse the special tags
            ConfigureDoxygen();

            // execute doxygen now
            string status;
            string msgs = doxygen.RunDoxygen(out status);

            ProcessOutputMessages(msgs);

            // check the output for existence
            DirectoryInfo di = new DirectoryInfo(doxygen.OutputDirectory);
            if (!di.Exists)
                throw new DirectoryNotFoundException("The doxygen output directory (" + di.FullName + ") cannot be found.");
            FileInfo[] fis = di.GetFiles("index.xml");
            if ( fis.Length == 0 )
                throw new FileNotFoundException("The doxygen output index.xml in directory " + di.FullName + " cannot be found.");

            // create the needed xml output from the doxygen generated output
            CreateXml(fis[0]);

            // do some validation
            CheckXmlSemantic();

            // clean up
            if (diTemp != null)
            {
                diTemp.Delete(true);
                doxygen.OutputDirectory = "";
            }
        }

        /// <summary>
        /// Creates the needed xml output from the doxygen generated output 
        /// </summary>
        /// <param name="fi">The index.xml file created by doxygen.</param>
        private void CreateXml(FileInfo fi)
        {
            // open the file
            XPathDocument xmlInput = new XPathDocument(fi.FullName);

            // configure a xslt transformation
            StringReader xsltText = new StringReader(Resource.Compound2DslInfos);
            XmlReader xsltReader = new XmlTextReader(xsltText);
            XslTransform xslt = new XslTransform();
            xslt.Load(xsltReader);
            xsltReader.Close();
            xsltText.Close();

            // transform the xml input to an in-memory-output
            result = new MemoryStream();
            XmlTextWriter writer = new XmlTextWriter(result, null);
            
            XsltArgumentList args = new XsltArgumentList();
            string directory = fi.Directory.FullName;
            if (!directory.EndsWith("\\"))
                directory += "\\";
            args.AddParam("directory", "", directory);

            xslt.Transform(xmlInput, args, writer);

            // writes the xslt output to the global stream 'result'
            writer.Flush();
        }


        /// <summary>
        /// Checks the created XML output for some semantic errors defined by SemanticCheck rules
        /// </summary>
        private void CheckXmlSemantic()
        {
            // gets the xml as a CodeDescription DOM
            CodeDescription doc = GetCodeDescription(false);

            // find all rules via reflection
            foreach (Type type in this.GetType().Assembly.GetTypes())
                if (!type.IsAbstract && type.IsSubclassOf(typeof(SemanticCheckBase)))
                {
                    // create and execute each rule
                    SemanticCheckBase chk = (SemanticCheckBase) type.GetConstructor(new[] { typeof(IList<ParsingMessage>) }).Invoke(new[] { messages });
                    chk.Check(doc);  
                }
        }

        /// <summary>
        /// Processes the output messages from doxygen and searchs for error messages
        /// </summary>
        /// <param name="msgs">The output created by doxygen.</param>
        private void ProcessOutputMessages(string msgs)
        {
            Regex regex = new Regex("^(?<file>.+):(?<line>[0-9]+):(?<type>[^:]+):(?<text>.+)$");
            ParsingMessage lastMsg = null;
            foreach (string msg in msgs.Split('\n'))
            {
                if (msg.Trim() == "")
                    continue;
                Match m = regex.Match(msg);         
                if (m.Success)
                {
                    lastMsg = new ParsingMessage();
                    lastMsg.File = m.Groups["file"].Value.Trim().Replace('/','\\');
                    lastMsg.Line = int.Parse(m.Groups["line"].Value);
                    lastMsg.Priority = ParsingMessagePriority.Information;
                    string type = m.Groups["type"].Value.Trim();
                    string text = m.Groups["text"].Value.Trim();
                    if (type != "Warning")
                        text = type + ": " + text;
                    lastMsg.Text = text;

                    this.messages.Add(lastMsg);
                }
                else
                {
                    // lines that does not match with the regex.                    
                    if ( lastMsg == null )
                    {
                        // create a new message for them
                        lastMsg = new ParsingMessage();
                        lastMsg.Text = msg.Trim();
                        lastMsg.Priority = ParsingMessagePriority.Information;
                        this.messages.Add(lastMsg);
                    }
                    else
                        // or if there is a message before: append the text
                        lastMsg.Text += " " + msg.Trim();
                }
            }
        }

        /// <summary>
        /// Configures the doxygen wrapper to parse the special tags i had defined
        /// </summary>
        private void ConfigureDoxygen()
        {
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("default"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("description"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("event"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("eventCallback"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("field"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("fieldenabled"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("getVariable"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("input"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("inputInit"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("output"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("outputInit"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("parameter"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("setVariable"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("enabled"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("multiplicity"));

            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("firmwarename"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("drivername"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("productname"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("vendorname"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("producturl"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("vendorurl"));
            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithParameter("version"));

            doxygen.Aliases.Add(DoxygenWrapper.XmlAliasWithoutParameter("enabled"));

            doxygen.Generate = DoxygenOutput.Xml;
            doxygen.HideUndocClasses = true;
            doxygen.HideUndocMembers = true;

            doxygen.ProjectName = "firmware2xml";
            doxygen.XmlProgrammlisting = true;
            doxygen.XmlOutput = ".";
        }

        /// <summary>
        /// Gets the output of the parser as a CodeDescription object.
        /// </summary>
        public CodeDescription GetCodeDescription()
        {
            return GetCodeDescription(true);
        }

        /// <summary>
        /// Gets the output of the parser as a CodeDescription object.
        /// </summary>
        /// <param name="fullClean">if true, all elements without any annotation will be removed from the output</param>
        private CodeDescription GetCodeDescription(bool fullClean)
        {
            XmlReader validationReader = GetValidationReader();

            XmlSerializer serializer = new XmlSerializer(typeof(CodeDescription));
            CodeDescription code = (CodeDescription)serializer.Deserialize(validationReader);
            validationReader.Close();

            if (fullClean)
            {
                // remove all elements without any annotation
                var cleanFunctions = from f in code.function
                                     where f.ItemElementName != ItemChoiceType.none
                                     select f;
                code.function = cleanFunctions.ToArray();
            }
            else
            {
                // remove all elements without any annotation
                var cleanFunctions = from f in code.function
                                     where
                                         !(f.description == null && f.field == null && f.Item == null &&
                                           !f.multiplicitySpecified && f.ItemElementName == ItemChoiceType.none)
                                     select f;
                code.function = cleanFunctions.ToArray();
            }
            return code;
        }

        /// <summary>
        /// Gets the parser output as a XDocument.
        /// </summary>
        /// <returns></returns>
        public XDocument GetXDocument()
        {
            XmlReader validationReader = GetValidationReader();

            XDocument document = XDocument.Load(validationReader);
            validationReader.Close();
            
            return document;
        }

        /// <summary>
        /// Gets the validation reader for the parser output
        /// </summary>
        /// <returns></returns>
        private XmlReader GetValidationReader()
        {
            result.Position = 0;

            XmlReaderSettings readerSettings = new XmlReaderSettings();
            XmlSchema schema = XmlSchema.Read(new StringReader(Resource.FirmwareInfo), validationHandler);
            readerSettings.Schemas.Add(schema);
            readerSettings.ValidationEventHandler += validationHandler;
            readerSettings.ValidationType = ValidationType.Schema;

            return XmlReader.Create(result, readerSettings);
        }

        /// <summary>
        /// Handler for Validation Errors used by the XmlReader 
        /// </summary>
        private void validationHandler(object sender, ValidationEventArgs e)
        {
            ParsingMessage pm = new ParsingMessage();
            pm.Text = e.Message;
            pm.Priority = ParsingMessagePriority.Error;
            pm.File = "INTERNAL ERROR";
            
            messages.Add(pm);
        }
    }
}
