using System.Collections.Generic;
using System.Linq;

namespace BenjaminSchroeter.Flow.Firmwareparser.SemanticChecks
{
    /// <summary>
    /// Semantic Check Rule to ensure that corresponding functions 
    /// (that are functions with the same name in the annotatation)
    /// form valid groups.
    /// 
    /// That is
    ///  - no mixup with different annotation types (e.g. an event and an input with the same name)
    ///  - all events needs a event init method
    ///  - no duplicate declarations (e.g. for GetVariable)
    /// </summary>
    class Chk002 : SemanticCheckBase
    {
        public Chk002(IList<ParsingMessage> messages) : base(messages)    {}

        public override void Check(CodeDescription doc)
        {
            // all function calls that correspond to one shape in the dsl
            var matchingFunctions = from f in doc.function 
                                    group f by f.Item into g
                                    select new { Name = g.Key, Functions = g , AnnotationTypes = from x in g select x.ItemElementName};

            // check each group
            foreach (var shape in matchingFunctions)
            {
                int countEvent = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.@event);
                int countEventCallback = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.eventCallback);
                int sumEvent = countEvent + countEventCallback;

                int countInput = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.input);
                int countInputInit = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.inputInit);
                int sumInput = countInput + countInputInit;

                int countOutput = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.output);
                int countOutputInit = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.outputInit);
                int sumOutput = countOutput + countOutputInit;

                int countGetVariable = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.getVariable);
                int countSetVariable = shape.Functions.Count(x => x.ItemElementName == ItemChoiceType.setVariable);
                int sumVariable = countGetVariable + countSetVariable;

                string[] functionNames = (from f in shape.Functions select f.name+"()").ToArray();
                string functList = string.Join(", ", functionNames);

                if ( (((sumEvent > 0 ? 1 : 0)
                        + (sumInput > 0 ? 1 : 0)
                        + (sumOutput > 0 ? 1 : 0)
                        + (sumVariable > 0 ? 1 : 0 ))) > 1 )
                    ReportError("002a","The element name '{0}' is used for more than one of input, output, variable and event. Please use unique names: {1}.", shape.Name, functList);

                // events
                if ( sumEvent > 0)
                {
                    if (countEvent == 0)
                        ReportError("002b", "For event '{0}' a function with @event annotation is missing: {1}.", shape.Name, functList);
                    if (countEventCallback == 0)
                        ReportError("002c", "For event '{0}' a function with @eventCallback annotation is missing: {1}.", shape.Name, functList);
                    if (countEvent > 1)
                        ReportError("002d", "For event '{0}' more than one function uses the @event annotation: {1}.", shape.Name, functList);
                    if (countEventCallback > 1)
                        ReportError("002e", "For event '{0}' more than one function uses the @eventCallback annotation: {1}.", shape.Name, functList);
                }

                // variable
                if (sumVariable > 0)
                {
                    if (countGetVariable > 1)
                        ReportError("002f", "More than one @getVariable function is defined for '{0}': {1}.", shape.Name, functList);

                    if (countSetVariable > 1)
                        ReportError("002g", "More than one @setVariable function is defined for '{0}': {1}.", shape.Name, functList);

                    if ( countGetVariable == 1 && countSetVariable == 1 )
                    {
                        function f1 = shape.Functions.First(x => x.ItemElementName == ItemChoiceType.getVariable);
                        function f2 = shape.Functions.First(x => x.ItemElementName == ItemChoiceType.setVariable);

                        if (f1.param != null && f1.param.Length == 1 && f2.param != null && f2.param.Length == 1)
                        {
                            string type1 = CodeDescription.GetValueTypeFromPointer(f1.param[0].type);
                            string type2 = CodeDescription.GetValueTypeFromPointer(f2.param[0].type);

                            if ( type1[0] != type2[0])
                                ReportError("002l", "The type of the parameter in both functions for '{0}' must be the same: {1}.", shape.Name, functList);
                        }
                    }
                }

                // input
                if (sumInput > 0)
                {
                    if (countInputInit > 1)
                        ReportError("002h", "For the input port '{0}' more than one function uses the @inputInit annotation: {1}.", shape.Name, functList);
                    if (countInput == 0)
                        ReportError("002i", "For the input port '{0}' at least one function with @input annotation is missing: {1}.", shape.Name, functList);                    
                }

                // output
                if (sumOutput > 0)
                {
                    if (countOutputInit > 1)
                        ReportError("002j", "For the output port '{0}' more than one function uses the @outputInit annotation: {1}.", shape.Name, functList);
                    if (countOutput== 0)
                        ReportError("002k", "For the output port '{0}' at least one function with @output annotation is missing: {1}.", shape.Name, functList);
                }
            }
        }
    }
}
