using System.Collections.Generic;

namespace BenjaminSchroeter.Flow.Firmwareparser.SemanticChecks
{
    /// <summary>
    /// Base class for semantic checks within the firmwarearser.
    /// 
    /// Each rule must implement a <see cref="Check"/> method and validate
    /// "something" on the given CodeDescription.
    /// </summary>
    public abstract class SemanticCheckBase
    {
        private IList<ParsingMessage> messages;

        /// <summary>
        /// Initializes a new instance of the <see cref="SemanticCheckBase"/> class.
        /// </summary>
        /// <param name="messages">All generated messages will be added to this list.</param>
        public SemanticCheckBase(IList<ParsingMessage> messages)
        {
            this.messages = messages;
        }

        /// <summary>
        /// Does the validation of this rule on the given CodeDescription
        /// </summary>
        /// <param name="doc">The CodeDescription to validate.</param>
        public abstract void Check(CodeDescription doc);

        /// <summary>
        /// Reports a validation error.
        /// </summary>
        protected void ReportError(string errorKey, string text, params string[] args)
        {
            Report(ParsingMessagePriority.Error, errorKey, text, args);
        }
        /// <summary>
        /// Reports a validation warning.
        /// </summary>
        protected void ReportWarning(string errorKey, string text, params string[] args)
        {
            Report(ParsingMessagePriority.Warning, errorKey, text, args);
        }
        /// <summary>
        /// Reports an information from validation.
        /// </summary>
        /// <param name="errorKey">The error key.</param>
        /// <param name="text">The text.</param>
        /// <param name="args">The args.</param>
        protected void ReportInformation(string errorKey, string text, params string[] args)
        {
            Report(ParsingMessagePriority.Information, errorKey, text, args);
        }

        /// <summary>
        /// Reports the a message to the validation messages.
        /// </summary>
        private void Report(ParsingMessagePriority priority, string errorKey, string text, params string[] args)
        {
            ParsingMessage p = new ParsingMessage();
            p.Priority = priority;
            p.Text = string.Format(text, args);
            p.ErrorKey = errorKey;

            messages.Add(p);
        }

        /// <summary>
        /// Determines whether the specified parameter type is value type.
        /// </summary>
        /// <param name="parameterType">Type of the parameter.</param>
        /// <returns>
        /// 	<c>true</c> if the specified parameter type is value type; otherwise, <c>false</c>.
        /// </returns>
        protected bool IsValue(string parameterType)
        {
            return !IsPointer(parameterType);
        }

        /// <summary>
        /// Determines whether the specified parameter type is pointer type.
        /// </summary>
        /// <param name="parameterType">Type of the parameter.</param>
        /// <returns>
        /// 	<c>true</c> if the specified parameter type is pointer type; otherwise, <c>false</c>.
        /// </returns>
        protected bool IsPointer(string parameterType)
        {
            return parameterType.EndsWith(" *");
        }
    }
}