﻿using System;
using System.Collections.Generic;
using System.Linq;
using BenjaminSchroeter.Dsl.Graphs;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Representation of the elements in the HardwareDescription model
    /// as a graph with the EnabledRelationships as edges
    /// </summary>
    class HardwareDescriptionGraph : Graph<EnableableDomainClass,EnableableRelationship>
    {
        private readonly Partition partition;

        /// <summary>
        /// Initializes a new instance of the <see cref="HardwareDescriptionGraph"/> class.
        /// </summary>
        /// <param name="model">The HardwareDescription model</param>
        public HardwareDescriptionGraph(HardwareDescriptionModel model)
        {
            this.partition = model.Partition;
        }

        /// <summary>
        /// All Verticeses of this Graph.
        /// </summary>
        /// <remarks>Thats are all EnableableDomainClass of the model.</remarks>
        /// <returns></returns>
        public override IEnumerable<EnableableDomainClass> Vertices()
        {
            return partition.ElementDirectory.AllElements.OfType<EnableableDomainClass>();
        }

        /// <summary>
        /// Children of the specified EnableableDomainClass.
        /// </summary>
        public override IEnumerable<EnableableDomainClass> Children(EnableableDomainClass v)
        {
            // find the source and source field of v
            EnableableShapeMappedDomainClass linkSource;
            Guid linkSourceGuid;
            if (v is EnableableShapeMappedDomainClass)
            {
                linkSource = (EnableableShapeMappedDomainClass)v;
                linkSourceGuid = Guid.Empty;
            }
            else if (v is InputField)
            {
                linkSource = (v as InputField).Input;
                linkSourceGuid = v.Guid;
            }
            else if (v is OutputField)
            {
                linkSource = (v as OutputField).Output;
                linkSourceGuid = v.Guid;
            }
            else
                yield break;

            // follow all incomming lins
            var sources = from l in EnableableRelationship.GetLinksToSources(linkSource)
                          where l.TargetEntry.Equals(linkSourceGuid)
                          select l;
            foreach (var source in sources)
                yield return source.GetSourceAsFieldOrShape();

            // follow all outgoing links
            var targets = from l in EnableableRelationship.GetLinksToTargets(linkSource)
                          where l.SourceEntry.Equals(linkSourceGuid)
                          select l;
            foreach (var target in targets)
                yield return target.GetTargetAsFieldOrShape();
        }

        /// <summary>
        /// Gets a value indicating whether this instance is directed.
        /// </summary>
        /// <value>always false</value>
        public override bool IsDirected
        {
            get { return false;  }
        }
    }
}
