using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Many elements of the Hardware Description DSL hava a string property "call".
    /// This property uses a special format that can be parsed into this CallParameter calls.
    /// 
    /// The CallParameter is a abstraction for this string and can be ask for the components
    /// and can be used to write the data back to the string.
    /// </summary>
    public class CallParameter
    {
        private static readonly Regex functionRegex = new Regex(@"\((.?)\)");

        /// <summary>
        /// Initializes a new instance of the <see cref="CallParameter"/> class.
        /// </summary>
        internal CallParameter()
        {
            ParameterPos = 0;
            ParameterReferenceType = "";
            EnabledParameterPos = 0;
            EnabledParameterReferenceType = "";
            Function = "";
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="CallParameter"/> class and fills it with the given parameter string.
        /// </summary>
        /// <remarks>
        /// The components of the string are seperated by ';'.
        /// A function call ends always with "()".
        /// A single number or a number with the prefex "enabled:" is allowed.
        /// </remarks>
        public CallParameter(string s) : this()
        {
            string[] parts = s.Split(';');
            foreach (string part in parts)
            {
                Match functionMatch = functionRegex.Match(part);
                if (functionMatch.Success )
                {
                    string prefix = functionMatch.Groups[1].Value;
                    this.Function = part.Substring(0, part.Length - 2 - prefix.Length);
                    if (ParameterReferenceType == "" && prefix != "")
                        ParameterReferenceType = prefix;
                }
                else if (part.StartsWith("enabled:",StringComparison.InvariantCultureIgnoreCase))
                {
                    int offset = 0;
                    if (part.Length > 9 && !char.IsNumber(part, 8))
                    {
                        offset = 1;
                        this.EnabledParameterReferenceType = part.Substring(8, 1);
                    }

                    string str = part.Substring(8 + offset);
                    int i;
                    if (int.TryParse(str, out i))
                        this.EnabledParameterPos = i;
                }
                else
                {
                    int offset = 0;
                    if ( part.Length > 1 && !char.IsNumber(part, 0))
                    {
                        offset = 1;
                        this.ParameterReferenceType = part.Substring(0, 1);
                    }

                    string str = part.Substring(offset);
                    int i;
                    if (int.TryParse(str, out i))
                        this.ParameterPos = i;
                }
            }
        }

        #region properties
        /// <summary>
        /// Gets or sets the function name of this call parameter.
        /// </summary>
        /// <value>The function name without a leading "()".</value>
        public string Function { get; set; }
        
        /// <summary>
        /// Gets or sets the parameter position of this call parameter within the corresponding function call.
        /// </summary>
        /// <remarks>Value 0 if there is no parameter position.</remarks>
        /// <value>The parameter position.</value>
        public int ParameterPos { get; set; }

        /// <summary>
        /// Gets or sets the type of the parameter reference. This should be one of "", "&", "*".
        /// </summary>
        public string ParameterReferenceType { get; set; }

        /// <summary>
        /// Gets or sets the parameter pos of the enabled flag within the corresponding function call.
        /// </summary>
        /// <remarks>Value 0 if there is no enabled parameter position.</remarks>
        /// <value>The enabled parameter position.</value>
        public int EnabledParameterPos { get; set; }

        /// <summary>
        /// Gets or sets the type of the enabled parameter reference. This should be one of "", "&", "*".
        /// </summary>
        public string EnabledParameterReferenceType { get; set; }
        #endregion

        #region Has... properties
        /// <summary>
        /// Gets a value indicating whether this call parameter has a function name.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this call parameter has a function name; otherwise, <c>false</c>.
        /// </value>
        public bool HasFunction
        {
            get
            {
                return !string.IsNullOrEmpty(Function);
            }
        }
        /// <summary>
        /// Gets a value indicating whether this call parameter has parameter.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this call parameter has parameter; otherwise, <c>false</c>.
        /// </value>
        public bool HasParameter
        {
            get
            {
                return ParameterPos != 0;
            }
        }
        /// <summary>
        /// Gets a value indicating whether this call parameter has enabled parameter.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this call parameter has enabled parameter; otherwise, <c>false</c>.
        /// </value>
        public bool HasEnabledParameter
        {
            get
            {
                return EnabledParameterPos != 0;
            }
        }
        #endregion

        /// <summary>
        /// Returns this instance in the formard described at <see cref="CallParameter"/>
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            List<string> l = new List<string>();

            if (HasFunction)
                l.Add(Function + "(" + ParameterReferenceType + ")");

            if (HasParameter)
                l.Add(ParameterReferenceType + ParameterPos);

            if (HasEnabledParameter)
                l.Add("enabled:" + EnabledParameterReferenceType + EnabledParameterPos);

            return string.Join(";", l.ToArray());
        }    
    }
}
