using System.Collections.Generic;
using BenjaminSchroeter.Flow.Firmwareparser;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// This collection of CallParameter is used to organize all function calls used to
    /// represent a Port (Input, Output or Event) shape in the firmware.
    /// </summary>
    /// <remarks>
    /// Each Port shape in the HardwareDescription DSL can be mapped to one or more functions
    /// in the firmware. The firmware will be  parsed and all functions mapping to one Port will
    /// be added to this CallParameterCollection.
    /// 
    /// The data obtained by all these function calls will be populated to a Port shape. 
    /// These data must be written to the corresponding domain class itself and also
    /// to the Fields an Parameters of the class.
    /// 
    /// First a number of times the AddFunction() must be called. 
    /// After that one of the Populate...() functions can be called.
    /// </remarks>
    class CallParameterCollection
    {
        private readonly Dictionary<string, CallParameter> fields = new Dictionary<string, CallParameter>();
        private readonly Dictionary<string, CallParameter> parameter = new Dictionary<string, CallParameter>();
        private readonly CallParameter call = new CallParameter();
        private readonly CallParameter initCall = new CallParameter();

        public CallParameterCollection() { }

        /// <summary>
        /// Adds a function from the Firmwareparser to this Collection.
        /// 
        /// The annotations of the function will be read and stored.
        /// If necessary new fields or parameters will be created.
        /// </summary>
        /// <param name="f">The function.</param>
        public void AddFunction(function f)
        {
            // if field is null, the field annotaion is missing
            if ( f.field == null )
            {
                // if param is null, the parameter annotation is missing
                if (f.param != null)
                    AddFunctionParameter(f.param);

                // if this is an init call, remember the function name
                if (f.ItemElementName == ItemChoiceType.inputInit || f.ItemElementName == ItemChoiceType.outputInit || f.ItemElementName == ItemChoiceType.@event)
                    initCall.Function = f.name;

                // if no init call (a regular call) remember name, too
                if (f.ItemElementName == ItemChoiceType.input || f.ItemElementName == ItemChoiceType.output)
                {
                    call.Function = f.name;
                }
            }
            else // only for input and output
            {
                CallParameter c = GetField(f.field);
                c.Function = f.name;
                if (f.param.Length == 1)
                {
                    if (f.param[0].type.EndsWith("*"))
                        c.ParameterReferenceType = "*";
                    if (f.param[0].type.EndsWith("&"))
                        c.ParameterReferenceType = "&";
                }
            }
        }

        /// <summary>
        /// Checks the parameter of a function call
        /// </summary>
        /// <param name="args"></param>
        private void AddFunctionParameter(params param[] args)
        {
            // check each parameter and count the position in the parameter list
            int pos = 0;
            foreach (param p in args)
            {
                string type = "";
                if ( p.type.EndsWith("*") )
                    type ="*";
                else if (p.type.EndsWith("&"))
                    type = "&";

                pos++;
                // depending on the annotation: store the number 
                switch (p.ItemElementName)
                {
                    case ItemChoiceType1.enabled:
                        initCall.EnabledParameterPos = pos;
                        initCall.EnabledParameterReferenceType = type;
                        break;
                    case ItemChoiceType1.field:
                        CallParameter field = GetField(p.Item);
                        field.ParameterPos = pos;
                        field.ParameterReferenceType = type;
                        break;
                    case ItemChoiceType1.fieldenabled:
                        CallParameter fieldenabeld = GetField(p.Item);
                        fieldenabeld.EnabledParameterPos = pos;
                        fieldenabeld.EnabledParameterReferenceType = type;
                        break;
                    case ItemChoiceType1.parameter:
                        CallParameter parameter = GetParameter(p.Item);
                        parameter.ParameterPos = pos;
                        parameter.ParameterReferenceType = type;
                        break;
                }            
            }
        }


        /// <summary>
        /// Gets the parameter from the dictionary (or creates a new one).
        /// </summary>
        /// <param name="name">The parameter name.</param>
        /// <returns></returns>
        private CallParameter GetParameter(string name)
        {
            // Create a new entry if no entry in the dictionary
            if ( !this.parameter.ContainsKey(name) )
                this.parameter.Add(name, new CallParameter());

            return this.parameter[name];
        }
        /// <summary>
        /// Gets the field from the dictionary (or creates a new one)
        /// </summary>
        /// <param name="name">The field name.</param>
        /// <returns></returns>
        private CallParameter GetField(string name)
        {
            // Create a new entry if no entry in the dictionary
            if (!this.fields.ContainsKey(name))
                this.fields.Add(name, new CallParameter());

            return this.fields[name];
        }

        /// <summary>
        /// Writes the data from this instance to the Input DomainClass.
        /// </summary>
        /// <param name="input"></param>
        public void PopulateCallData(Input input)
        {
            PopulateCallData(input as Port);

            foreach (InputField f in input.Fields)
                if (this.fields.ContainsKey(f.Name))
                    f.Call = this.fields[f.Name].ToString();

            input.Call = this.call.ToString();
        }

        /// <summary>
        /// Writes the data from this instance to the Output DomainClass.
        /// </summary>
        /// <param name="output"></param>
        public void PopulateCallData(Output output)
        {
            PopulateCallData(output as Port);

            foreach (OutputField f in output.Fields)
                if (this.fields.ContainsKey(f.Name))
                    f.Call = this.fields[f.Name].ToString();

            output.Call = this.call.ToString();
        }
        
        /// <summary>
        /// Writes the data from this instance to the Event DomainClass.
        /// </summary>
        /// <param name="e"></param>
        public void PopulateCallData(Event e)
        {
            PopulateCallData(e as Port);

            foreach (EventField f in e.Fields)
                if (this.fields.ContainsKey(f.Name))
                    f.Call = this.fields[f.Name].ToString();
        }

        /// <summary>
        /// Writes the data from this instance to the Port DomainClass.
        /// </summary>
        /// <remarks>
        /// This method is used internaly to cover the Parameter existing 
        /// in each of Input, Output and Event.</remarks>
        /// <param name="port"></param>
        private void PopulateCallData(Port port)
        {
            foreach (Parameter p in port.Parameters)
                if (this.parameter.ContainsKey(p.Name))
                    p.Call = this.parameter[p.Name].ToString();

            port.InitCall = this.initCall.ToString();
        }
    }
}