using System.Collections.ObjectModel;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// When a new DataType is changed in this model: inform the DataTypeConverter.
    /// </summary>
    /// <remarks>
    /// The DataTypeConverter is used to show the dropdown list for the type field 
    /// in the properties window.
    /// </remarks>
    [RuleOn(typeof(DataType), FireTime = TimeToFire.LocalCommit)]
    class DataTypeChangeRule:ChangeRule
    {
        public override void ElementPropertyChanged(ElementPropertyChangedEventArgs e)
        {
            // only name changed must be reported
            if ( e.DomainProperty.Id.Equals(DataType.NameDomainPropertyId))
            {
                // remove the old item from the DataTypeConverter
                string oldName = e.OldValue as string;
                if (oldName != null)
                    DataTypes.DataTypeConverter.RemoveDataType(e.ModelElement.Store, oldName);

                // and add the new item
                DataType newType = e.ModelElement as DataType;
                if ( newType != null)
                    DataTypes.DataTypeConverter.AddDataType(e.ModelElement.Store, newType.Name, newType.CType, newType.DslType );                 

                // change the value of the type field for every element in this diagram it used the old one
                if ( oldName != null && newType != null)
                    ChangeTypeName(e.ModelElement.Partition, oldName, newType.Name);
           }
        }

        /// <summary>
        /// Changes the name of the type from the old one to the new for every
        /// element in the given partition that have a Type property.
        /// </summary>
        /// <param name="partition">The partition.</param>
        /// <param name="oldName">The old name.</param>
        /// <param name="newName">The new name.</param>
        private void ChangeTypeName(Partition partition, string oldName, string newName)
        {
            if ( string.IsNullOrEmpty(oldName) || string.IsNullOrEmpty(newName) || oldName==newName)
                return;

            ReadOnlyCollection<ModelElement> variables = partition.ElementDirectory.FindElements(Variable.DomainClassId);
            foreach (Variable variable in variables)
                if (variable.Type != null && variable.Type.Value == oldName)
                    variable.Type.Value = newName;

            ReadOnlyCollection<ModelElement> fields = partition.ElementDirectory.FindElements(EnableableFieldBase.DomainClassId);
            foreach (EnableableFieldBase field in fields)
                if (field.Type != null && field.Type.Value == oldName)
                    field.Type.Value = newName;

            ReadOnlyCollection<ModelElement> fields2 = partition.ElementDirectory.FindElements(FieldBase.DomainClassId);
            foreach (FieldBase field in fields2)
                if (field.Type != null && field.Type.Value == oldName)
                    field.Type.Value = newName;
        }
    }
}
