using System;
using BenjaminSchroeter.Dsl.CompartmentMapping;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Partial class with a few methods used by the classes created by the
    /// CompartmentMapping. See Mapping.xml / Mapping_xGen.cs in this folder.
    /// </summary>
    partial class EnableableRelationshipBuilderInstance
    {
        /// <summary>
        /// Make a connection between the given pair of source and target elements
        /// </summary>
        /// <param name="source">The model element to use as the source of the connection</param>
        /// <param name="sourcePartType">The type of the source. (head or entry)</param>
        /// <param name="sourceEntry">The model element to use as the source entry of the conneciton</param>
        /// <param name="target">The model element to use as the target of the connection</param>
        /// <param name="targetPartType">The type of the target. (head or entry)</param>
        /// <param name="targetEntry">The model elment to use as the target entry of the connection</param>
        /// <returns>
        /// A link representing the created connection
        /// </returns>
        protected override ElementLink CreateElementLink(EnableableShapeMappedDomainClass source, SelectedCompartmentPartType sourcePartType, ModelElement sourceEntry, EnableableShapeMappedDomainClass target, SelectedCompartmentPartType targetPartType, ModelElement targetEntry)
        {
            EnableableRelationship link = new EnableableRelationship(source, target);

            if (sourcePartType == SelectedCompartmentPartType.None || sourcePartType == SelectedCompartmentPartType.Head)
                link.SourceEntry = Guid.Empty;
            else if (sourcePartType == SelectedCompartmentPartType.Element)
                link.SourceEntry = GetFieldGuid(sourceEntry);

            if (targetPartType == SelectedCompartmentPartType.None || targetPartType == SelectedCompartmentPartType.Head)
                link.TargetEntry = Guid.Empty;
            else if (targetPartType == SelectedCompartmentPartType.Element)
                link.TargetEntry = GetFieldGuid(targetEntry);

            // and now one addition: if the source is enabled, disable the target
            bool sourceEnabeld = false;
            if (sourcePartType == SelectedCompartmentPartType.Head || sourcePartType == SelectedCompartmentPartType.None)
                sourceEnabeld = source.Enabled;
            else
            {
                EnableableFieldBase field = sourceEntry as EnableableFieldBase;
                if (field != null)
                    sourceEnabeld = field.Enabled;
            }
            if (sourceEnabeld) // only if source is enabeld
            {
                if (targetPartType == SelectedCompartmentPartType.Head || targetPartType == SelectedCompartmentPartType.None)
                    target.Enabled = false;
                else
                {
                    EnableableFieldBase field = targetEntry as EnableableFieldBase;
                    if (field != null)
                        field.Enabled = false;
                }
            }


            return link;
        }

        /// <summary>
        /// Checks if the given connection has the entry as source
        /// </summary>
        /// <param name="entry">the entry if a compartment shape</param>
        /// <param name="connection">the connection</param>
        /// <returns>
        /// true, if the connection has the entry as source
        /// </returns>
        public override bool IsEntryConnectionSource(ModelElement entry, EnableableRelationship connection)
        {
            if (entry == null)
                return connection.SourceEntry == Guid.Empty;

            return connection.SourceEntry == GetFieldGuid(entry);
        }

        /// <summary>
        /// Checks if the given connection has the entry as target
        /// </summary>
        /// <param name="entry">the entry if a compartment shape</param>
        /// <param name="connection">the connection</param>
        /// <returns>
        /// true, if the connection has the entry as target
        /// </returns>
        public override bool IsEntryConnectionTarget(ModelElement entry, EnableableRelationship connection)
        {
            if (entry == null)
                return connection.TargetEntry == Guid.Empty;

            return connection.TargetEntry == GetFieldGuid(entry);
        }

        /// <summary>
        /// Test whether a given model element is acceptable to this ConnectionBuilder as the source of a connection.
        /// 
        /// Every element that has a enabled property is good for this ConnectionBuilder.
        /// That is:
        ///  - Variabled
        ///  - Head of Input, Output and Event
        ///  - Fileds of Input and Output Shapes
        /// </summary>
        /// <param name="candidate">The model element to test.</param>
        /// <param name="partType">The type of the source. (head or entry)</param>
        /// <param name="candidateEntry">The model element to use as the source entry of the conneciton</param>
        /// <returns>
        /// Whether the element can be used as the source of a connection.
        /// </returns>
        protected override bool CanAcceptAsCompartmentSource(EnableableShapeMappedDomainClass candidate, SelectedCompartmentPartType partType, ModelElement candidateEntry)
        {
            if ( candidate is Variable)
                return true;

            // else: Input, Output, Event
            if (partType == SelectedCompartmentPartType.Head)
                return true;  // Head if them is allowed

            // if not the head but the element of a shape is selected
            if (partType == SelectedCompartmentPartType.Element)
                // only input and output fields are allowed
                // not parameter or event fields (they do not have an enabeld property)
                return candidateEntry is InputField || candidateEntry is OutputField;

            return false; // not allowed
        }

        /// <summary>
        /// Test whether a given model element is acceptable to this ConnectionBuilder as the target of a connection.
        /// 
        /// Uses the same logic, and in fact the same method, as <see cref="CanAcceptAsCompartmentSource"/>.
        /// </summary>
        /// <param name="candidate"></param>
        /// <param name="partType"></param>
        /// <param name="candidateEntry"></param>
        /// <returns>
        /// Whether the elements can be used as the source and target of a connection
        /// </returns>
        protected override bool CanAcceptAsCompartmentTarget(EnableableShapeMappedDomainClass candidate, SelectedCompartmentPartType partType, ModelElement candidateEntry)
        {
            return CanAcceptAsCompartmentSource(candidate, partType, candidateEntry);
        }

        /// <summary>
        /// Gets the GUID of a field.
        /// 
        /// Just one place for the casts and selection...
        /// </summary>
        /// <param name="field">The field.</param>
        /// <returns></returns>
        private Guid GetFieldGuid(ModelElement field)
        {
            if (field is InputField)
                return (field as InputField).Guid;

            if (field is OutputField)
                return (field as OutputField).Guid;

            if (field is EventField)
                return (field as EventField).Guid;

            return Guid.Empty;
        }
    }
}
