using System;
using System.Collections.Generic;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Each Model Element has a Guid. This class creates new Guids for Model Elements
    /// and is used by the <see cref="ModelCreator"/>.
    /// 
    /// The object will always return the same guid.
    /// </summary>
    /// <remarks>
    /// In addition to that the internal cache can be initialized from the Elements
    /// living already in an existing model. With that method I can ensure that
    /// after reparsing the firmware of a Hardwaremodel the elements with the 
    /// same name keep the same Guid.
    /// </remarks>
    internal class GuidCreator
    {
        private readonly Dictionary<string, Guid> guids;

        public GuidCreator()
        {
            guids = new Dictionary<string, Guid>();
        }

        /// <summary>
        /// Gets the GUID for the given element and its arguments.
        /// </summary>
        /// <remarks>
        /// A unique key will be created from the element type and all the string in the argument list.
        /// </remarks>
        /// <param name="element">The element.</param>
        /// <param name="args">The args.</param>
        /// <returns>The Guid for this element.</returns>
        public Guid GetGuid(object element, params string[] args)
        {
            // create a string key from element and args
            string key = Key(element, args);

            // return a guid if there is one in the dictionary
            if (guids.ContainsKey(key))
                return guids[key];

            // create a new one
            Guid g = Guid.NewGuid();
            guids.Add(key, g);

            return g;
        }

        /// <summary>
        /// Creates a unique string key from the element type and the string arguments.
        /// </summary>
        /// <param name="element">The element.</param>
        /// <param name="args">The args.</param>
        /// <returns></returns>
        private string Key(object element, params string[] args)
        {
            return element.GetType().Name + "\\/" + string.Join(".#/", args);
        }

        #region Init from model
        /// <summary>
        /// Inits the guids from a HardwareDescriptionModel to ensure that new elements
        /// with the same name keep their guid.
        /// </summary>
        /// <param name="model">The model.</param>
        public void InitGuidsFromModel(HardwareDescriptionModel model)
        {
            // iterate over all platforms and all elements in that platform
            // and call for everything a AddGuid() method

            foreach (Platform p in model.Platforms)
            {
                AddGuid(p.Guid, p, p.Name);

                foreach (Variable v in p.Variables)
                    AddGuid(v.Guid, v, v.Name);

                foreach (Input i in p.Inputs)
                {
                    AddGuid(i.Guid, i, i.Name);
                    AddParameterGuids(i);
                    AddFieldGuids(i);
                }

                foreach (Output o in p.Outputs)
                {
                    AddGuid(o.Guid, o, o.Name);
                    AddParameterGuids(o);
                    AddFieldGuids(o);
                }

                foreach (Event e in p.Events)
                {
                    AddGuid(e.Guid, e, e.Name);
                    AddParameterGuids(e);
                    AddFieldGuids(e);
                }
            }
        }

        private void AddFieldGuids(Input port)
        {
            foreach (InputField f in port.Fields)
                AddGuid(f.Guid, f, port.Name, f.Name);
        }
        private void AddFieldGuids(Output port)
        {
            foreach (OutputField f in port.Fields)
                AddGuid(f.Guid, f, port.Name, f.Name);
        }
        private void AddFieldGuids(Event port)
        {
            foreach (EventField f in port.Fields)
                AddGuid(f.Guid, f, port.Name, f.Name);
        }
        
        private void AddParameterGuids(Port port)
        {
            foreach (Parameter p in port.Parameters)
                AddGuid(p.Guid, p, port.Name, p.Name);
        }

        private void AddGuid(Guid guid, object element, params string[] args)
        {
            guids.Add(Key(element, args), guid);
        }
        #endregion
    }
}