using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using BenjaminSchroeter.Flow.DataTypes;
using BenjaminSchroeter.Flow.Firmwareparser;
using Microsoft.VisualStudio.Modeling;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Creates a HardwareDescription Model from the code structure of the parsed firmware
    /// </summary>
    /// <remarks>
    /// Uses the Firmwareparser CodeDescription.
    /// </remarks>
    public class ModelCreator
    {
        #region properties
        /// <summary>
        /// The CodeDescription parsed by the firmwareparser
        /// </summary>
        /// <value>The code.</value>
        public CodeDescription code { get; private set; }

        /// <summary>
        /// Gets or sets the newly creates model.
        /// </summary>
        /// <value>The model.</value>
        public HardwareDescriptionModel model { get; private set; }

        /// <summary>
        /// Gets or sets the filenames used to create the code model.
        /// </summary>
        /// <value>The filenames.</value>
        public string[] filenames { get; private set; }

        /// <summary>
        /// Gets or sets the default platform of the model.
        /// This is the platform everything will be generated in.
        /// </summary>
        /// <value>The default platform.</value>
        private Platform defaultPlatform { get; set;}

        private readonly IList<string> messages = new List<string>();
        /// <summary>
        /// Gets the messages created by the ModelCreator.
        /// </summary>
        /// <value>The messages.</value>
        public IList<string> Messages
        {
            get { return messages; }
        }

        /// <summary>
        /// The Stroe of the model to use
        /// </summary>
        private Store store;
        
        /// <summary>
        /// A list of all datatypes found on my way when creating the model.
        /// </summary>
        private HashSet<string> dataTypes;

        /// <summary>
        /// The GuidCreator that will be used to create all Guids used for
        /// elements when creating this model.
        /// </summary>
        private GuidCreator guidCreator;

        /// <summary>
        /// This class will be used to create the DataTypes of this model
        /// </summary>
        private DataTypeCreator dataTypeCreator;

        private ModelSettingsCache settingsCache;
        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="ModelCreator"/> class.
        /// </summary>
        /// <param name="code">The code created by the firmwarepaser.</param>
        /// <param name="name">The filenames used to create the CodeDescription.</param>
        /// <param name="model">The model to generate code in.</param>
        public ModelCreator(CodeDescription code, string[] name, HardwareDescriptionModel model)
        {
            this.filenames = name;
            this.code = code;
            this.model = model;

            // creates the GuidCreator and initialize it with the guids from the given model
            // this guids will be reused for elements with the same name
            guidCreator = new GuidCreator();
            guidCreator.InitGuidsFromModel(model);

            dataTypeCreator = new DataTypeCreator();
            dataTypeCreator.InitDataTypesFromModel(model);

            settingsCache = new ModelSettingsCache();
            settingsCache.InitFromModel(model);
        }

        /// <summary>
        /// Creates the model elements from the parsed code.
        /// </summary>
        public void CreateModelElements()
        {            
            // clear the datatypes hashset
            dataTypes = new HashSet<string>();

            // set the store
            store = model.Store;

            // all changes must be done in a transaction
            using (Transaction tr = store.TransactionManager.BeginTransaction("Create HardwareDescription from parsed code"))
            {
                // creates a Platform. All following elements will be created within this platform.
                // Stored in the global field defaultPlatform.
                if (IsDriver())
                    defaultPlatform = new DriverPlatform(model.Partition);
                else
                    defaultPlatform = new HardwarePlatform(model.Partition);
                model.Platforms.Add(defaultPlatform);

                // Fill the properties of the Model
                FillModelProperties();

                // sorts and groups the functions
                // so that all functions that belong to the same model element will be in one group
                Dictionary<string, List<function>> functionGroups = SortFunctions(code.function);

                // create an element in the model for each function group
                foreach (List<function> functionGroup in functionGroups.Values)
                {
                    switch (functionGroup[0].ItemElementName)
                    {
                        case ItemChoiceType.@event:
                        case ItemChoiceType.eventCallback:
                            CreateEvent(functionGroup);
                            break;

                        case ItemChoiceType.input:
                        case ItemChoiceType.inputInit:
                            CreateInput(functionGroup);
                            break;
                        
                        case ItemChoiceType.output:
                        case ItemChoiceType.outputInit:
                            CreateOutput(functionGroup);
                            break;
                        
                        case ItemChoiceType.getVariable:
                        case ItemChoiceType.setVariable:
                            CreateVariable(functionGroup);
                            break;
                    }
                }

                CreateDataTypes();

                // commit the transaction
                tr.Commit();
            }

            using (Transaction t = store.TransactionManager.BeginTransaction("Rearrange the created Shapes."))
            {
                settingsCache.RestoreSettings(model);
                t.Commit();
            }

            // and clear the global variables
            store = null;
            defaultPlatform = null;
        }

        /// <summary>
        /// Creates the data types in the model for each datatype used in the source code
        /// </summary>
        private void CreateDataTypes()
        {
            // find the type strings from all elements with a Type attribute
            var types1 = model.Partition.ElementDirectory.FindElements(Variable.DomainClassId).Select(x => (x as Variable).Type.Value);
            var types2 = model.Partition.ElementDirectory.FindElements(EnableableFieldBase.DomainClassId).Select(x => (x as EnableableFieldBase).Type.Value);
            var types3 = model.Partition.ElementDirectory.FindElements(FieldBase.DomainClassId).Select(x => (x as FieldBase).Type.Value);

            // concat the type lists and make them distinct
            var types = types1.Concat(types2.Concat(types3)).Distinct();

            // for each type list that is not empty
            foreach (string type in types.Where(s => !string.IsNullOrEmpty(s)))
            {
                string name;
                DataType t = dataTypeCreator.CreateDataType(type, this.store, out name);
                model.DataTypes.Add(t);

                // the name must be set AFTER adding to trigger the Name-Change-Rule.
                t.Name = name;
            }
        }

        #region Create... (Model Elements)
        /// <summary>
        /// Creates a Variable from the data in this functiongroup and adds it to the defaut platform
        /// </summary>
        /// <param name="functions">The functions.</param>
        private void CreateVariable(List<function> functions)
        {
            // validate the group
            if ( !ValidateFunctionGroup(functions, ItemChoiceType.setVariable, ItemChoiceType.getVariable) )
                return;

            // crate the variable
            Variable v = new Variable(store);
            v.Name = functions[0].Item;
            v.Type = new DataTypeString(CodeDescription.GetValueTypeFromPointer(functions[0].param[0].type ));
            v.Guid = guidCreator.GetGuid(v, v.Name);
            foreach (function f in functions)
            {
                if ( !string.IsNullOrEmpty(f.description)  )
                    v.Description = f.description; 

                if (f.ItemElementName == ItemChoiceType.getVariable)
                    v.CallGetter = f.name + "(*)";
                if (f.ItemElementName == ItemChoiceType.setVariable)
                    v.CallSetter = f.name + "()";
            }

            // type of the variable
            dataTypes.Add(v.Type.Value);

            defaultPlatform.Variables.Add(v);
        }

        /// <summary>
        /// Creates an Output from the data in this functiongroup and adds it to the defaut platform
        /// </summary>
        /// <param name="functions">The functions.</param>        
        private void CreateOutput(List<function> functions)
        {
            // validate the group
            if (!ValidateFunctionGroup(functions, ItemChoiceType.outputInit, ItemChoiceType.output))
                return;

            // create the output
            Output o = new Output(store);
            CallParameterCollection cp = new CallParameterCollection();

            o.Name = functions[0].Item;
            o.Guid = guidCreator.GetGuid(o, o.Name);

            foreach (function f in functions)
            {
                // add the functions to the call parameter collection
                cp.AddFunction(f);

                if (!string.IsNullOrEmpty(f.description) && f.field == null)  
                    o.Description = f.description;

                // create fields if needed
                if (f.ItemElementName == ItemChoiceType.output)
                {
                    if (f.field != null)
                    {
                        OutputField field = new OutputField(store);
                        field.Name = f.field;
                        field.Description = f.description ?? "";
                        field.Type = new DataTypeString( CodeDescription.GetValueTypeFromPointer(f.param[0].type));
                        field.Guid = guidCreator.GetGuid(field, o.Name, field.Name);

                        o.Fields.Add(field);
                    }
                    else
                    {
                        CreateField(o, f);
                    }
                }

                // create parameter if needed
                if (f.ItemElementName == ItemChoiceType.outputInit)
                    CreateParameter(o, f);
            }

            cp.PopulateCallData(o);

            // set the optional flag of fields
            foreach (OutputField field in o.Fields)
                field.Optional = field.CallParameter.HasFunction;

            defaultPlatform.Outputs.Add(o);
        }

        /// <summary>
        /// Creates an Input from the data in this functiongroup and adds it to the defaut platform
        /// </summary>
        /// <param name="functions">The functions.</param>   
        private void CreateInput(List<function> functions)
        {
            // validate the group
            if (!ValidateFunctionGroup(functions, ItemChoiceType.inputInit, ItemChoiceType.input))
                return;

            // creates the input
            Input i = new Input(store);
            CallParameterCollection cp = new CallParameterCollection();

            i.Name = functions[0].Item;
            i.Guid = guidCreator.GetGuid(i, i.Name);

            foreach (function f in functions)
            {
                // add the functions to the call parameter collection
                cp.AddFunction(f);

                if (!string.IsNullOrEmpty(f.description) && f.field == null)
                    i.Description = f.description;

                // create fields if needed
                if (f.ItemElementName == ItemChoiceType.input)
                {
                    if (f.field != null)
                    {
                        InputField field = new InputField(store);
                        field.Name = f.field;
                        field.Description = f.description ?? "";
                        field.Type = new DataTypeString(CodeDescription.GetValueTypeFromPointer(f.param[0].type));
                        field.Guid = guidCreator.GetGuid(field, i.Name, field.Name);

                        i.Fields.Add(field);
                    }
                    else
                    {
                        CreateField(i, f);
                    }
                }
                
                // create parameter if needed
                if (f.ItemElementName == ItemChoiceType.inputInit)
                    CreateParameter(i, f);
            }

            cp.PopulateCallData(i);

            defaultPlatform.Inputs.Add(i);
        }

        /// <summary>
        /// Creates an Event from the data in this functiongroup and adds it to the defaut platform
        /// </summary>
        /// <param name="functions">The functions.</param>   
        private void CreateEvent(List<function> functions)
        {
            // validate the group
            if (!ValidateFunctionGroup(functions, ItemChoiceType.@event, ItemChoiceType.eventCallback))
                return;

            // create the event
            Event e = new Event(store);
            CallParameterCollection cp = new CallParameterCollection();
            e.Name = functions[0].Item;
            e.Guid = guidCreator.GetGuid(e, e.Name);
            e.Multiplicity = 1;  // defualt value

            foreach (function f in functions)
            {
                // add the functions to the call parameter collection
                cp.AddFunction(f);
                if (!string.IsNullOrEmpty(f.description))
                    e.Description = f.description;

                if (f.multiplicitySpecified)
                    e.Multiplicity = (int) f.multiplicity;

                if (f.ItemElementName == ItemChoiceType.@event)
                    CreateParameter(e, f);

                if (f.ItemElementName == ItemChoiceType.eventCallback)
                    CreateField(e, f);
            }

            cp.PopulateCallData(e);

            defaultPlatform.Events.Add(e);
        }

        /// <summary>
        /// Creates the fields for a Port from the annotations of a given function
        /// </summary>
        /// <param name="port">The port.</param>
        /// <param name="function">The function.</param>
        private void CreateField(Port port, function function)
        {
            if (function.param == null)
                return;

            foreach (param functionParameter in function.param)
            {
                if (functionParameter.ItemElementName == ItemChoiceType1.field)
                {
                    if (port is Output)
                    {
                        OutputField f = new OutputField(store);
                        f.Name = functionParameter.Item ?? "";
                        f.Description = functionParameter.description ?? "";
                        f.Type = new DataTypeString(CodeDescription.GetValueTypeFromPointer(functionParameter.type));
                        f.Guid = guidCreator.GetGuid(f, port.Name, f.Name);

                        ((Output)port).Fields.Add(f);

                        dataTypes.Add(f.Type.Value);
                    }
                    else if (port is Input)
                    {
                        InputField f = new InputField(store);
                        f.Name = functionParameter.Item ?? "";
                        f.Description = functionParameter.description ?? "";
                        f.Type = new DataTypeString(CodeDescription.GetValueTypeFromPointer(functionParameter.type));
                        f.Guid = guidCreator.GetGuid(f, port.Name, f.Name);

                        ((Input)port).Fields.Add(f);

                        dataTypes.Add(f.Type.Value);
                    }
                    else if (port is Event)
                    {
                        EventField f = new EventField(store);
                        f.Name = functionParameter.Item ?? "";
                        f.Description = functionParameter.description ?? "";
                        f.Type = new DataTypeString(CodeDescription.GetValueTypeFromPointer(functionParameter.type));
                        f.Guid = guidCreator.GetGuid(f, port.Name, f.Name);

                        ((Event)port).Fields.Add(f);

                        dataTypes.Add(f.Type.Value);
                    }
                }
            }
        }

        /// <summary>
        /// Creates the parameters for a Port from the annotations of a given function
        /// </summary>
        /// <param name="port">The port.</param>
        /// <param name="function">The function.</param>
        private void CreateParameter(Port port, function function)
        {
            if (function.param == null)
                return;

            int pos = -1;
            foreach (param functionParameter in function.param)
            {
                pos++;

                if (functionParameter.ItemElementName == ItemChoiceType1.parameter)
                {
                    Parameter p = new Parameter(store);
                    p.Name = functionParameter.Item ?? "";
                    p.Value = functionParameter.@default ?? "";
                    p.Description = functionParameter.description ?? "";
                    p.Type = new DataTypeString( CodeDescription.GetValueTypeFromPointer(functionParameter.type));
                    p.Guid = guidCreator.GetGuid(p, port.Name, p.Name);

                    port.Parameters.Add(p);

                    dataTypes.Add(p.Type.Value);
                }
            }
        }
        #endregion

        /// <summary>
        /// Validates the functiongroup to check that all functions only
        /// belong to the allowed function types
        /// </summary>
        /// <param name="functions">The functions.</param>
        /// <param name="allowedFunctionTypes">The allowed function types.</param>
        /// <returns></returns>
        private bool ValidateFunctionGroup(List<function> functions, params ItemChoiceType[] allowedFunctionTypes)
        {
            foreach (function function in functions)
            {
                bool isValid = false;
                foreach (ItemChoiceType functionType in allowedFunctionTypes)
                    if ( function.ItemElementName == functionType )
                        isValid = true;
            
                if( !isValid )
                {
                    messages.Add(string.Format("Declaration of Element '{0}' is not valid (wrong types).", function.Item));
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Sorts the functions and groups it according to the name of the modele element
        /// that will be created by these functions.
        /// </summary>
        /// <param name="functions">The functions.</param>
        /// <returns></returns>
        private Dictionary<string, List<function>> SortFunctions(function[] functions)
        {
            Dictionary<string, List<function>> functionGroups = new Dictionary<string, List<function>>();

            if (functions == null)
                return new Dictionary<string, List<function>>();

            foreach (function f in functions)
            {
                if ( !functionGroups.ContainsKey(f.Item) )
                    functionGroups.Add(f.Item, new List<function>());

                functionGroups[f.Item].Add(f);
            }

            return functionGroups;
        }

        /// <summary>
        /// Removes all Elements from the Model
        /// </summary>
        public void ClearModel()
        {
            using (Transaction t = model.Store.TransactionManager.BeginTransaction("ModelCreator.ClearModel() clears the Model."))
            {
                model.DataTypes.Clear();
                model.Platforms.Clear();

                t.Commit();
            }
        }

        #region model properties
        /// <summary>
        /// Fills the model properties with the properties found in the source code annotation.
        /// </summary>
        /// <remarks>
        /// Fields with missing source code annotations will be cleard (set to "").
        /// </remarks>
        private void FillModelProperties()
        {
            if (IsDriver())
                defaultPlatform.Name = code.drivername ?? "";
            else
                defaultPlatform.Name = code.firmwarename ?? "";

            defaultPlatform.Version = code.version ?? "";
            defaultPlatform.ProductName = code.productname ?? "";
            defaultPlatform.VendorName = code.vendorname ?? "";
            defaultPlatform.ProductUrl = code.producturl ?? "";
            defaultPlatform.VendorUrl = code.vendorurl ?? "";
            defaultPlatform.Description = code.description ?? "";

            // find the first header file
            foreach (string file in this.filenames)
            {
                if (Path.GetExtension(file).Equals(".h", StringComparison.InvariantCultureIgnoreCase))
                {
                    defaultPlatform.HeaderFile = Path.GetFileName(file);
                    break;
                }
            }
            
            defaultPlatform.Guid = guidCreator.GetGuid(defaultPlatform, defaultPlatform.Name);
        }

        /// <summary>
        /// Determines whether this instance is a driver.
        /// </summary>
        /// <returns><c>true</c> if this instance is driver; otherwise, <c>false</c>.</returns>
        public bool IsDriver()
        {
            return string.IsNullOrEmpty(code.firmwarename);
        }
        #endregion

        #region compiler hint work around
        private void ForTheCompiler()
        {
            // i have to to add single values to the generated code from xsd.exe
            // if i ever need to regenerate the code this line will remember me to do it again (the compiler cannot compile)
            
            // but be careful: this value MUST be the first of the enum
            ItemChoiceType x = ItemChoiceType.none;

            ItemChoiceType1 y = ItemChoiceType1.none;
        }
        #endregion
    }
}
