using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.Modeling.Diagrams;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Caches the position of all shapes before parsing the firmware
    /// to rearrange the shapes after the creation of the new model.
    /// </summary>
    class ModelSettingsCache
    {
        private Dictionary<Guid, RectangleD> shapePositions;
        private RelationshipInfo[] enableableRelationships;
        private CommentInfo[] comments;

        /// <summary>
        /// Inits the cache from the model.
        /// </summary>
        /// <param name="model">The model.</param>
        public void InitFromModel(HardwareDescriptionModel model)
        {
            // position of variable shapes
            shapePositions = new Dictionary<Guid, RectangleD>();

            var vars = model.Partition.ElementDirectory.FindElements(VariableShape.DomainClassId);
            foreach (VariableShape shape in vars)
            {
                Variable variable = (shape.ModelElement as Variable);
                if (variable != null)
                    shapePositions.Add(variable.Guid, shape.Bounds);
            }

            // position of compartment shapes
            var ports = model.Partition.ElementDirectory.FindElements(PortShape.DomainClassId);
            foreach (PortShape shape in ports)
            {
                Port port = (shape.ModelElement as Port);
                if (port != null)
                    shapePositions.Add(port.Guid, shape.Bounds);
            }

            // Enableable Relationship
            var links =
                from link in
                    model.Partition.ElementDirectory.FindElements(EnableableRelationship.DomainClassId).Cast
                    <EnableableRelationship>()
                select new
                         RelationshipInfo
                         {
                             SourceGuid = link.Source.Guid,
                             SourceEntry = link.SourceEntry,
                             TargetGuid = link.Target.Guid,
                             TargetEntry = link.TargetEntry,
                         };
            enableableRelationships = links.ToArray();

            // comments
            var ccc =
                from c in model.Partition.ElementDirectory.FindElements(Comment.DomainClassId).Cast<Comment>()
                select new CommentInfo
                           { 
                               Comment = c,
                               RefClasses = c.EnableableDomainClasses.Select(e => e.Guid).ToArray()
                            };
            comments = ccc.ToArray();
        }

        /// <summary>
        /// Writes everything back to the new model.
        /// </summary>
        /// <param name="model">The model.</param>
        public void RestoreSettings(HardwareDescriptionModel model)
        {
            // position of variable shapes
            var varShapes = model.Partition.ElementDirectory.FindElements(VariableShape.DomainClassId);
            foreach (VariableShape shape in varShapes)
            {
                Variable variable = (shape.ModelElement as Variable);
                if (variable != null)
                    if (shapePositions.ContainsKey(variable.Guid))
                        shape.Bounds = shapePositions[variable.Guid];
            }

            // position of compartment shapes
            var portShapes = model.Partition.ElementDirectory.FindElements(PortShape.DomainClassId);
            foreach (PortShape shape in portShapes)
            {
                Port port = (shape.ModelElement as Port);
                if (port != null)
                    if (shapePositions.ContainsKey(port.Guid))
                    {
                        // for port shapes (compartment shapes) the height must not be changed.
                        var bounds = shapePositions[port.Guid];
                        RectangleD newB = new RectangleD(bounds.X, bounds.Y, bounds.Width, shape.Bounds.Height);
                        shape.Bounds = newB;
                    }
            }

            // Enableable Relationship
            var elemnts = model.Partition.ElementDirectory.FindElements(EnableableShapeMappedDomainClass.DomainClassId).Cast<EnableableShapeMappedDomainClass>();
            foreach (RelationshipInfo info in enableableRelationships)
            {
                // try to find the soruce and target
                EnableableShapeMappedDomainClass source = elemnts.FirstOrDefault(e => e.Guid.Equals(info.SourceGuid));
                if (!IsEnableableRelationshipValidInNewModel(source, info.SourceEntry))
                    continue;
                EnableableShapeMappedDomainClass target = elemnts.FirstOrDefault(e => e.Guid.Equals(info.TargetGuid));
                if (!IsEnableableRelationshipValidInNewModel(target, info.TargetEntry))
                    continue;

                // if we come to here. the source and target are set (!=null)
                // and the fields does exist!
                EnableableRelationship r = new EnableableRelationship(source, target);
                r.SourceEntry = info.SourceEntry;
                r.TargetEntry = info.TargetEntry;
            }

            //// comments
            foreach (CommentInfo info in comments)
            {
                foreach (Guid shapeguid in info.RefClasses)
                {
                    EnableableShapeMappedDomainClass target = elemnts.FirstOrDefault(e => e.Guid.Equals(shapeguid));
                    if (target != null)
                        new CommentReferencesEnableableDomainClasses(info.Comment, target);
                }
                
            }
        }

        /// <summary>
        /// Determines whether the enableable relationship is valid in the new model. That means a
        /// element with the given field exists.
        /// </summary>
        private bool IsEnableableRelationshipValidInNewModel(EnableableShapeMappedDomainClass element, Guid entry)
        {
            if (element == null)
                return false;

            if (entry.Equals(Guid.Empty))
                return true;

            if (element is Input && !(element as Input).Fields.Any(f => f.Guid.Equals(entry)))
                return false;
            if (element is Output && !(element as Output).Fields.Any(f => f.Guid.Equals(entry)))
                return false;
            if (element is Event && !(element as Event).Fields.Any(f => f.Guid.Equals(entry)))
                return false;

            return true;
        }

        private struct RelationshipInfo
        {
            public Guid SourceGuid;
            public Guid SourceEntry;
            public Guid TargetGuid;
            public Guid TargetEntry;
        }

        private struct CommentInfo
        {

            public Comment Comment;
            public Guid[] RefClasses;
        }
    }
}
