using System.Linq;
using Microsoft.VisualStudio.Modeling.Validation;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    [ValidationState(ValidationState.Enabled)]
    partial class EnableableFieldBase
    {
        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateType(ValidationContext context)
        {
            // check for empty types
            if (this.Type == null || string.IsNullOrEmpty(this.Type.Value))
            {
                string msg = string.Format("The Type of the {1} '{0}' must not be empty.", this.Name, this.GetType().Name);
                context.LogError(msg, "Variable.ValidateType", this);
            }
            else
            {
                // the type must be defined in the model
                HardwareDescriptionModel model = (HardwareDescriptionModel) this.Partition.ElementDirectory.FindElements(HardwareDescriptionModel.DomainClassId)[0];
                if (model.DataTypes.FirstOrDefault(d => d.Name == this.Type.Value) == null)
                {
                    string msg = string.Format("The Type '{0}' of the {2} '{1}' is not defined in this Hardware Description.", this.Type.Value, this.Name, this.GetType().Name);
                    context.LogError(msg, "Variable.ValidateType", this);
                }
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateCall(ValidationContext context)
        {
            CallParameter c = this.CallParameter;

            if ( !(c.HasFunction ^ c.HasParameter) )
            {
                // do the right things for input and output fields
                string parentname = "";
                if (this is InputField)
                    parentname = (this as InputField).Name;
                else if (this is OutputField)
                    parentname = (this as OutputField).Name;

                string msg = string.Format("The Call Property of Field '{0}' of '{1}' is invalid. You must provide the position in the parameter list or a function name but not both.", this.Name, parentname);
                context.LogError(msg, "EventField.ValidateCall", this);
            }
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateInitCall(ValidationContext context)
        {
            if ( this.CallParameter.HasEnabledParameter )
            {
                // do the right things for input and output fields
                bool parentHasInitCall = false;
                string parentname = "";
                if (this is InputField)
                {
                    parentname = (this as InputField).Name;
                    parentHasInitCall = (this as InputField).Input.CallParameterInit.HasFunction;
                }
                else if (this is OutputField)
                {
                    parentname = (this as OutputField).Name;
                    parentHasInitCall = (this as OutputField).Output.CallParameterInit.HasFunction;
                }

                if (!parentHasInitCall)
                {
                    string msg = string.Format("The InitCall Property of '{0}' is invalid. Please provide a Function Name since you have defined an enabled position on field '{1}'.",parentname, this.Name);
                    context.LogError(msg, "EnableableFieldBase.ValidateInitCall", this);
                }
            }
        }
    }
}