using System.Collections.Generic;
using BenjaminSchroeter.Dsl;
using Microsoft.VisualStudio.Modeling.Validation;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    [ValidationState(ValidationState.Enabled)]
    partial class Port
    {
        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateParameter(ValidationContext context)
        {
            ValidationHelper.CheckNameUniqueness(context,
                            this.Parameters,
                            "Port.ValidateParameter",
                            x => x.Name);
        }

        protected CallParameter ValidateInputOutputCall(ValidationContext context, CallParameter call)
        {
            if ( call.HasParameter || call.HasEnabledParameter )
            {
                string msg = string.Format("The Call Property of {1} '{0}' is invalid. Please provide only the Function Name.", this.Name, this.GetType().Name);
                context.LogError(msg, "Port.ValidateInputOutputCall", this);
            }

            return call;
        }

        [ValidationMethod(ValidationCategories.Menu | ValidationCategories.Save | ValidationCategories.Open)]
        private void ValidateInitCall(ValidationContext context)
        {
            CallParameter c = this.CallParameterInit;
            if ( this is Input || this is Output || this is Event )
            {
                if ( (c.HasEnabledParameter && !c.HasFunction) || c.HasParameter)
                {
                    string msg = string.Format("The InitCall Property of {1} '{0}' is invalid. You can provide a Function Name or a Function Name and 'enabled' Parameter.", this.Name, this.GetType().Name);
                    context.LogError(msg, "Port.ValidateInitCall", this);
                }
            }

            if ( (c.HasEnabledParameter || this.Parameters.Count > 0) && !c.HasFunction )
            {
                string msg = string.Format("The InitCall Property of {1} '{0}' is invalid. Please provide a Function Name since you have defined an enabled position or parameter for this shape.", this.Name, this.GetType().Name);
                context.LogError(msg, "Port.ValidateInitCall", this);                
            }
        }

        protected void ValidateFunctionParameterUniqueness(ValidationContext context, List<int> callPos, List<int> initPos)
        {
            initPos.Add(this.CallParameterInit.EnabledParameterPos);
            foreach (Parameter parameter in this.Parameters)
                initPos.Add(parameter.CallParameter.ParameterPos);

            // remove 0 (0 is used for "not a parameter")
            callPos.RemoveAll(x => x == 0);
            initPos.RemoveAll(x => x == 0);

            if (!IsPosListValid(callPos) )
            {
                string msg = string.Format("The Call Properties of {1} '{0}' are invalid. Some Function Parameter Positions are used more than once or the Function Parameter are not counting consistently from 1.", this.Name, this.GetType().Name);
                context.LogError(msg, "Port.ValidateFunctionParameterUniqueness", this);
            }
            if (!IsPosListValid(initPos))
            {
                string msg = string.Format("The Init Call Properties of {1} '{0}' are invalid. Some Function Parameter Positions are used more than once or the Function Parameter are not counting consistently from 1.", this.Name, this.GetType().Name);
                context.LogError(msg, "Port.ValidateFunctionParameterUniqueness", this);
            }
        }
        /// <summary>
        /// Determines whether the list of positions is valid.
        /// 
        /// The list is valid if it contains all numbers from 1 without a gap. E.g. 1,2,3....,n where n = list.Length.
        /// </summary>
        private bool IsPosListValid(List<int> list)
        {
            list.Sort();

            int x = 1;
            foreach (int i in list)
            {
                if (i != x)
                    return false;

                x++;
            }

            return true;
        }
    }
}
