using System;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Handler methods for the Parse Firmware command.
    /// </summary>
    class ImportDriverCommandHandler
    {
        public const uint cmdId = 0x2010;
        private const string FileFilterDriver = "*.driverInfo|*.driverInfo|*.*|*.*";

        /// <summary>
        /// Queries the status for the ImportDriver command.
        /// </summary>
        /// <param name="sender">The sender.</param>
            /// <param name="isHardwaredescription">if set to <c>true</c> a hardwaredescription model is selected.</param>
            /// <param name="isSwimlaneSelected">if set to <c>true</c> a is swimlane selected.</param>
        /// <param name="filename">The filename of the selected file.</param>
        public static void QueryStatus(object sender, bool isHardwaredescription, bool isSwimlaneSelected, string filename, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            bool fileExtensionOk = filename.EndsWith(Constants.DesignerFileExtensionHardwareInfo,StringComparison.InvariantCultureIgnoreCase);

            cmd.Enabled = cmd.Visible = isHardwaredescription && isSwimlaneSelected && fileExtensionOk && IsImportDriverServiceAvailable(serviceProvider);
        }

        /// <summary>
        /// Invokes the command.
        /// </summary>
        /// <param name="swimlane">A swimlane of the HardwareDescription model.</param>
        public static void Invoke(SwimlaneShape swimlane, IServiceProvider serviceProvider)
        {
            // find the hardware platform of the model
            Platform selectedPlatform = (Platform)swimlane.ModelElement;
            HardwareDescriptionModel model = selectedPlatform.HardwareDescriptionModel;
            HardwarePlatform mainPlatform = (HardwarePlatform)model.Platforms.FirstOrDefault(p => p is HardwarePlatform);
            if ( mainPlatform == null )
                return;

            // === ask for the driver file using the ImportDriverService ===
            IImportDriverService importDriverService = serviceProvider.GetService(typeof(SImportDriverService)) as IImportDriverService;
            Debug.Assert(importDriverService != null, "IImportDriverService");
            string driverInfoFile = importDriverService.GetDriverInfo();
            if ( driverInfoFile == "")
                return;

            // == open the driver file ===
            HardwareDescriptionModelData driverModel;
            using (Store s = new Store())
            {
                s.LoadDomainModels(new[] {typeof (CoreDesignSurfaceDomainModel), typeof (HardwareDescriptionDomainModel)});
                using (Transaction t = s.TransactionManager.BeginTransaction("loading driver"))
                {
                    HardwareDescriptionModel m = HardwareDescriptionSerializationHelper.Instance.LoadModel(s, driverInfoFile, null, null);
                    driverModel = m.ToDataClass();
                    t.Rollback();
                }
            }

            // === check to avoid loading of the same driver twice ===
            foreach (PlatformData p in driverModel.Platforms)
                if ( model.Platforms.Any( x => x.Guid.Equals(p.Guid) ) )
                {
                    MessageBox.Show(
                        "The selected driver is already imported to the current model.\n It cannot be loaded a second time.",
                        "Duplicated driver loaded", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                    importDriverService.Close();
                    return;
                }

            // === check the compability of the driver with this hardware
            foreach (DriverPlatformData platform in driverModel.Platforms.OfType<DriverPlatformData>())
                if ( !platform.CompatibilityInfo.CompatibleHardware.Contains(mainPlatform.Guid))
                    if ( MessageBox.Show("The selected driver is not compatible with the hardware platfom of the current model.\n\nLoad anyway?\n(If you unsure do NOT load the driver.)","Incompatible dirver",MessageBoxButtons.YesNo,MessageBoxIcon.Question,MessageBoxDefaultButton.Button2) == DialogResult.No)
                    {
                        importDriverService.Close();
                        return;
                    }

            // === check for datatypes
            foreach (DataTypeData type in driverModel.DataTypes)
            {
                DataType hardwaretype = model.DataTypes.FirstOrDefault(x => x.Name.Equals(type.Name));
                if ( hardwaretype == null )
                {
                    MessageBox.Show(
                        string.Format("The selected driver cannot be loaded, since the data type '{0}' is missing in the hardwaredescription.", type.Name ),
                        "Driver load error", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                    importDriverService.Close();
                    return;
                }

                if ( hardwaretype.CType != type.CType 
                    || hardwaretype.DslType != type.DslType )
                {
                    MessageBox.Show(
                        string.Format("The selected driver cannot be loaded, since the definiton of data type '{0}' in the hardwaredescription is different from the driver.", type.Name),
                        "Driver load error", MessageBoxButtons.OK, MessageBoxIcon.Hand);
                    importDriverService.Close();
                    return;
                }
            }
  
            // everything fine to here. import the files
            importDriverService.ImportDriverToProject();
            importDriverService.Close();

            // == finally: import the driver ==
            using (Transaction t = model.Store.TransactionManager.BeginTransaction("Adding driver to HardwareDescription"))
            {
                // == create new swimlane and fill it ===
                foreach (DriverPlatformData platform in driverModel.Platforms.OfType<DriverPlatformData>())
                {
                    Platform p = platform.ToDomainClass(model.Partition);
                    model.Platforms.Add(p);
                }

                // === import the relationships for the enabled flag
                foreach (DriverPlatformData platform in driverModel.Platforms.OfType<DriverPlatformData>())
                    ImportRelationships(platform.CompatibilityInfo.EnableableRelations, model);
                
                t.Commit();
            }

        }

        /// <summary>
        /// Imports the relationships to disable elements of the HardwareDescription model from the driver.
        /// </summary>
        /// <param name="relations">The relationships to create.</param>
        /// <param name="model">The HardwareDescriptionModel.</param>
        private static void ImportRelationships(List<DriverCompatibility.EnableableRelation> relations, HardwareDescriptionModel model)
        {
            foreach (DriverCompatibility.EnableableRelation relation in relations)
            {
                EnableableDomainClass a = FindModelElement(model, relation.DriverElement);
                EnableableDomainClass b = FindModelElement(model, relation.HardwareElement);

                if ( a == null || b == null )
                    continue;

                EnableableShapeMappedDomainClass soruce = GetElement(a);
                EnableableShapeMappedDomainClass target = GetElement(b);

                RoleAssignment soruceRole = new RoleAssignment(EnableableRelationship.SourceDomainRoleId, soruce);
                RoleAssignment targetRole = new RoleAssignment(EnableableRelationship.TargetDomainRoleId, target);
                EnableableRelationship link = new EnableableRelationship(model.Partition, soruceRole, targetRole);
                link.SourceEntry = GetEntry(a);
                link.TargetEntry = GetEntry(b);

                if ( link.GetSourceAsFieldOrShape().Enabled )
                    link.GetTargetAsFieldOrShape().Enabled = false;
            }
        }

        private static EnableableShapeMappedDomainClass GetElement(EnableableDomainClass x)
        {
            if ( x is EnableableShapeMappedDomainClass)
                return (EnableableShapeMappedDomainClass) x;

            if (x is InputField)
                return (x as InputField).Input;

            if (x is OutputField)
                return (x as OutputField).Output;

            return null;
        }

        private static Guid GetEntry(EnableableDomainClass x)
        {
            if (x is EnableableFieldBase)
                return (x as EnableableFieldBase).Guid;

            return Guid.Empty;
        }

        private static EnableableDomainClass FindModelElement(HardwareDescriptionModel model, Guid guid)
        {
            var x = from e in model.Partition.ElementDirectory.AllElements.OfType<EnableableDomainClass>()
                    where e.Guid.Equals(guid)
                        select e;

            if (x.Count() == 0)
                return null;

            return x.ElementAt(0);
        }

        #region service availibility
        // -1 = not checked yet
        // 0  = not available (false)
        // 1  = avaiilable (true)
        private static int isImportDriverServiceAvailable = -1;
        private static bool IsImportDriverServiceAvailable(IServiceProvider serviceProvider)
        {
            if ( isImportDriverServiceAvailable == -1 )
            {
                IImportDriverService x = serviceProvider.GetService(typeof (SImportDriverService)) as IImportDriverService;
                isImportDriverServiceAvailable = x == null ? 0 : 1;
            }

            return isImportDriverServiceAvailable >= 1;
        }
        #endregion
    }
}
