using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.Flow.Firmwareparser;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Modeling.Validation;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.Flow.HardwareDescription
{
    /// <summary>
    /// Service to Parse the firmware .c or .h file and create a Hardware Description model
    /// </summary>
    internal class FirmwareParserService : IFirmwareParser, SFirmwareParser
    {
        private const string FileFilterCode = "*.c;*.h|*.c;*.h|*.c|*.c|*.h|*.h|*.*|*.*";

        private readonly IServiceProvider serviceProvider;
        private readonly ErrorListProvider errorList;
        private readonly FirmwareParserOptionPage optionPage;

        /// <summary>
        /// Initializes a new instance of the <see cref="FirmwareParserService"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        /// <param name="optionPage">The option page.</param>
        /// <param name="errorList">The error list.</param>
        internal FirmwareParserService(IServiceProvider serviceProvider, FirmwareParserOptionPage optionPage, ErrorListProvider errorList)
        {
            this.serviceProvider = serviceProvider;
            this.errorList = errorList;
            this.optionPage = optionPage;
        }

        /// <summary>
        /// The "Parse Firmware" command was invoked on the given project item.
        /// </summary>
        /// <param name="projectItem">The project item.</param>
        public void ParseFirmwareCommandInvoke(ProjectItem projectItem)
        {
            // === get the model file name ===
            string selectedModelFileName = projectItem.get_FileNames(0);
            if (!(selectedModelFileName.EndsWith("." + Constants.DesignerFileExtensionHardwareDescription, StringComparison.InvariantCultureIgnoreCase)
                   || selectedModelFileName.EndsWith("." + Constants.DesignerFileExtensionDriverDescription, StringComparison.InvariantCultureIgnoreCase)))
                return;

            // === get the doxygen settings
            string doxygenExe = "";
            if (optionPage != null)
                doxygenExe = optionPage.DoxygenExecutable;
            if (!File.Exists(doxygenExe) || !doxygenExe.EndsWith(".exe"))
            {
                MessageBox.Show("Cannot find doxygen.exe. Please check the settings at Tools -> Options... -> Hardware Description -> Firmware Parser.");
                return;
            }

            // === open the file (if it is already open nothing will happen) ===
            Window w = projectItem.Open(EnvDTE.Constants.vsViewKindPrimary);
            w.Activate();

            // === get the model from rdt ===
            IVsRunningDocumentTable rdt = (IVsRunningDocumentTable)serviceProvider.GetService(typeof(IVsRunningDocumentTable));
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");
            HardwareDescriptionDocData docData = rdt.GetDocDataForOpenDocument(selectedModelFileName) as HardwareDescriptionDocData;
            Debug.Assert(docData != null, "HardwareDescriptionDocData");

            // === show open file dialog ===
            OpenFileDialog d = new OpenFileDialog { Multiselect = true, Filter = FileFilterCode };
            if (d.ShowDialog() != DialogResult.OK) 
                return;

            // === clear error list ===
            errorList.Tasks.Clear();
            errorList.Refresh();

            // === parse file to CodeDescription ===
            CodeDescription code;
            using (FirmwareParser parser = new FirmwareParser())
            {
                parser.DoxygenExecutable = doxygenExe;
                parser.Parse(d.FileNames);
                code = parser.GetCodeDescription();

                // show all warnings from parser
                if (parser.Messages.Count > 0)
                {
                    foreach (ParsingMessage message in parser.Messages)
                    {
                        Task t = new Task
                        {
                            Text = message.Text,
                            Document = message.File,
                            Line = message.Line,
                            Column = 0
                        };

                        switch (message.Priority)
                        {
                            case ParsingMessagePriority.Information:
                                t.Priority = TaskPriority.Low; 
                                break;
                            case ParsingMessagePriority.Warning:
                                t.Priority = TaskPriority.Normal;
                                break;
                            case ParsingMessagePriority.Error:
                                t.Priority = TaskPriority.High;
                                break;
                        }

                        errorList.Tasks.Add(t);
                    }
                    errorList.Refresh();
                    errorList.Show();

                    // === can the model be created? ===
                    if (Enumerable.Any(parser.Messages, m => m.Priority == ParsingMessagePriority.Error))
                    {
                        MessageBox.Show(
                            "There were errors parsing the file. The errors are shown in the error list. A new model could not be created and the current model was not changed.",
                            "Parsing errors", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                        return;
                    }
                }
            }


            // === create model elements ===
            HardwareDescriptionModel model = (HardwareDescriptionModel)docData.RootElement;
            ModelCreator mc = new ModelCreator(code, d.FileNames, model);
            mc.ClearModel();
            mc.CreateModelElements();

            // show all warnings from model creator
            if (mc.Messages.Count > 0)
            {
                foreach (string s in mc.Messages)
                {
                    Task t = new Task { Priority = TaskPriority.Low, Text = s };
                    errorList.Tasks.Add(t);
                }
                errorList.Refresh();
                errorList.Show();
            }

            // maybe a driver was created on a .hardware-file (or the other way: a hardware description on a .driver file)
            // rename the file
            bool fileIsHardware = projectItem.get_FileNames(0).EndsWith("." + Constants.DesignerFileExtensionHardwareDescription, StringComparison.InvariantCultureIgnoreCase);
            bool firmwareIsHardware = !mc.IsDriver();

            if (fileIsHardware != firmwareIsHardware)
            {
                string newExtension = "." + (firmwareIsHardware
                                          ? Constants.DesignerFileExtensionHardwareDescription
                                          : Constants.DesignerFileExtensionDriverDescription);

                string newName =  Path.GetFileName(Path.ChangeExtension(selectedModelFileName, newExtension));

                string firmwaretype = firmwareIsHardware ? "Hardware" : "Driver";
                // we must rename
                if (MessageBox.Show(string.Format("You parsed the firmware of a {0} into a {1} file. The file will be renamed to '{2}'.", firmwaretype, Path.GetExtension(selectedModelFileName), newName), "Rename Model", MessageBoxButtons.OKCancel, MessageBoxIcon.Information) == DialogResult.OK)
                    projectItem.Name = newName;
            }

            // now validate the file
            docData.ValidationController.Validate(docData.Store, ValidationCategories.Open);
        }
    }
}
