using System;
using System.Diagnostics;
using BenjaminSchroeter.Flow.HardwareDescription.Data;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio.Modeling;
using Microsoft.VisualStudio.Modeling.Diagrams;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.Flow.HardwareDescription.DataService
{
    /// <summary>
    /// DataService to receive the Hardwaredescription of a given project.
    /// </summary>
    public class HardwareDescriptionDataService : IHardwareDescriptionDataService, SHardwareDescriptionDataService
    {
        private readonly IServiceProvider serviceProvider;

        /// <summary>
        /// Initializes a new instance of the <see cref="HardwareDescriptionDataService"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        internal HardwareDescriptionDataService(IServiceProvider serviceProvider)
        {
            this.serviceProvider = serviceProvider;
        }

        /// <summary>
        /// Gets the hardware description from the given project
        /// </summary>
        /// <param name="projectHierarchy">The project hierarchy item.</param>
        /// <returns></returns>
        /// <exception cref="WrongNumberOfHardwareDescriptionException">There are none or more than one Hardware Descriptions in the project</exception>
        public HardwareDescriptionModelData GetHardwareDescription(IVsHierarchy projectHierarchy)
        {
            IVsRunningDocumentTable rdt = (IVsRunningDocumentTable)serviceProvider.GetService(typeof(IVsRunningDocumentTable));
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");

            // === search the project hierarchy for all models (with filename .hardwareinfo) ===
            HierarchyItem modelItem = null;
            int c = 0;
            foreach (HierarchyItem item in projectHierarchy.FindFileItems(file => file.EndsWith("." + Constants.DesignerFileExtensionHardwareInfo, StringComparison.InvariantCultureIgnoreCase)))
            {
                modelItem = item;
                c++;
            }

            // = if no .hardwareinfo file is found, do a fallback and look for .hardware fiels
            if ( c == 0 )
            {
                foreach (HierarchyItem item in projectHierarchy.FindFileItems(file => file.EndsWith("." + Constants.DesignerFileExtensionHardwareDescription, StringComparison.InvariantCultureIgnoreCase)))
                {
                    modelItem = item;
                    c++;
                }                
            }

            if (c != 1)
                throw new WrongNumberOfHardwareDescriptionException(c);

            HardwareDescriptionModel model;
            // === get the models from rdt... ===
            HardwareDescriptionDocData docdata = rdt.GetDocDataForOpenDocument(modelItem) as HardwareDescriptionDocData;
            if (docdata != null && docdata.RootElement is HardwareDescriptionModel)
                model = (HardwareDescriptionModel) docdata.RootElement;
            else
            {
                // === ...or load them from disk  ===
                Store store = new Store();
                store.LoadDomainModels(new[] { typeof(CoreDesignSurfaceDomainModel), typeof(HardwareDescriptionDomainModel) });

                using (Transaction t = store.TransactionManager.BeginTransaction("Load from File"))
                {
                    model = HardwareDescriptionSerializationHelper.Instance.LoadModel(store, modelItem.FullFileName, null, null);
                    t.Commit();
                }
            }
            Debug.Assert(model != null,"Loaded model should be not null.");

            // === return the data from model ===
            HardwareDescriptionModelData data = model.ToDataClass();
            return data;
        }
    }
}
