using System.Collections.Generic;
using System.Linq;

namespace BenjaminSchroeter.Dsl.Graphs
{
    /// <summary>
    /// A simplel implementation of the depth first search.
    /// </summary>
    /// <typeparam name="V">Type of Vertices of the Graph</typeparam>
    /// <typeparam name="E">Type of Edges of the Graph</typeparam>
    public class DepthFirstSearch<V, E> where V : class
    {
        private enum VertexColor
        {
            white,
            gray,
            black
        }

        #region properties
        /// <summary>
        /// The Graph
        /// </summary>
        private readonly Graph<V, E> graph;
        /// <summary>
        /// The color of each vertex
        /// </summary>
        private readonly Dictionary<V, VertexColor> colors;

        /// <summary>
        /// Gets the order in which the vertices were visited.
        /// </summary>
        public IList<V> Order { get; private set;}
        /// <summary>
        /// Gets the connected components of this Graph.
        /// </summary>
        public IList<HashSet<V>> ConnectedComponents { get; private set; }
        /// <summary>
        /// Gets the connected components that form a circle.
        /// </summary>
        public IList<HashSet<V>> ConnectedCircleComponents { get; private set; }
        /// <summary>
        /// Gets all circles found in the graph.
        /// </summary>
        /// <value>The circles.</value>
        public IList<V> Circles { get; private set;  }
        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="DepthFirstSearch&lt;V, E&gt;"/> class.
        /// </summary>
        /// <param name="graph">The graph.</param>
        public DepthFirstSearch(Graph<V,E> graph)
        {
            this.graph = graph;
            
            this.colors = new Dictionary<V, VertexColor>();
            this.Order = new List<V>();
            this.ConnectedComponents = new List<HashSet<V>>();
            this.ConnectedCircleComponents = new List<HashSet<V>>();
            this.Circles = new List<V>();

            // do the search
            Dfs();
        }

        /// <summary>
        /// This implements the depth first search
        /// </summary>
        private void Dfs()
        {
            // color all vertices white
            foreach (V u in graph.Vertices())
                colors.Add(u, VertexColor.white);

            foreach (V u in graph.Vertices())
                if (colors[u] == VertexColor.white)
                {
                    // the following two variables are fill in the Dfs_Visit call
                    bool cycleDetected = false;
                    HashSet<V> currentComponent = new HashSet<V>();

                    Dfs_Visit(u, currentComponent, ref cycleDetected);
                    
                    // remember the newly found component
                    ConnectedComponents.Add(currentComponent);
                    if ( cycleDetected )
                        ConnectedCircleComponents.Add(currentComponent);
                }
        }

        private void Dfs_Visit(V u, HashSet<V> currentComponent, ref bool cycleDetected)
        {
            colors[u] = VertexColor.gray;
            currentComponent.Add(u);

            foreach (V v in graph.Children(u))
            {
               if ( colors[v] == VertexColor.white )
               {
                   Dfs_Visit(v, currentComponent, ref cycleDetected);
               }
               else if (    (colors[v] == VertexColor.gray && graph.IsDirected)
                         || (colors[v] == VertexColor.black && !graph.IsDirected) )
               {
                   cycleDetected = true;
                   this.Circles.Add(v);
               }
            }

            colors[u] = VertexColor.black;
            Order.Add(u);
        }

        /// <summary>
        /// Finds the connected component of a given vertex.
        /// </summary>
        /// <param name="element">The element.</param>
        /// <returns></returns>
        public HashSet<V> FindConnectedComponent(V element)
        {
            foreach (HashSet<V> set in this.ConnectedComponents)
                if ( set.Contains(element) )
                    return set;

            return null;
        }

        /// <summary>
        /// Gets the <see cref="Order"/> reversed
        /// </summary>
        public IEnumerable<V> ReverseOrder
        {
            get
            {
                return Order.Reverse();
            }
        }

        /// <summary>
        /// Gets a value indicating whether a cycle was detected in this graph.
        /// </summary>
        /// <value><c>true</c> if a cycle was detected; otherwise, <c>false</c>.</value>
        public bool CycleDetected
        {
            get
            {
                return this.ConnectedCircleComponents.Count > 0;
            }
        }
    }
}