using System.Collections.Generic;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.VsxTools
{
    /// <summary>
    /// Represents an item of the IVsHierarchy
    /// </summary>
    public class HierarchyItem
    {
        /// <summary>
        /// Gets or sets the IVsHierarchy of this Item.
        /// </summary>
        public IVsHierarchy VsHierarchy { get; private set; }
        /// <summary>
        /// Gets or sets the item id.
        /// </summary>
        public uint ItemId { get; private set; }

        /// <summary>
        /// Initializes a new instance of the <see cref="HierarchyItem"/> class.
        /// </summary>
        /// <param name="vsHierarchy">The IVsHierarchy.</param>
        /// <param name="itemId">The item id.</param>
        public HierarchyItem(IVsHierarchy vsHierarchy, uint itemId)
        {
            this.VsHierarchy = vsHierarchy;
            this.ItemId = itemId;
        }

        /// <summary>
        /// Gets a property of this item
        /// </summary>
        /// <typeparam name="T">The type of the property</typeparam>
        /// <param name="property">The property.</param>
        /// <returns>The value of the property or a default value for type T</returns>
        public T GetProperty<T>(__VSHPROPID property)
        {
            if (!IsValid)
                return default(T);

            object p = GetProperty(property);
            if (p is T)
                return (T)p;
            else
                return default(T);
        }

        /// <summary>
        /// Gets a property of this item
        /// </summary>
        /// <param name="property">The property.</param>
        /// <returns>The value of the property or nullT</returns>
        private object GetProperty(__VSHPROPID property)
        {
            if (!IsValid)
                return null;

            object pvar;
            VsHierarchy.GetProperty(ItemId, (int) property, out pvar);
            return pvar;
        }

        /// <summary>
        /// Gets the parent item of this item.
        /// </summary>
        /// <returns></returns>
        public HierarchyItem GetParent()
        {
            if ( IsRoot )
                return null;

            int parentId = GetProperty<int>(__VSHPROPID.VSHPROPID_Parent);

            return new HierarchyItem(this.VsHierarchy, (uint) parentId);
        }

        /// <summary>
        /// Gets the root item of this item.
        /// </summary>
        /// <returns></returns>
        public HierarchyItem GetRoot()
        {
            HierarchyItem i = this;
            while (!i.IsRoot)
                i = i.GetParent();

            return i;
        }

        /// <summary>
        /// Gets all child HierarchyItem of this item.
        /// </summary>
        /// <returns></returns>
        public IEnumerable<HierarchyItem> GetChildren()
        {
            if (!IsValid)
                yield break;

            int childId = this.GetProperty<int>(__VSHPROPID.VSHPROPID_FirstChild);
            while ((uint)childId != VSConstants.VSITEMID_NIL)
            {
                HierarchyItem child = new HierarchyItem(VsHierarchy, (uint)childId);
                yield return child;

                childId = child.GetProperty<int>(__VSHPROPID.VSHPROPID_NextSibling);
            }
        }

        /// <summary>
        /// Gets all child HierarchyItem of this item in a recursiv manner.
        /// </summary>
        /// <returns></returns>
        public IEnumerable<HierarchyItem> GetChildrenRecursiv()
        {
            if (!IsValid)
                yield break;

            foreach (HierarchyItem item in this.GetChildren())
            {
                yield return item;

                foreach (HierarchyItem i2 in item.GetChildrenRecursiv())
                    yield return i2;
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is the root item of a hierarchy.
        /// </summary>
        /// <value><c>true</c> if this instance is the root item; otherwise, <c>false</c>.</value>
        public bool IsRoot
        {
            get
            {
                if (!IsValid)
                    return false;

                return this.ItemId == VSConstants.VSITEMID_ROOT;
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is valid. That is: the VsHierarchy is not null
        /// </summary>
        /// <value><c>true</c> if this instance is valid; otherwise, <c>false</c>.</value>
        public bool IsValid
        {
            get
            {
                return VsHierarchy != null;
            }
        }

        #region some properties shortcuts
        /// <summary>
        /// Gets the name of this item.
        /// </summary>
        /// <remarks>This is a shortcut for the property __VSHPROPID.VSHPROPID_Name</remarks>
        /// <value>The name.</value>
        public string Name
        {
            get
            {
                if (!IsValid)
                    return "(invalid hierarchy)";

                return GetProperty<string>(__VSHPROPID.VSHPROPID_Name);
            }
        }

        /// <summary>
        /// Gets the name of the file.
        /// 
        /// File name specified on the FileSave menu.
        /// </summary>
        /// <remarks>This is a shortcut for the property __VSHPROPID.VSHPROPID_SaveName</remarks>
        /// <value>The name of the file or "".</value>
        public string FileName
        {
            get
            {
                return GetProperty<string>(__VSHPROPID.VSHPROPID_SaveName) ?? "";
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is non member item.
        /// 
        /// Specifies that item is not considered a member of the hierarchy. 
        /// Required if hierarchy supports (supersetted) views of its contents, 
        /// that is, ShowAllFiles displays files in directory that are not members of the project.
        /// </summary>
        /// <remarks>This is a shortcut for the property __VSHPROPID.VSHPROPID_IsNonMemberItem</remarks>
        /// <value>
        /// 	<c>true</c> if this instance is non member item; otherwise, <c>false</c>.
        /// </value>
        public bool IsNonMemberItem
        {
            get
            {
                return GetProperty<bool>(__VSHPROPID.VSHPROPID_IsNonMemberItem);
            }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is non local storage.
        /// 
        /// Specifies whether project storage is local. IPersistFileFormat is used for saving project regardless of whether it is a local "file" or not.
        /// </summary>
        /// <remarks>This is a shortcut for the property __VSHPROPID.VSHPROPID_IsNonLocalStorage</remarks>
        /// <value>
        /// 	<c>true</c> if this instance is non local storage; otherwise, <c>false</c>.
        /// </value>
        public bool IsNonLocalStorage
        {
            get
            {
                return GetProperty<bool>(__VSHPROPID.VSHPROPID_IsNonLocalStorage);
            }
        }
        #endregion

        /// <summary>
        /// Gets the full name of the file.
        /// </summary>
        /// <value>The full name of the file.</value>
        public string FullFileName
        {
            get
            {
                string s;
                VsHierarchy.GetCanonicalName(ItemId, out s);

                return s ?? "";
            }
        }
    }
}