﻿using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.VsxTools
{
    /// <summary>
    /// Some extension methods for the IVsMonitorSelection interface
    /// </summary>
    public static class IVsMonitorSelectionExtensions
    {
        /// <summary>
        /// Gets the hierarchy of the user selection in the solution explorer.
        /// </summary>
        /// <remarks>
        /// If multiple itmes are selected this returns the hierarchy, too. But only if
        /// all items belong to the same hierarchy.
        /// </remarks>
        /// <param name="monitorSelection">The IVsMonitorSelection instance.</param>
        /// <returns>The hierarchy or null</returns>
        public static IVsHierarchy GetHierarchyOfSelection(this IVsMonitorSelection monitorSelection)
        {
            IntPtr ppHierPtr;
            uint pitemid;
            IVsMultiItemSelect ppMIS;
            IntPtr ppSC;
            monitorSelection.GetCurrentSelection(out ppHierPtr, out pitemid, out ppMIS, out ppSC);

            // only one item is selected or all selected items belong to the same project hierarchy
            if (ppHierPtr != IntPtr.Zero)
            {
                IVsHierarchy ppHier = Marshal.GetObjectForIUnknown(ppHierPtr) as IVsHierarchy;
                Debug.Assert(ppHier != null, "IVsHierarchy from IVsMonitorSelection");
                return ppHier;
            }

            return null;
        }

        /// <summary>
        /// Gets the selected item if only one item is selected in solution explorer.
        /// </summary>
        /// <param name="monitorSelection">The IVsMonitorSelection instance.</param>
        /// <returns>The HierarchyItem or null</returns>
        public static HierarchyItem GetSingleSelectedItem(this IVsMonitorSelection monitorSelection)
        {
            IntPtr ppHierPtr;
            uint pitemid;
            IVsMultiItemSelect ppMIS;
            IntPtr ppSC;
            monitorSelection.GetCurrentSelection(out ppHierPtr, out pitemid, out ppMIS, out ppSC);

            // multiple selection or whatever...
            if (pitemid == VSConstants.VSITEMID_SELECTION ||  ppHierPtr == IntPtr.Zero)
                return null;

            IVsHierarchy ppHier = Marshal.GetObjectForIUnknown(ppHierPtr) as IVsHierarchy;
            Debug.Assert(ppHier != null, "IVsHierarchy from IVsMonitorSelection");

            return new HierarchyItem(ppHier,pitemid);
        }

        /// <summary>
        /// Gets the filename of the selected file in solution explorer.
        /// </summary>
        /// <param name="monitorSelection">The IVsMonitorSelection instance.</param>
        /// <returns>The filename oder any empty string</returns>
        public static string GetSelectedFileName(this IVsMonitorSelection monitorSelection)
        {
            HierarchyItem item = monitorSelection.GetSingleSelectedItem();
            if (item == null)
                return "";
            return item.FileName;
        }

        /// <summary>
        /// Gets the full path of the selected file in solution explorer.
        /// </summary>
        /// <param name="monitorSelection">The IVsMonitorSelection instance.</param>
        /// <returns>The full path as a string or "".</returns>
        public static string GetSelectedFullFileName(this IVsMonitorSelection monitorSelection)
        {
            HierarchyItem item = monitorSelection.GetSingleSelectedItem();
            return item.FullFileName;
        }
    }
}