﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using EnvDTE;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.VsxTools
{
    /// <summary>
    /// Some extension methods for the IVsRunningDocumentTable interface
    /// </summary>
    public static class IVsRunningDocumentTableExtensions
    {
        /// <summary>
        /// Gets the DocData for an already open document.
        /// </summary>
        /// <param name="rdt">The IVsRunningDocumentTable instance.</param>
        /// <param name="filename">The filename.</param>
        /// <returns>The DocData or null if the document is not open.</returns>
        public static object GetDocDataForOpenDocument(this IVsRunningDocumentTable rdt, string filename)
        {
            IVsHierarchy ppHier;
            uint pitemid;
            uint pdwCookie;
            IntPtr ppunkDocData;
            int ret = rdt.FindAndLockDocument((uint) _VSRDTFLAGS.RDT_NoLock,
                                              filename, out ppHier, out pitemid, out ppunkDocData, out pdwCookie);

            

            if (ret != VSConstants.S_OK)
                return null;

            return Marshal.GetObjectForIUnknown(ppunkDocData);
        }

        /// <summary>
        /// Gets the DocData for an already open document.
        /// </summary>
        /// <param name="rdt">The IVsRunningDocumentTable instance.</param>
        /// <param name="item">The HierarchyItem if the document to get.</param>
        /// <returns>The DocData or null if the document is not open.</returns>
        public static object GetDocDataForOpenDocument(this IVsRunningDocumentTable rdt, HierarchyItem item)
        {
            uint pgrfRDTFlags;
            uint pdwReadLocks;
            uint pdwEditLocks;
            string pbstrMkDocument;
            IVsHierarchy pphier;
            uint pitemid;
            IntPtr punkDocData;

            foreach (uint u in GetItems(rdt))
            {
                rdt.GetDocumentInfo(u, out pgrfRDTFlags, out pdwReadLocks, out pdwEditLocks, out pbstrMkDocument, out pphier, out pitemid, out punkDocData);
                if (pphier == item.VsHierarchy && pitemid == item.ItemId)
                    return Marshal.GetObjectForIUnknown(punkDocData);
            }

            return null;
        }

        /// <summary>
        /// Gets the DocData for all open documents.
        /// </summary>
        /// <param name="rdt">The IVsRunningDocumentTable instance.</param>
        /// <returns></returns>
        public static IEnumerable<RdtInfo> GetDocDataForOpenDocuments(this IVsRunningDocumentTable rdt)
        {
            uint pgrfRDTFlags;
            uint pdwReadLocks;
            uint pdwEditLocks;
            string pbstrMkDocument;
            IVsHierarchy pphier;
            uint pitemid;
            IntPtr punkDocData;

            foreach (uint u in GetItems(rdt))
            {
                rdt.GetDocumentInfo(u, out pgrfRDTFlags, out pdwReadLocks, out pdwEditLocks, out pbstrMkDocument, out pphier, out pitemid, out punkDocData);
                RdtInfo info = new RdtInfo(pbstrMkDocument, new HierarchyItem(pphier,pitemid),Marshal.GetObjectForIUnknown(punkDocData));
                yield return info;
            }
        }

        public struct RdtInfo
        {
            private string mkDocument;
            private HierarchyItem item;
            private object docData;

            public string MkDocument
            {
                get { return mkDocument; }
            }

            public HierarchyItem Item
            {
                get { return item; }
            }

            public object DocData
            {
                get { return docData; }
            }

            public RdtInfo(string mkdocument, HierarchyItem item, object docData)
            {
                mkDocument = mkdocument;
                this.item = item;
                this.docData = docData;
            }
        }

        /// <summary>
        /// Gets all document cookies of the entries in the RDT.
        /// </summary>
        /// <param name="rdt">The IVsRunningDocumentTable instance.</param>
        /// <returns></returns>
        private static IEnumerable<uint> GetItems(this IVsRunningDocumentTable rdt)
        {
            IEnumRunningDocuments ppenum;
            rdt.GetRunningDocumentsEnum(out ppenum);

            uint pceltFetched = 1;
            uint[] rgelt = new uint[1];

            while (pceltFetched > 0)
            {
                ppenum.Next(1, rgelt, out pceltFetched);

                if (pceltFetched > 0)
                    yield return rgelt[0];
            }
        }

        /// <summary>
        /// Saves all open documents that belong to the given project instance.
        /// </summary>
        /// <param name="rdt">The IVsRunningDocumentTable instance.</param>
        /// <param name="project">The project.</param>
        /// <param name="silent">if set to <c>true</c> do a silent save (do not ask the user).</param>
        public static void SaveAllDocumentsFromProject(this IVsRunningDocumentTable rdt, Project project, bool silent)
        {
            VSSAVEFLAGS saveflag = silent ? VSSAVEFLAGS.VSSAVE_SilentSave : VSSAVEFLAGS.VSSAVE_Save;

            // iterate all open documents
            foreach (IVsRunningDocumentTableExtensions.RdtInfo info in rdt.GetDocDataForOpenDocuments())
            {
                IVsPersistDocData2 docData = info.DocData as IVsPersistDocData2;
                if (docData != null && info.Item.IsValid)
                {
                    HierarchyItem projectFromRdt = info.Item.GetRoot();
                    // if it belongs to the given project
                    if (projectFromRdt.FullFileName.Equals(project.FullName, StringComparison.InvariantCultureIgnoreCase))
                    {
                        int i;
                        docData.IsDocDataDirty(out i);
                        if (i == 1) // and if it dirty
                        {
                            string a;
                            int b;
                            docData.SaveDocData(saveflag, out a, out b);
                        }
                    }
                }
            }

            if (project.IsDirty)
                project.Save(null);

        }
    }
}
