using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using BenjaminSchroeter.Flow.CodeGen;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.TextManager.Interop;
using Constants=EnvDTE.Constants;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// This service provides an interface to open a (custom) editor for
    /// CodeBlock shapes. For CodeBlocks the user needs some place where he
    /// can put the code for this block. 
    /// The DataFlow DSL Editor will use this service whenever the
    /// user request to edit the code of the CodeBlock.
    /// </summary>
    public class CodeBlockEditor : ICodeBlockEditor, SCodeBlockEditor
    {
        private const string USER_CODE_DIRECTORY = "UserCodeblocks";

        private readonly IServiceProvider serviceProvider;

        /// <summary>
        /// Initializes a new instance of the <see cref="CodeBlockEditor"/> class.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public CodeBlockEditor(IServiceProvider serviceProvider)
        {
            this.serviceProvider = serviceProvider;
        }

        /// <summary>
        /// Opens the code block editor.
        /// </summary>
        /// <param name="codeblock">The codeblock to open in the editor.</param>
        /// <param name="modelName">Name of the dataflow model.</param>
        /// <param name="hierarchy">The hierarchy item of the dataflow model.</param>
        public void OpenCodeBlockEditor(CodeBlock codeblock, string modelName, IVsHierarchy hierarchy)
        {
            bool created = false;

            // get the name of the codeblock function in c code
            string name = CodeBlockGenerator.GenerateFunctionName(modelName, codeblock, new UniqueNameProvider());
            // use this name as filename
            string filename = name + ".c";

            // search for a file in this project with the given name
            Project project = hierarchy.ToDteProject();
            Debug.Assert(project != null, "Hierarchy is not a Project");
            List<ProjectItem> files = project.FindAllFiles();
            ProjectItem codeBlockFile = files.Where(p => p.get_FileNames(0).EndsWith(filename, StringComparison.InvariantCultureIgnoreCase)).FirstOrDefault();
            
            // if file not found in the projec: add it and create it if nessesary
            if ( codeBlockFile == null )
            {
                // the directory for user files 
                string pathOfProject = Path.GetDirectoryName(hierarchy.GetRoot().FullFileName);
                DirectoryInfo di = new DirectoryInfo(Path.Combine(pathOfProject, USER_CODE_DIRECTORY));
                if ( !di.Exists )
                    di.Create();

                // full filename of the codeblock .c file in the user code directory
                string fullfilename = Path.Combine(di.FullName, filename);
                FileInfo fi = new FileInfo(fullfilename);
                if ( !fi.Exists )
                {   
                    // create the file if necessary
                    using (FileStream x = fi.Create() )
                    {
                        x.Close();
                    }
                    created = true;
                }

                // add the file to the project
                codeBlockFile = project.ProjectItems.AddFromFile(fullfilename);
            }

            // open the file
            if ( codeBlockFile != null )
            {
                Window w = codeBlockFile.Open(Constants.vsViewKindPrimary);
                w.Activate();

                // now: whats with the content of the file?
                FileContent(codeBlockFile.get_FileNames(0), created, codeblock, modelName, serviceProvider);
            }
        }

        /// <summary>
        /// Creates or edits the content of the codeblock file.
        /// </summary>
        /// <param name="fullfilename">The fullfilename.</param>
        /// <param name="created">if set to <c>true</c> the file was created in this run.</param>
        /// <param name="codeblock">The codeblock.</param>
        /// <param name="modelname">The modelname.</param>
        /// <param name="serviceProvider">The service provider.</param>
        private void FileContent(string fullfilename, bool created, CodeBlock codeblock, string modelname, IServiceProvider serviceProvider)
        {
            // the whole signature of the codeblock as a c function definition
            string signature = CodeBlockGenerator.GenerateFunctionSignature(modelname, codeblock, new UniqueNameProvider());

            IVsRunningDocumentTable rdt = serviceProvider.GetService(typeof(IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "rdt");

            // receive the (open) file from rdt as IVsTextLines instance
            IVsTextLines textDoc = rdt.GetDocDataForOpenDocument(fullfilename) as IVsTextLines;
            if (textDoc == null)
                return;

            object e;
            textDoc.CreateEditPoint(0, 0, out e);
            EditPoint ep = e as EditPoint;
            if (ep != null)
            {
                // get the whole file content
                string content = ep.GetText(int.MaxValue);    
                
                // does our signature already exists?
                if ( !DoesFileContainsSignature(content, signature) )
                {
                    // finally: insert the code
                    ep.Insert(CodeBlockCode(codeblock, modelname, signature));

                    // if the file already exists. add a marker
                    if (!created)
                        ep.Insert(PreviousCodeMarker());
                }
            }
        }

        /// <summary>
        /// Checks if the file contains a given signature.
        /// </summary>
        /// <remarks>All whitespaces will be removed for this test</remarks>
        private bool DoesFileContainsSignature(string content, string signature)
        {
            // finds all Whitespaces one or many in a row (Space, Tab, Newline) 
            Regex whitespaces = new Regex(@"(\s)+",RegexOptions.Multiline);

            // clean up the contetn and the signature
            string cleanContent = whitespaces.Replace(content, " ");
            string cleanSignature = whitespaces.Replace(signature, " ");

            // now: search
            return cleanContent.Contains(cleanSignature);
        }

        /// <summary>
        /// Generates the code to insert into the file
        /// </summary>
        private static string CodeBlockCode(CodeBlock codeblock, string modelname, string signature)
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendFormatLine("/* This file contains user code for the");
            sb.AppendFormatLine("   Codeblock '{0}'", codeblock.Name);
            sb.AppendFormatLine("   from the Model '{0}' */", modelname);
            sb.AppendLine();
            sb.AppendLine("// the following line must match exactly the definition in the model");
            sb.AppendLine("// please do not change this line:");
            sb.AppendLine(signature);
            sb.AppendLine("{");
            sb.AppendLine(CodeBlockGenerator.GenerateFunctionContent(codeblock, "/* TODO: Insert your code here */"));
            sb.AppendLine("}");
            sb.AppendLine();

            return sb.ToString();
        }

        private static string PreviousCodeMarker()
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendLine("/**************************************************");
            sb.AppendLine(" * TODO: previous code of this file, please check *");
            sb.AppendLine(" **************************************************/");

            return sb.ToString();
        }
    }
}
