using System;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.Zip;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// Comamnd Handler for the "Create Flow Driver Package..." entry
    /// in the Project context menu of .driver files.
    /// </summary>
    class CreateFlowDriverPackageHandler
    {
        public const int cmdidCreateDriverPackage = 0x0200;

        private static readonly string[] hiddenFileExtensionsForDialog = {Const.FileExtensionDriverInfo, 
                                                                             Const.FileExtensionDriverDescription, 
                                                                             Const.FileExtensionHardwareInfo, 
                                                                             Const.FileExtensionHardwareDescription,
                                                                             Const.FileExtensionDiagramAddition };

        /// <summary>
        /// Queries the status of the command.
        /// </summary>
        /// <remarks>The command will only be visible when exactly one file is selected and this file is a .driver file.</remarks>
        /// <param name="sender">The sender.</param>
        /// <param name="serviceProvider">The service provider.</param>
        public static void QueryStatus(object sender, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            DTE dte = serviceProvider.GetService(typeof(_DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");

            // multiple selection or whatever..
            if (dte.SelectedItems.Count != 1)
            {
                cmd.Visible = false;
                return;
            }

            string filename = dte.SelectedItems.Item(1).Name;
            cmd.Visible = filename.EndsWith(Const.FileExtensionDriverDescription, StringComparison.InvariantCultureIgnoreCase);
        }

        /// <summary>
        /// Invokes the command
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void Invoke(IServiceProvider serviceProvider)
        {
            DTE dte = serviceProvider.GetService(typeof(_DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");

            // multiple selection or whatever..
            if (dte.SelectedItems.Count != 1)
                return;

            // get the selected project item (a .driver file)
            ProjectItem projectItem = dte.SelectedItems.Item(1).ProjectItem;
            // and the current project
            Project project = projectItem.ContainingProject;

            // == save all open files form the current project
            IVsRunningDocumentTable rdt = serviceProvider.GetService(typeof(IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");
            rdt.SaveAllDocumentsFromProject(project, true);

            // == show dialog (to select files from project)
            SelectProjectFilesDialog filesDialog = new SelectProjectFilesDialog(project, hiddenFileExtensionsForDialog);
            if ( filesDialog.ShowDialog() != DialogResult.OK )
                return;

            // == save file dialog (for the driver package)
            SaveFileDialog fileDialog = new SaveFileDialog();
            fileDialog.Filter = "*.flowdrv|*.flowdrv|*.*|*.*";
            if (fileDialog.ShowDialog() != DialogResult.OK)
                return;

            // == create driver package
            // all files from the first dialog
            string rootPath = Path.GetDirectoryName(project.FullName);
            List<string> files = (from p in filesDialog.SelectedProjectItems
                        select PathUtil.RelativePathTo(rootPath, p.get_FileNames(0))).ToList();
            // add the .driver and dirver.diagram file
            string driverFileName = projectItem.get_FileNames(0);
            string driverDiagramFileName = driverFileName + ".diagram";
            if ( !files.Any( s => s.EndsWith(driverFileName,StringComparison.InvariantCultureIgnoreCase) ))
                files.Add(PathUtil.RelativePathTo(rootPath, driverFileName));
            if ( File.Exists(driverDiagramFileName) && !files.Any(s => s.EndsWith(driverDiagramFileName, StringComparison.InvariantCultureIgnoreCase)))
                files.Add(PathUtil.RelativePathTo(rootPath, driverDiagramFileName));

            // the constructor creates the zip file! (i know, bad design, but thats the way this class works)
            new ZipFileCompressor(fileDialog.FileName, rootPath, files.ToArray(), true, true);
        }
    }
}
