using System;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Serialization;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.TextTemplating.VSHost;
using Microsoft.Win32;
using Process=System.Diagnostics.Process;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// Comamnd Handler for the "Create Flow Platform Setup..." entry
    /// in the Project context and main menu.
    /// </summary>
    class CreateFlowPlatformSetupHandler
    {
        public const int cmdidCreatePlatformSetup = 0x0100;
        public const int cmdidCreatePlatformSetup2 = 0x0105;

        public const string guidCreatePlatformSetupString = "58EEA246-6759-4939-AE0E-E6BF52070997";
        public static readonly Guid guidCreatePlatformSetup = new Guid(guidCreatePlatformSetupString);

        /// <summary>
        /// Queries the status of the Command.
        /// </summary>
        /// <remarks>The command is only visible if a Scatter Web Dsl project is selected and a .hardware file is present in the project.</remarks>
        public static void QueryStatus(object sender, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");
            object[] projects = dte.ActiveSolutionProjects as object[];
            Debug.Assert(projects != null);
            if (projects.Length != 1)
            {
                cmd.Visible = false;
                return;
            }

            Project project = projects[0] as Project;
            Debug.Assert(project != null, "No project selected while using a project add context menu.");

            // is it a Scatter Web dsl project for hardware vendors?
            cmd.Visible = project.IsScatterWebDslHardwarePlatformProject();
        }

        /// <summary>
        /// Invokes the handler and create a flow platform setup.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void Invoke(IServiceProvider serviceProvider)
        {
            // == get the project
            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");
            object[] projects = dte.ActiveSolutionProjects as object[];
            Debug.Assert(projects != null);
            if (projects.Length != 1)
                return;

            Project project = projects[0] as Project;
            string projectFile = project.FullName;

            // -- check for a setup.xml file in the same directory
            DirectoryInfo di = new DirectoryInfo( Path.GetDirectoryName(projectFile));
            FileInfo setupXml = di.GetFiles("setup.xml").FirstOrDefault();
            FlowHardwarePlatformSetup settings;
            if ( setupXml == null )
                settings = CreateEmptySettings(project);
            else
            {
                try
                {
                    XmlReader xmlReader = new XmlTextReader(setupXml.FullName);
                    XmlSerializer deserializer = new XmlSerializer(typeof(FlowHardwarePlatformSetup));
                    settings = (FlowHardwarePlatformSetup)deserializer.Deserialize(xmlReader);
                    xmlReader.Close();
                }
                catch
                {
                    settings = CreateEmptySettings(project);
                }
            }

            // == open the settings dialog
            FlowPlatformSetupDialog d = new FlowPlatformSetupDialog(settings, di.FullName);
            if ( d.ShowDialog() != DialogResult.OK )
                return;
            // - save the settings file
            string setupFile = Path.Combine(di.FullName, "setup.xml");
            XmlTextWriter xmlWriter = new XmlTextWriter(setupFile, Encoding.Default);
            XmlSerializer serializer = new XmlSerializer(typeof(FlowHardwarePlatformSetup));
            serializer.Serialize(xmlWriter, settings);
            xmlWriter.Close();

            // == save all open files form the current project
            IVsRunningDocumentTable rdt = serviceProvider.GetService(typeof (IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");
            rdt.SaveAllDocumentsFromProject(project, true);

            // == create the in memory tt file
            string nsiTt = string.Format(Resources.SetupNsiTt, projectFile, setupFile);

            // == transform tt file to nsi
            ITextTemplating templatingService = serviceProvider.GetService(typeof(STextTemplating)) as ITextTemplating;
            Debug.Assert(templatingService != null, "ITextTemplating");

            TextTemplatingCallback callback = new TextTemplatingCallback();
            string nsiScriptContent = templatingService.ProcessTemplate(projectFile  +".tt", nsiTt, callback, null);
            string nsiFile = Path.ChangeExtension(setupFile, callback.FileExtension);
            StreamWriter sw = new StreamWriter( nsiFile );
            sw.Write(nsiScriptContent);
            sw.Close();

            // == execute nsi
            // - try to find the nsi path in the registry
            string exe = GetNsisPath();
            if (exe == "")
            {
                MessageBox.Show(
                    "NSIS (Nullsoft Scriptable Install System) cannot be found. Please install it from http://nsis.sourceforge.net to create a Flow Platform Setup.",
                    "NSIS not found", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                
                return;
            }

            Process.Start(exe, "\"" + nsiFile + "\"");
        }

        /// <summary>
        /// Gets the path of the nsis executable. If the path cannot be found, "" will be returend.
        /// </summary>
        /// <returns></returns>
        private static string GetNsisPath()
        {
            try
            {
                // search for the path in the registry
                RegistryKey k1 = Registry.LocalMachine.OpenSubKey("Software");
                RegistryKey k2 = k1.OpenSubKey("NSIS");
                string dir = k2.GetValue("", "") as string;
                k2.Close();
                k1.Close();
                
                // add the file name to the path
                string exe = Path.Combine(dir, "makensisw.exe");

                // and check if it exists
                if ( File.Exists(exe))
                    return exe;
                else
                    return "";
            }
            catch
            {
                return "";
            }
        }

        /// <summary>
        /// Creates the empty FlowHardwarePlatformSetup instance for the setup.xml file.
        /// </summary>
        private static FlowHardwarePlatformSetup CreateEmptySettings(Project project)
        {
            FlowHardwarePlatformSetup s = new FlowHardwarePlatformSetup();
            s.ShortName = project.Name;
            s.FullName = project.Name;
            s.SetupName = project.Name + "Setup.exe";
            s.Guid = Guid.NewGuid().ToString();
            return s;
        }
    }
}
