using System;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.VsxTools;
using EnvDTE;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// Comamnd Handler for the "Add [Datastructure, Dataflow]" context menu entry
    /// in the Solution Add Item menu.
    /// </summary>
    internal class AddDslItemHandler
    {
        public const int cmdidAddDatastructures = 0x3FE7;
        public const int cmdidAddDataflow = 0x3FF2;
        public const int cmdidAddEmptyDriver = 0x3005;

        public const int cmdidAddDatastructures2 = 0x3FE8;
        public const int cmdidAddDataflow2 = 0x3FF3;
        public const int cmdidAddEmptyDriver2 = 0x3003;

        public const string guidAddDslString = "E701C545-52D1-4729-BC88-C28629A1AD9D";
        public static readonly Guid guidAddDsl = new Guid(guidAddDslString);

        /// <summary>
        /// Queries the status of the command for User Menu Items, that are Datastructures and Dataflow
        /// </summary>
        /// <remarks>
        /// The command is only visible if a Scatter Web Dsl project is selected.
        /// </remarks>
        /// <param name="sender">The sender.</param>
        /// <param name="serviceProvider">The service provider.</param>
        public static void QueryStatusUserItems(object sender, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            DTE dte = serviceProvider.GetService(typeof (DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");
            object[] projects = dte.ActiveSolutionProjects as object[];
            Debug.Assert(projects != null);
            if (projects.Length == 0)
            {
                cmd.Visible = false;
                return;
            }
            
            Project project = projects[0] as Project;
            Debug.Assert(project != null, "No project selected while using a project add context menu.");

            // is it a Scatter Web dsl project?
            cmd.Visible = project.IsScatterWebDslProject();
        }

        /// <summary>
        /// Queries the status of the command for Vendor Menu Items, that is the Empty Flow Driver
        /// </summary>
        /// <remarks>
        /// The command is only visible if a Scatter Web Dsl project is selected and the showHardwareVenderOptions option is set
        /// </remarks>
        /// <param name="sender">The sender.</param>
        /// <param name="serviceProvider">The service provider.</param>
        public static void QueryStatusVendorItems(object sender, IServiceProvider serviceProvider, bool showHardwareVenderOptions)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            if (!showHardwareVenderOptions)
            {
                cmd.Visible = cmd.Enabled = false;
            }
            else
            {
                cmd.Visible = cmd.Enabled = true;
                QueryStatusUserItems(cmd, serviceProvider);                
            }
        }

        /// <summary>
        /// Handler / Invoke method for "Add datastructes..."
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void AddDatastructes(IServiceProvider serviceProvider)
        {
            AddDsl(serviceProvider, "datastructures", Const.FileExtensionDatastructures, "Datastructures", DatastructuresTtContent);
        }

        /// <summary>
        /// Handler / Invoke method for "Add dataflow..."
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void AddDataflow(IServiceProvider serviceProvider)
        {
            AddDsl(serviceProvider, "dataflow", Const.FileExtensionDataFlow, "Dataflow", DataflowTtContent);
        }

        /// <summary>
        /// Handler / Invoke method for "Add empty flow driver..."
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void AddEmptyDriver(IServiceProvider serviceProvider)
        {
            AddDsl(serviceProvider, "flow driver", Const.FileExtensionDriverDescription, "Driver", null);
        }
        
        /// <summary>
        /// Adds a new Dsl model item to the project
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        /// <param name="FileDescription">The description of the file to show in error messages.</param>
        /// <param name="Extension">The extension of the new dsl file.</param>
        /// <param name="DefaultName">The default name (prefix) of the new file.</param>
        /// <param name="content">The content for the corresponding tt file or null if there is no tt to add</param>
        private static void AddDsl(IServiceProvider serviceProvider, string FileDescription, string Extension, string DefaultName, Func<Project, string, string> content)
        {
            // get services
            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");
            object[] projects = dte.ActiveSolutionProjects as object[];
            Debug.Assert(projects != null);
            Debug.Assert(projects.Length >= 1,"No project selected while using a project add context menu.");
            Project project = projects[0] as Project;
            Debug.Assert(project != null, "No project selected while using a project add context menu.");

            // the path to add the items
            string projectDir = Path.GetDirectoryName(project.FullName);

            // find a default name
            string defaultName = FindNonExistingFile(Path.Combine(projectDir, DefaultName + "{0}" + Extension));
            defaultName = Path.GetFileNameWithoutExtension(defaultName);

            // ask the user for a name
            InputBox d = new InputBox("Add " +FileDescription, "Please provide a name for the " + FileDescription + ":", defaultName);
            if (d.ShowDialog() == DialogResult.OK)  
            {
                string filePrefix = Path.Combine(projectDir, d.Input);
                
                string dslFileName = filePrefix + Extension;
                string diagramFileName = dslFileName + Const.FileExtensionDiagramAddition;

                // check for existing name / file and cancel
                if (File.Exists(dslFileName))
                {
                    MessageBox.Show("A " + FileDescription + " with the name '" + d.Input + Extension + "' already exists.","Error adding " + FileDescription,MessageBoxButtons.OK,MessageBoxIcon.Exclamation);
                    return;
                }
                
                // create empty files
                File.Create(dslFileName).Close();
                File.Create(diagramFileName).Close();
                
                // create tt file
                string ttFileName = null;
                if (content != null)
                {
                    ttFileName = Path.Combine(Path.Combine(projectDir, "GeneratedCode"), d.Input) + ".tt";
                    if (!Directory.Exists(Path.GetDirectoryName(ttFileName)))
                        Directory.CreateDirectory(Path.GetDirectoryName(ttFileName));
                    StreamWriter ttFile = new StreamWriter(ttFileName);
                    ttFile.Write(content(project, dslFileName));
                    ttFile.Close();
                }

                // add to the project
                ProjectItem dslItem = project.ProjectItems.AddFromFile(dslFileName);
                project.ProjectItems.AddFromFile(diagramFileName);
                if ( ttFileName != null)
                    project.ProjectItems.AddFromFile(ttFileName);

                // open the newly created file
                Window w = dslItem.Open(Constants.vsViewKindPrimary);
                w.Activate();
            }
        }

        /// <summary>
        /// Creates the content of the tt file for new Dataflows.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <param name="dslFileName">Name of the DSL file.</param>
        /// <returns></returns>
        private static string DataflowTtContent(Project project, string dslFileName)
        {
            // get the name of the hardware description
            string hardwareFileName = GetHardwareInfoFileName(project);

            // make the file names relativ to the template
            hardwareFileName = PathUtil.RelativePathTo( Path.GetDirectoryName(dslFileName) , hardwareFileName);
            dslFileName = Path.GetFileName(dslFileName);
            
            // get the tt content from resources
            StreamReader sr = new StreamReader(new MemoryStream(Resources.Dataflow));
            string s = sr.ReadToEnd();
            sr.Close();

            // insert the file names to the template
            s = s.Replace("{{{dataflow}}}", dslFileName);
            s = s.Replace("{{{hardware}}}", hardwareFileName);
            
            return s;
        }

        /// <summary>
        /// Creates the content of the tt file for new Datastructures.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <param name="dslFileName">Name of the DSL file.</param>
        /// <returns></returns>
        private static string DatastructuresTtContent(Project project, string dslFileName)
        {
            // get the name of the hardware description
            string hardwareFileName = GetHardwareInfoFileName(project);

            // make the file names relativ to the template
            hardwareFileName = PathUtil.RelativePathTo(Path.GetDirectoryName(dslFileName), hardwareFileName);
            dslFileName = Path.GetFileName(dslFileName);

            // get the tt content from resources
            StreamReader sr = new StreamReader(new MemoryStream(Resources.Datastruct));
            string s = sr.ReadToEnd();
            sr.Close();

            // insert the file names to the template
            s = s.Replace("{{{datastruct}}}", dslFileName);
            s = s.Replace("{{{hardware}}}", hardwareFileName);

            return s;
        }

        /// <summary>
        /// Gets the name of the hardware info file of the given project.
        /// </summary>
        /// <param name="project">The project.</param>
        /// <returns></returns>
        private static string GetHardwareInfoFileName(Project project)
        {
            // find the hardware info
            List<ProjectItem> items = project.FindAllFiles();
            ProjectItem hardware = items.FirstOrDefault(x => x.get_FileNames(0).EndsWith(Const.FileExtensionHardwareInfo, StringComparison.InvariantCultureIgnoreCase));
            if ( hardware == null )
                hardware = items.FirstOrDefault(x => x.get_FileNames(0).EndsWith(Const.FileExtensionHardwareDescription, StringComparison.InvariantCultureIgnoreCase));
            return hardware == null ? "ERROR_NO_HARDWAREINFO" : hardware.get_FileNames(0);
        }

        /// <summary>
        /// Finds a non existing file based on the template.
        /// </summary>
        /// <example>FindNonExistingFile(@"C:\output{0}.txt");</example>
        /// <param name="template">The template.</param>
        /// <returns></returns>
        private static string FindNonExistingFile(string template)
        {
            int i = 1;
            string filename;
            do
            {
                filename = string.Format(template, i++);
            } while (File.Exists(filename));

            return filename;
        }
    }
}