using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using BenjaminSchroeter.VsxTools;
using EnvDTE;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// Event Handler for the Rename Project Item Event
    /// </summary>
    /// <remarks>Will be registerd from the package.</remarks>
    internal class RenameProjectItemHandler
    {
        /// <summary>
        /// All File extensions this handler will observe
        /// </summary>
        private static readonly string[] HandelFileTypes = new[] { Const.FileExtensionDataFlow, 
                                                                    Const.FileExtensionDatastructures, 
                                                                    Const.FileExtensionHardwareDescription,
                                                                    Const.FileExtensionHardwareInfo };

        /// <summary>
        /// If a file with one of the extensions above will be renamed,
        /// other project files with the same name but extensions from 
        /// this list will be renamed, too.
        /// </summary>
        private static readonly string[] RenameAlsoFileTypes = new[] { "{0}.c", 
                                                                        "{0}.tt", 
                                                                        "{0}{1}" + Const.FileExtensionDiagramAddition }; // {0} = filename , {1} = extension of removed file

        public static void ProjectItemsEvents_ItemRenamed(ProjectItem item, string oldName)
        {
            Project project = item.ContainingProject;

            // only for our project type
            if (!project.IsScatterWebDslProject())
                return;

            // only for some file types (HandelFileTypes)
            string newName = item.get_FileNames(0);
            string newNameWithoutExtension = Path.GetFileNameWithoutExtension(newName);
            string newExtension = Path.GetExtension(newName);
            if (!HandelFileTypes.Any(n => newName.EndsWith(n, StringComparison.InvariantCultureIgnoreCase)))
                return;

            // create patterns to delete
            string oldNameWithoutExtension = Path.GetFileNameWithoutExtension(oldName);
            string oldExtension = Path.GetExtension(oldName);
            var pattersToRename = RenameAlsoFileTypes.Select(x => string.Format(x, oldNameWithoutExtension, oldExtension));

            // find all types to rename
            List<ProjectItem> allProjectItems = project.FindAllFiles();

            List<ProjectItem> toRename = new List<ProjectItem>();
            foreach (ProjectItem i in allProjectItems)
            {
                string filename = i.get_FileNames(0);
                if (pattersToRename.Any(x => filename.EndsWith(x, StringComparison.InvariantCultureIgnoreCase)))
                    toRename.Add(i);
            }

            // the actual renaming
            foreach (ProjectItem projectItem in toRename)
            {
                string newItemName = projectItem.Name.Replace(oldNameWithoutExtension, newNameWithoutExtension);
                newItemName = newItemName.Replace(oldExtension, newExtension);
                projectItem.Name = newItemName; 
            }

            // check all tt files and change the file name, too
            var allTtFiles = from i in allProjectItems
                             where i.get_FileNames(0).EndsWith(".tt", StringComparison.InvariantCultureIgnoreCase)
                             select i;
            foreach (ProjectItem file in allTtFiles)
                ReplaceInTt(file.get_FileNames(0), oldName, Path.GetFileName(newName));
        }

        /// <summary>
        /// Replaces a filename in tt file
        /// </summary>
        /// <param name="ttFile">The tt file.</param>
        /// <param name="name">The old name.</param>
        /// <param name="newName">The new name.</param>
        private static void ReplaceInTt(string ttFile, string name, string newName)
        {
            if ( !File.Exists(ttFile))
                return;

            StreamReader sr = null;
            StreamWriter sw = null;

            try
            {
                sr = new StreamReader(ttFile);
                string content = sr.ReadToEnd();
                Encoding enc = sr.CurrentEncoding;
                sr.Close();
                if (content.Contains(name + "'"))
                {
                    content = content.Replace(name + "'", newName + "'");
                    sw = new StreamWriter(ttFile, false, enc);
                    sw.Write(content);
                    sw.Close();
                }
            }
            catch { }
            finally
            {
                if (sr != null)
                    sr.Close();
                if (sw != null)
                    sw.Close();
            }
        }
    }
}
