using System;
using System.Collections.Generic;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.VsxTools;
using DiveDeeper.VsxLibrary.Core;
using EnvDTE;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.TextManager.Interop;
using Microsoft.VisualStudio.TextTemplating.VSHost;
using OutputWindow=DiveDeeper.VsxLibrary.Core.OutputWindow;
using OutputWindowPane=DiveDeeper.VsxLibrary.Core.OutputWindowPane;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// Comamnd Handler for the "Generate Code" button
    /// in the toolbar of the solution explorer
    /// </summary>
    internal class TransformTemplatesHandler
    {
        public const int cmdidTransformTemplates = 0x402C;
        public const string guidTransformTemplatesString = "CA6F3901-D96B-41FC-8B0B-DB0AB458A08E";
        public static readonly Guid guidTransformTemplates = new Guid(guidTransformTemplatesString);

        private static readonly object lockObj = new object();

        /// <summary>
        /// Queries the status of the command
        /// </summary>
        /// <remarks>This command is only visible when a Scatter Web Dsl Project is selected in the solution explorer.</remarks>
        /// <param name="sender">The sender.</param>
        /// <param name="serviceProvider">The service provider.</param>
        public static void QueryStatus(object sender, IServiceProvider serviceProvider)
        {
            MenuCommand cmd = sender as MenuCommand;
            Debug.Assert(cmd != null, "sender is not a MenuCommand");

            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");
            IEnumerable<Project> projects = GetProjects(dte);

            foreach (Project project in projects)
                if (project.IsScatterWebDslProject())
                {
                    cmd.Visible = true;
                    return;
                }

            cmd.Visible = false;
        }

        /// <summary>
        /// Invokes the command.
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        public static void Invoke(IServiceProvider serviceProvider)
        {
            // use locking to avoid multiple instance running at the same time
            lock (lockObj)
            {
                // first do the validation
                IGlobalValidation v = serviceProvider.GetService(typeof(SGlobalValidation)) as IGlobalValidation;
                Debug.Assert(v != null, "IGlobalValidation");
                if ( v.Validate() > 0 )
                    if ( MessageBox.Show("There were validation errors (see Error List).\n\nDo you realy want to generate code? The code may not compile!","Validation",MessageBoxButtons.YesNo,MessageBoxIcon.Question,MessageBoxDefaultButton.Button2) == DialogResult.No)
                        return;

                // now generate code
                TransformTemplatesHandler h = new TransformTemplatesHandler();
                h.TransformTemplates(serviceProvider);                
            }
        }

        #region Generator
        private DTE dte;
        private ITextTemplating templatingService;
        private IVsRunningDocumentTable rdt;

        /// <summary>
        /// Transforms all templates in the selected projects 
        /// or if no projects selected use all projects
        /// </summary>
        /// <param name="serviceProvider">The service provider.</param>
        private void TransformTemplates(IServiceProvider serviceProvider)
        {
            // get all needed services
            dte = serviceProvider.GetService(typeof (DTE)) as DTE;
            Debug.Assert(dte != null, "DTE");

            templatingService = serviceProvider.GetService(typeof (STextTemplating)) as ITextTemplating;
            Debug.Assert(templatingService != null, "ITextTemplating");

            rdt = serviceProvider.GetService(typeof (IVsRunningDocumentTable)) as IVsRunningDocumentTable;
            Debug.Assert(rdt != null, "IVsRunningDocumentTable");

            // init (and clear) output pane
            OutputWindowPane pane = OutputWindow.CreatePane(typeof (MyDebugPane));
            pane.Activate();
            pane.Clear();
            dte.ExecuteCommand("View.Output", "");

            // if the same error in a texttemplate occures multiple time supress the duplicates
            templatingService.BeginErrorSession();

            // iterate all the projects
            foreach (Project project in GetProjects(dte))
            {
                // but only work on Scatter Web Dsl projects
                if (!project.IsScatterWebDslProject())
                    continue;

                pane.WriteLine("=== Processing templates of project {0} ===", project.Name);

                // save all open files (only in the selected projects)
                rdt.SaveAllDocumentsFromProject(project, true);

                // find all children of the current project
                List<ProjectItem> items = project.FindAllFiles();

                // for all templates
                var templates = from i in items
                                where i.get_FileNames(0).EndsWith(".tt", StringComparison.InvariantCultureIgnoreCase)
                                select i;
                foreach (ProjectItem item in templates)
                {
                    // now, finally, do transforming of one template by the time
                    pane.Write("Transforming {0} ...", item.Name);
                    if (TransformProjectItem(item))
                        pane.WriteLine(" done.");
                    else
                        pane.WriteLine(" ERROR.");

                }
                pane.WriteLine("");
            }


            templatingService.EndErrorSession();
        }

        /// <summary>
        /// Transforms the project item (that must be a tt file).
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns></returns>
        private bool TransformProjectItem(ProjectItem item)
        {
            // get the content of the template
            string content;
            string ttFileName = item.get_FileNames(0);
            using(StreamReader reader = new StreamReader(ttFileName))
            {
                content = reader.ReadToEnd();
                reader.Close();
            }

            // process the template
            TextTemplatingCallback templatingCallback = new TextTemplatingCallback();
            string output = templatingService.ProcessTemplate(ttFileName, content, templatingCallback, null);

            if ( templatingCallback.FileExtension == null ) // that appends if the user cancels
                return false;

            // create new filename (same as the template but with oter extension)
            string newFileName = Path.ChangeExtension(ttFileName, templatingCallback.FileExtension);

            bool contentWritten = false;

            // if the file is open in VS, write to this open file
            // search the RDT for the output file
            IVsPersistDocData docData = rdt.GetDocDataForOpenDocument(newFileName) as IVsPersistDocData;
            IVsTextLines textDoc = docData as IVsTextLines;
            if ( docData != null && textDoc != null )
            {
                // the file is currently open in a text editor
                object e;
                textDoc.CreateEditPoint(0, 0, out e);
                EditPoint ep = e as EditPoint;
                if (ep != null)
                {
                    // replace content of the file in the open editor
                    ep.Delete(int.MaxValue);
                    ep.Insert(output);
                    contentWritten = true;
                }

                int y;
                string x;
                docData.SaveDocData(VSSAVEFLAGS.VSSAVE_SilentSave, out x, out y);
            }
            
            // otherwise (the file is not open in VS) write to the file on disk
            if (!contentWritten)
                using (StreamWriter writer = new StreamWriter(newFileName, false, templatingCallback.Encoding))  
                {
                    writer.Write(output);
                    writer.Flush();
                    writer.Close();
                }

            // search for file in this project and maybe add it
            List<ProjectItem> items = item.ContainingProject.FindAllFiles();
            ProjectItem fileInProject = items.FirstOrDefault(i => i.get_FileNames(0).Equals(newFileName, StringComparison.InvariantCultureIgnoreCase));
            if (fileInProject == null)
            {
                // add this file to the project
                item.ContainingProject.ProjectItems.AddFromFile(newFileName);

                // and save the project file (the main.tt template will load it - if not saved it will miss this file)
                item.ContainingProject.Save(null);
            }

            return templatingCallback.ErrorCount == 0;
        }

        /// <summary>
        /// Gets the selected projects of the current solution
        /// or if no projects selected, get all projects.
        /// </summary>
        /// <param name="dte">The DTE.</param>
        /// <returns></returns>
        private static IEnumerable<Project> GetProjects(DTE dte)
        {
            object[] projects = dte.ActiveSolutionProjects as object[];
            if (projects != null)
            {
                if ( projects.Length > 0 )
                    foreach (Project p in projects)
                        yield return p;
                else
                    foreach (Project p in dte.Solution.Projects)
                        if ( p.Object != null) // a filter for the virtual project "misc files"
                            yield return p;
            }
        }

        [PaneName("Transform Templates")]
        [InitiallyVisible(true)]
        [ThreadSafe(true)]
        [System.Runtime.InteropServices.Guid("4BBC590A-CBB0-4bdc-AA9A-5ACA7F6DF993")]
        private class MyDebugPane : OutputPaneDefinition { }
        #endregion
    }
}
