using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using BenjaminSchroeter.Flow.HardwareDescription;
using BenjaminSchroeter.VsxTools;
using EnvDTE;
using Microsoft.VisualStudio.Zip;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    /// <summary>
    /// This service provides the UI and Project interaction
    /// to import a driver package.
    /// The Hardware Dsl Editor uses this service when the user
    /// requests to add a driver. The Service must determine which driver
    /// (with a file dialog or whatever?), add all files to the
    /// Project and than provides the .driverinfo file for the Hardware 
    /// Description Dsl Editor.    /// </summary>
    class ImportDriverService : SImportDriverService, IImportDriverService, IDisposable
    {
        private readonly IServiceProvider serviceProvider;
        private ZipFileDecompressor decomp;
        private Project project;

        public ImportDriverService(IServiceProvider serviceProvider)
        {
            this.serviceProvider = serviceProvider;
        }

        /// <summary>
        /// Opens a driver archiv after showing a open file dialog
        /// and returns the filename of a .driverinfo file.
        /// </summary>
        /// <returns>Filename of the .driverinfo file from the driver archive or "" if no driver should be loaded.</returns>
        public string GetDriverInfo()
        {
            // find the project and the driver directory
            DTE dte = serviceProvider.GetService(typeof(DTE)) as DTE;
            object[] projects = dte.ActiveSolutionProjects as object[];
            if (projects.Count() != 1)
                return "";
            project = projects[0] as Project;
            if (project == null)
                return "";

            // == ask for the driverInfo file
            OpenFileDialog d = new OpenFileDialog();
            d.Filter = "*.flowdrv|*.flowdrv|*.*|*.*";
            if (d.ShowDialog() != DialogResult.OK)
                return "";

            // == try open the zip file
            decomp = new ZipFileDecompressor(d.FileName, false);
            opened = true;
            // lets find the .driverinfo file
            IEnumerable<ZipEntry> driverInfos = from e in this.decomp.ZipFileEntries.Cast<ZipEntry>()
                             where !e.IsADirectory
                                   && e.FileName.EndsWith(Const.FileExtensionDriverInfo,StringComparison.InvariantCultureIgnoreCase)
                             select e;
            // if there is no .driverinfo, maybe we can find a .driver file
            if ( driverInfos.Count() == 0)
                driverInfos = from e in this.decomp.ZipFileEntries.Cast<ZipEntry>()
                                 where !e.IsADirectory
                                       && e.FileName.EndsWith(Const.FileExtensionDriverDescription ,StringComparison.InvariantCultureIgnoreCase)
                                 select e;
            // realy no file found -> exit
            if ( driverInfos.Count() != 1)
                return "";

            // extract the driver file to the temp directory
            string driverInfoFile = Path.Combine(Path.GetTempPath(), Path.GetFileName(driverInfos.First().FileName));
            this.decomp.UncompressZipEntry(driverInfos.First(),Path.GetTempPath(),true);

            return driverInfoFile;
        }

        /// <summary>
        /// Imports the driver source code files to the project.
        /// </summary>
        public void ImportDriverToProject()
        {
            string driversDir = Path.Combine(Path.GetDirectoryName(project.FullName), "Drivers");

            // extract all other files to the drivers directory
            if (!Directory.Exists(driversDir))
                Directory.CreateDirectory(driversDir);
            var fileEntries = from e in this.decomp.ZipFileEntries.Cast<ZipEntry>()
                              where !e.IsADirectory
                                    && !e.FileName.EndsWith(Const.FileExtensionDriverInfo, StringComparison.InvariantCultureIgnoreCase)
                                    && !e.FileName.EndsWith(Const.FileExtensionDriverDescription, StringComparison.InvariantCultureIgnoreCase)
                                    && !e.FileName.EndsWith(Const.FileExtensionDiagramAddition, StringComparison.InvariantCultureIgnoreCase)
                              select e;
            string tempfile = Path.GetTempFileName();
            File.Delete(tempfile);
            Directory.CreateDirectory(tempfile);

            foreach (ZipEntry entry in fileEntries)
            {
                decomp.UncompressZipEntry(entry,tempfile,true);
                string fileName = Path.Combine(tempfile, entry.FileName);
                string newFileName = Path.Combine(driversDir,Path.GetFileName(entry.FileName));
                if ( File.Exists(newFileName) )
                    File.Delete(newFileName);
                File.Move(fileName,newFileName);
            
                if (!project.FindAllFiles().Any( x => x.get_FileNames(0).Equals(newFileName,StringComparison.InvariantCultureIgnoreCase) ))
                    project.ProjectItems.AddFromFile(newFileName);
            }
        }

        private bool opened = false;
        public void Close()
        {
            if ( decomp != null && opened )
            {
                decomp.Close();
                opened = false;
            }
        }

        public void Dispose()
        {
            this.Close();
        }
    }
}
