using System;
using System.Collections.Generic;
using System.Windows.Forms;
using BenjaminSchroeter.VsxTools;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    class RdtListener : IVsRunningDocTableEvents3, IDisposable
    {
        // copies of the guid defined in the three packages
        private const string DataFlowEditorFactoryId = "29bba8ff-8f30-4464-b6b1-0484e3f27bbc";
        private const string DatastructuresEditorFactoryId = "914e7427-f8c6-4815-a0ce-00e1176d798f";
        private const string HardwareDescriptionEditorFactoryId = "640e3c2c-eeb4-429b-9b91-1b3578206b46";

        private readonly Dictionary<Guid, string> exclusiveDocData;

        private IVsRunningDocumentTable rdt;

        /// <summary>
        /// Initializes a new instance of the <see cref="RdtListener"/> class.
        /// </summary>
        /// <param name="rdt">The RDT.</param>
        public RdtListener(IVsRunningDocumentTable rdt)
        {
            this.rdt = rdt;

            exclusiveDocData = new Dictionary<Guid, string>();
            exclusiveDocData.Add(new Guid(HardwareDescriptionEditorFactoryId), "Hardware Description Model");
            exclusiveDocData.Add(new Guid(DatastructuresEditorFactoryId), "Datastructures Model");
            exclusiveDocData.Add(new Guid(DataFlowEditorFactoryId), "Data Flow Model");
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            this.rdt = null;
        }

        /// <summary>
        /// Called before document window is shown.
        /// 
        /// I will check here the RDT for other open documents
        /// and if it is not allowed to open this document
        /// close it immediately.
        /// </summary>
        public int OnBeforeDocumentWindowShow(uint docCookie, int fFirstShow, IVsWindowFrame pFrame)
        {
            // only check the following on the first opening of a document
            if ( ! (fFirstShow == 1) )
                return VSConstants.S_OK;

            // get the DocData and Guid of the EditorFactory of the document that is opening
            object obj;
            pFrame.GetProperty((int) __VSFPROPID.VSFPROPID_DocData, out obj);
            DocData openingDocData = obj as DocData;
            if ( openingDocData == null )
                return VSConstants.S_OK;
            Guid openingDocDataEditorFactory;
            openingDocData.GetClassID(out openingDocDataEditorFactory);            

            // if the opening element is a member of the exclusive list
            if (exclusiveDocData.ContainsKey(openingDocDataEditorFactory))
            {
                bool allowOpening = true;
                // iterate all open documents
                foreach (IVsRunningDocumentTableExtensions.RdtInfo currentRdtInfo in rdt.GetDocDataForOpenDocuments())
                {   
                    // get the DocData and EditorFactory Guid of the current document in this loop
                    DocData currentDocData = currentRdtInfo.DocData as DocData;
                    if ( currentDocData == null )
                        continue;
                    Guid currentEditorFactory;
                    currentDocData.GetClassID(out currentEditorFactory);

                    // if the opening and the current document are of the same type: do nothing
                    if ( currentEditorFactory == openingDocDataEditorFactory )
                        continue;

                    // if the current document is also a member of the exclusive list
                    if ( exclusiveDocData.ContainsKey(currentEditorFactory) )
                    {
                        // forbit opening
                        allowOpening = false;
                        // exit the loop
                        break; 
                    }
                }

                // if opening is forbidden
                if (!allowOpening)
                {
                    // message to the user
                    string current = "";
                    List<string> otherList = new List<string>();
                    foreach (KeyValuePair<Guid, string> pair in exclusiveDocData)
                    {
                        if (pair.Key == openingDocDataEditorFactory)
                            current = pair.Value;
                        else
                            otherList.Add(pair.Value);
                    }
                    string other = string.Join(", ", otherList.ToArray());

                    string msg = string.Format("You can not open a {0} while models with anoter type ({1}) are already opened in the desginer.\n\nPlease close the other models first.", current, other);
                    MessageBox.Show(msg,"Opening " + current,MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    
                    // and close the opening document right now
                    pFrame.CloseFrame((uint) __FRAMECLOSE.FRAMECLOSE_NoSave);
                }
            }

            return VSConstants.S_OK;
        }

        #region more empty IVsRunningDocTableEvents3 Members

        public int OnAfterFirstDocumentLock(uint docCookie, uint dwRDTLockType, uint dwReadLocksRemaining,
                                            uint dwEditLocksRemaining)
        {
            return VSConstants.S_OK;
        }

        public int OnBeforeLastDocumentUnlock(uint docCookie, uint dwRDTLockType, uint dwReadLocksRemaining,
                                              uint dwEditLocksRemaining)
        {
            return VSConstants.S_OK;
        }

        public int OnAfterSave(uint docCookie)
        {
            return VSConstants.S_OK;
        }

        public int OnAfterAttributeChange(uint docCookie, uint grfAttribs)
        {
            return VSConstants.S_OK;
        }

        public int OnAfterDocumentWindowHide(uint docCookie, IVsWindowFrame pFrame)
        {
            return VSConstants.S_OK;
        }

        public int OnAfterAttributeChangeEx(uint docCookie, uint grfAttribs, IVsHierarchy pHierOld, uint itemidOld,
                                            string pszMkDocumentOld, IVsHierarchy pHierNew, uint itemidNew,
                                            string pszMkDocumentNew)
        {
            return VSConstants.S_OK;
        }

        public int OnBeforeSave(uint docCookie)
        {
            return VSConstants.S_OK;
        }

        #endregion
    }
}
