using System;
using System.ComponentModel.Design;
using System.Runtime.InteropServices;
using BenjaminSchroeter.Flow.DataFlow;
using BenjaminSchroeter.GlobalValidation;
using BenjaminSchroeter.Flow.HardwareDescription;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Modeling.Shell;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;

namespace BenjaminSchroeter.Flow.UiSupportPackage
{
    // package branding
    [InstalledProductRegistration(true, "", "", "")]

    // package load key
    [ProvideLoadKey("Standard", "1.0", "Flow UiSupportPackage", "Benjamin Schroeter", 200)]

    // services
    [ProvideService(typeof(SGlobalValidation))]
    [ProvideService(typeof(SCodeBlockEditor))]
    [ProvideService(typeof(SImportDriverService))]

    // option page
    [ProvideOptionPage(typeof(UiSupportPackageOptionPage), "Flow", "Options", 105, 110, true)]

    [PackageRegistration(UseManagedResourcesOnly = true)]
    [DefaultRegistryRoot("Software\\Microsoft\\VisualStudio\\9.0")]
    [ProvideMenuResource(1000, 33)]
    [ProvideAutoLoad("f1536ef8-92ec-443c-9ed7-fdadf150da82")] //Microsoft.VisualStudio.VSConstants.UICONTEXT_SolutionExists
    [Guid(UiSupportPackageConstants.guidVsPackagePkgString)]
    public sealed class UiSupportPackage : Package, IVsSolutionEvents3, IVsInstalledProduct
    {
        public UiSupportPackage()
        {
        }

        protected override void Initialize()
        {
            base.Initialize();

            // command handler
            OleMenuCommandService menuCommandService = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            AddAddDatastructuresHandler(menuCommandService, AddDslItemHandler.cmdidAddDatastructures);
            AddAddDatastructuresHandler(menuCommandService, AddDslItemHandler.cmdidAddDatastructures2);
            AddAddDataflowHandler(menuCommandService, AddDslItemHandler.cmdidAddDataflow);
            AddAddDataflowHandler(menuCommandService, AddDslItemHandler.cmdidAddDataflow2);
            AddAddEmptyDriverHandler(menuCommandService, AddDslItemHandler.cmdidAddEmptyDriver);
            AddAddEmptyDriverHandler(menuCommandService, AddDslItemHandler.cmdidAddEmptyDriver2);

            AddCreateFlowPlatformSetupHandler(menuCommandService, CreateFlowPlatformSetupHandler.cmdidCreatePlatformSetup);
            AddCreateFlowPlatformSetupHandler(menuCommandService, CreateFlowPlatformSetupHandler.cmdidCreatePlatformSetup2);

            AddTransformTemplatesHandler(menuCommandService);
            AddParseFirmwareHandler(menuCommandService);
            AddCreateFlowDriverPackageHandler(menuCommandService);
            AddValidateProjectHandler(menuCommandService);

            // register for project item events
            AdviseProjectItemEvents();
            // register for solution events
            AdviseSolutionEvents();

            // register services
            IServiceContainer serviceContainer = this;
            serviceContainer.AddService(typeof(SCodeBlockEditor), createDataService, true);
            serviceContainer.AddService(typeof(SGlobalValidation),createGlobalValidation,true);
            serviceContainer.AddService(typeof(SImportDriverService),createImportDriverService,true);

            
            AdviseRDTEvents();
        }

        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);

            // unregister the events
            UnAdviseProjectItemEvents();
            UnAdviseSolutionEvents();
            UnAdviseRDTEvents();
        }

        #region register command handler
        /// <summary>
        /// Adds the "add datastructures" handler.
        /// </summary>
        /// <remarks>On the solution "add item" contextmenu of the solutionexplorer</remarks>
        private void AddAddDatastructuresHandler(IMenuCommandService menuCommandService, int cmdId)
        {
            CommandID menuCommandID = new CommandID(AddDslItemHandler.guidAddDsl, cmdId);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { AddDslItemHandler.QueryStatusUserItems(sender, this); },
                delegate { AddDslItemHandler.AddDatastructes(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Adds the "add dataflow" handler.
        /// </summary>
        /// <remarks>On the solution "add item" contextmenu of the solutionexplorer</remarks>
        private void AddAddDataflowHandler(IMenuCommandService menuCommandService, int cmdId)
        {
            CommandID menuCommandID = new CommandID(AddDslItemHandler.guidAddDsl, cmdId);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { AddDslItemHandler.QueryStatusUserItems(sender, this); },
                delegate { AddDslItemHandler.AddDataflow(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }
        
        /// <summary>
        /// Adds the "add empty flow driver" handler.
        /// </summary>
        /// <remarks>On the solution "add item" contextmenu of the solutionexplorer</remarks>
        private void AddAddEmptyDriverHandler(IMenuCommandService menuCommandService, int cmdId)
        {
            CommandID menuCommandID = new CommandID(AddDslItemHandler.guidAddDsl, cmdId);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { AddDslItemHandler.QueryStatusVendorItems(sender, this, ShowHardwareVendorOptions()); },
                delegate { AddDslItemHandler.AddEmptyDriver(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }
        private bool ShowHardwareVendorOptions()
        {
            UiSupportPackageOptionPage page = this.GetDialogPage(typeof (UiSupportPackageOptionPage)) as UiSupportPackageOptionPage;
            if ( page == null )
                return false;

            return page.ShowHardwareVendorOptions;
        }

        /// <summary>
        /// Adds the "transform templates" handler.
        /// </summary>
        /// <remarks>In the toolbar of the solution explorer.</remarks>
        /// <param name="menuCommandService">The menu command service.</param>
        private void AddTransformTemplatesHandler(IMenuCommandService menuCommandService)
        {
            CommandID menuCommandID = new CommandID(TransformTemplatesHandler.guidTransformTemplates, TransformTemplatesHandler.cmdidTransformTemplates);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { TransformTemplatesHandler.QueryStatus(sender, this); },
                delegate { TransformTemplatesHandler.Invoke(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Adds the "validate all" handler.
        /// </summary>
        /// <remarks>In the toolbar of the solution explorer.</remarks>
        /// <param name="menuCommandService">The menu command service.</param>
        private void AddValidateProjectHandler(IMenuCommandService menuCommandService)
        {
            CommandID menuCommandID = new CommandID(ValidateProjectHandler.guidValidateProject, ValidateProjectHandler.cmdidValidateProject);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { ValidateProjectHandler.QueryStatus(sender, this); },
                delegate { ValidateProjectHandler.Invoke(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Adds the "parse firmware" handler.
        /// </summary>
        /// <remarks>On the contextmenu of the solution explorer for .hardware and .driver files</remarks>
        /// <param name="menuCommandService">The menu command service.</param>
        private void AddParseFirmwareHandler(IMenuCommandService menuCommandService)
        {
            CommandID menuCommandID = new CommandID(ParseFirmwareHandler.guidParseFirmware, ParseFirmwareHandler.cmdidParseFirmware);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { ParseFirmwareHandler.QueryStatus(sender, this); },
                delegate { ParseFirmwareHandler.Invoke(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Adds the "Create Flow Platform Setup" handler.
        /// </summary>
        /// <remarks>On the contextmenu of the solution explorer for projects and in the Project main menu</remarks>
        /// <param name="menuCommandService">The menu command service.</param>
        /// <param name="cmdId">The CMD id.</param>
        private void AddCreateFlowPlatformSetupHandler(IMenuCommandService menuCommandService, int cmdId)
        {
            CommandID menuCommandID = new CommandID(CreateFlowPlatformSetupHandler.guidCreatePlatformSetup,cmdId);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { CreateFlowPlatformSetupHandler.QueryStatus(sender, this); },
                delegate { CreateFlowPlatformSetupHandler.Invoke(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }

        /// <summary>
        /// Adds the "create flow driver package" handler.
        /// </summary>
        /// <remarks>On the contextmenu of the solution explorer for .driver files</remarks>
        /// <param name="menuCommandService">The menu command service.</param>
        private void AddCreateFlowDriverPackageHandler(IMenuCommandService menuCommandService)
        {
            CommandID menuCommandID = new CommandID(CreateFlowPlatformSetupHandler.guidCreatePlatformSetup, CreateFlowDriverPackageHandler.cmdidCreateDriverPackage);
            DynamicStatusMenuCommand menuItem = new DynamicStatusMenuCommand(
                delegate(object sender, EventArgs e) { CreateFlowDriverPackageHandler.QueryStatus(sender, this); },
                delegate { CreateFlowDriverPackageHandler.Invoke(this); },
                menuCommandID);
            menuCommandService.AddCommand(menuItem);
        }
        #endregion

        #region ProjectItemEvents Members
        private EnvDTE.ProjectItemsEvents projectItemsEvents;

        private EnvDTE.ProjectItemsEvents ProjectItemsEvents
        {
            get
            {
                if (projectItemsEvents == null)
                {
                    EnvDTE.DTE dte = GetService(typeof(EnvDTE.DTE)) as EnvDTE.DTE;
                    if (dte != null)
                        projectItemsEvents = dte.Events.GetObject("ProjectItemsEvents") as EnvDTE.ProjectItemsEvents;
                }

                return projectItemsEvents;
            }
        }

        private void AdviseProjectItemEvents()
        {
            if (ProjectItemsEvents != null)
            {
                //ProjectItemsEvents.ItemAdded += new EnvDTE._dispProjectItemsEvents_ItemAddedEventHandler(ProjectItemsEvents_ItemAdded);
                ProjectItemsEvents.ItemRemoved += RemoveProjectItemHandler.ProjectItemsEvents_ItemRemoved;
                ProjectItemsEvents.ItemRenamed += RenameProjectItemHandler.ProjectItemsEvents_ItemRenamed;
            }
        }

        private void UnAdviseProjectItemEvents()
        {
            if (ProjectItemsEvents != null)
            {
                //ProjectItemsEvents.ItemAdded -= ProjectItemsEvents_ItemAdded;
                ProjectItemsEvents.ItemRemoved -= RemoveProjectItemHandler.ProjectItemsEvents_ItemRemoved;
                ProjectItemsEvents.ItemRenamed -= RenameProjectItemHandler.ProjectItemsEvents_ItemRenamed;
            }
        }
        #endregion

        #region SolutionEvents Members
        private uint solutionEventsCookie;

        private void AdviseSolutionEvents()
        {
            Microsoft.VisualStudio.Shell.Interop.IVsSolution solution = GetService(typeof(SVsSolution)) as IVsSolution;
            if (solution != null) solution.AdviseSolutionEvents(this, out solutionEventsCookie);
        }

        private void UnAdviseSolutionEvents()
        {
            Microsoft.VisualStudio.Shell.Interop.IVsSolution solution = GetService(typeof(SVsSolution)) as IVsSolution;

            if (solution != null && solutionEventsCookie != 0)
            {
                solution.UnadviseSolutionEvents(solutionEventsCookie);
                solutionEventsCookie = 0;
            }
        }

        public int OnBeforeCloseProject(IVsHierarchy pHierarchy, int fRemoved)
        {
            ClearGlobalValidationErrorList();
            return VSConstants.S_OK;
        }

        public int OnBeforeCloseSolution(object pUnkReserved)
        {
            ClearGlobalValidationErrorList();
            return VSConstants.S_OK;
        }

        #region empty default implementation
        public int OnAfterCloseSolution(object pUnkReserved)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterLoadProject(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pStubHierarchy, Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pRealHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterOpenProject(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy, int fAdded)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterOpenSolution(object pUnkReserved, int fNewSolution)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnBeforeUnloadProject(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pRealHierarchy, Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pStubHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnQueryCloseProject(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy, int fRemoving, ref int pfCancel)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnQueryCloseSolution(object pUnkReserved, ref int pfCancel)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnQueryUnloadProject(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pRealHierarchy, ref int pfCancel)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterClosingChildren(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterMergeSolution(object pUnkReserved)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnAfterOpeningChildren(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnBeforeClosingChildren(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }

        public int OnBeforeOpeningChildren(Microsoft.VisualStudio.Shell.Interop.IVsHierarchy pHierarchy)
        {
            return Microsoft.VisualStudio.VSConstants.S_OK;
        }
        #endregion
        #endregion

        #region service
        #region global validation
        /// <summary>
        /// Creates the SGlobalValidation service
        /// </summary>
        /// <param name="container">The container.</param>
        /// <param name="serviceType">Type of the service.</param>
        /// <returns></returns>
        private object createGlobalValidation(IServiceContainer container, Type serviceType)
        {
            if (container != this)
                return null;

            if (typeof(SGlobalValidation) == serviceType)
                return new GlobalValidationService(this);

            return null;
        }

        private void ClearGlobalValidationErrorList()
        {
            IGlobalValidation v = this.GetService(typeof(SGlobalValidation)) as IGlobalValidation;
            if (v != null)
                v.ClearErrorListItems();
        }
        #endregion
        private object createDataService(IServiceContainer container, Type serviceType)
        {
            if (container != this)
                return null;

            if (serviceType == typeof(SCodeBlockEditor))
                return new CodeBlockEditor(this);

            return null;
        }

        private object createImportDriverService(IServiceContainer container, Type serviceType)
        {
            if (container != this)
                return null;

            if (serviceType == typeof(SImportDriverService))
                return new ImportDriverService(this);

            return null;
        }
        #endregion

        #region Implementation of IVsInstalledProduct
        public int OfficialName(out string pbstrName)
        {
            pbstrName = "FLOW - data driven development";
            return VSConstants.S_OK;
        }

        public int ProductID(out string pbstrPID)
        {
            pbstrPID = "1.0";
            return VSConstants.S_OK;
        }

        public int ProductDetails(out string pbstrProductDetails)
        {
            pbstrProductDetails =
@"FLOW - data driven development for embedded systems
Extensions for Visual Studio 2008 to create software for embeded sensor nodes in a data driven way by Benjamin Schrter 2008.
http://flow.irgendwie.net";
            return VSConstants.S_OK;
        }

        public int IdIcoLogoForAboutbox(out uint pIdIco)
        {
            pIdIco = 32121;
            return VSConstants.S_OK;
        }

        public int IdBmpSplash(out uint pIdBmp)
        {
            IdIcoLogoForAboutbox(out pIdBmp);
            return VSConstants.S_OK;
        }
        #endregion

        #region RunningDocTableEvents Members
        private uint rdtEventsCookie;
        private RdtListener rdtListener;

        private void AdviseRDTEvents()
        {
            // when i'm debugging and so i do not want this functionality ;-)
            if ( IsDesignTimeRunMode() )
                return;
            
            Microsoft.VisualStudio.Shell.Interop.IVsRunningDocumentTable rdt = GetService(typeof(SVsRunningDocumentTable)) as IVsRunningDocumentTable;
            rdtListener = new RdtListener(rdt);
            rdt.AdviseRunningDocTableEvents(rdtListener, out rdtEventsCookie);
        }

        private void UnAdviseRDTEvents()
        {
            Microsoft.VisualStudio.Shell.Interop.IVsRunningDocumentTable rdt = GetService(typeof(SVsRunningDocumentTable)) as IVsRunningDocumentTable;

            if (rdt != null && rdtEventsCookie != 0)
            {
                rdt.UnadviseRunningDocTableEvents(rdtEventsCookie);
                rdtEventsCookie = 0;

                rdtListener.Dispose();
                rdtListener = null;
            }
        }
        
        protected bool IsDesignTimeRunMode()
        {
            int num;
            string str;
            IVsAppCommandLine service = (IVsAppCommandLine)GetService(typeof(IVsAppCommandLine));
            ErrorHandler.ThrowOnFailure(service.GetOption("DesignTimeRun", out num, out str));
            return (num == 1);
        }
       #endregion
    }
}